/*
 * Copyright (C) 2007 Brocade Communications
 *
 * This is low level harware sempahore driver.
 */

/* Usual kernel heasders */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>	
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/major.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/spinlock.h>

/* Header specific to Brocade */
#include <brocade/sin.h>
#include <brocade/platform.h>
#include <brocade/swbd.h>
#include <platforms/4xx/silkworm_cpld.h>
#include <brocade/cpld.h>
#include <brocade/silkworm_hwsem.h>

/* This is a database of hw sem params based on bd number */
struct _silkworm_hwsem_param_ {
	int		swbd;		/* if it is supported for this bd */
	int		capable;	/* what can be done with sema4 */

	/* These represent offsets to fpga base for various ops */
	uint 		lock_base;	/* offset to request lock */
	uint 		unlock_base;	/* offset to request unlock */
	uint 		reset_base;	/* offset to request reset */
	uint 		status_base;	/* offset to request status */
	uint 		index_base;	/* offset to select a sema4 */

	/* These are the values to be compared with */ 
	uint		available;	/* Value representing unlocked state */
	uint		unavailable;	/* Value representing other has locked state */
	uint		locked;		/* Value representing this has state */
	uint		invalid;	/* Value representing invalid state */
}; 

typedef struct _silkworm_hwsem_param_ silkworm_hwsem_param;

static silkworm_hwsem_param  sem_param[] =
{
	[0] =  {
		.swbd		= SWBD62,
		.capable	= SILKWORM_HWSEM_RESET_CAPABLE | SILKWORM_HWSEM_RESET_AT_BOOT,
		.lock_base	= 0x43,
		.unlock_base	= 0x44,
		.reset_base	= 0x45,
		.status_base	= 0x46,
		.index_base	= 0x47,
		.available	= 0x0,	/* Bits 00 */
		.unavailable	= 0x2,	/* Bits 10 */
		.locked		= 0x1,	/* Bits 01 Value for CPU ownership */
		.invalid	= 0x3,	/* Bits 11 */
	},

	[1] = {
		.swbd		= SWBD63,
		.capable	= 0,
		.lock_base	= 0x10,
		.unlock_base	= 0x11,
		.reset_base	= 0x12,
		.status_base	= 0x13,
		.index_base	= 0x14,
		.available	= 0x0,	/* Bits 00 */
		.unavailable	= 0x1,	/* Bits 01 */
		.locked		= 0x2,	/* Bits 10 Value for co-CPU */
		.invalid	= 0x3, 	/* Bits 11 */
	},

	[2] =  {
		.swbd		= SWBD77,
		.capable	= SILKWORM_HWSEM_RESET_CAPABLE | SILKWORM_HWSEM_RESET_AT_BOOT,
		.lock_base	= 0x43,
		.unlock_base	= 0x44,
		.reset_base	= 0x45,
		.status_base	= 0x46,
		.index_base	= 0x47,
		.available	= 0x0,	/* Bits 00 */
		.unavailable	= 0x2,	/* Bits 10 */
		.locked		= 0x1,	/* Bits 01 Value for CPU ownership */
		.invalid	= 0x3,	/* Bits 11 */
	},
};

static int hwsem_bd_index = -1;

static spinlock_t  hwsem_lock;

#define this_hwsem	(&sem_param[hwsem_bd_index])

/*
 * This routine initializes the hwsem interface
 * returns 0 on success -1 otherwise
 */
int __init silkworm_hwsem_init(int bd)
{
	int i;
	int max = (sizeof(sem_param) / sizeof (silkworm_hwsem_param));
	spin_lock_init(&hwsem_lock);
	for (i=0; i < max; ++i) {
		if (sem_param[i].swbd == bd) {
			hwsem_bd_index = i;
			break;
		}
	}

	if (hwsem_bd_index < 0) {
		printk(KERN_DEBUG "SILKWORM_HWSEM: This BD %d is not supported\n", bd);
		return -1;
	}
	
	printk(KERN_INFO "SILKWORM_HWSEM: capability %x index %d\n", this_hwsem->capable, hwsem_bd_index);
	/* Reset the semaphores if we are asked to do so on this BD */
	if ((this_hwsem->capable & SILKWORM_HWSEM_RESET_CAPABLE) &&
		(this_hwsem->capable & SILKWORM_HWSEM_RESET_AT_BOOT)) {

		printk(KERN_INFO "SILKWORM_HWSEM: Reseting all HW semaphores\n");

		for (i=0; i < SILKWORM_HWSEM_MAX; ++i)
			if (silkworm_hwsem_reset((unsigned char )i))
				printk(KERN_ERR "SILKWORM_HWSEM: Error reseting semaphore %d\n", i);
	}

	printk(KERN_INFO "SILKWORM_HWSEM: HW semaphores subsystem initialized for SWBD%d.\n", bd);

	return 0;
}

/*
 * silkworm_hwsem_trylock: Try to lock the HW semaphore
 *
 * This routine attempts to reserve a hardware smeaphore.
 *
 * Input	: semaphore number
 * Output	: none
 * Return 	: 0 on success -error otherwise
 * Error 	: -ENOSYS - BD not supported
 *       	: -EINVAL - Semaphore number is out of bound
 *       	: -EBUSY  - Semaphore is owned by other CPU
 *       	: -EFAULT - H/W Error occurred while reading semaphore
 */

int silkworm_hwsem_trylock(unsigned char semid)
{
	unsigned char  val;
	unsigned long flags;
	int ret = -ENOSYS;

	if (hwsem_bd_index < 0) {
		printk(KERN_ERR "SILKWORM_HWSEM: Lock: HW semaphore is not supported\n");
		return -ENOSYS;
	}

	if (semid >= SILKWORM_HWSEM_MAX)
		return -EINVAL;

	spin_lock_irqsave(&hwsem_lock, flags);

	/* Select the Semaphore number register */

	if (cpld_write_register(this_hwsem->index_base, semid)) {
		/* Error in write */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to write CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Try to own the semaphore */
	if (cpld_read_register(this_hwsem->lock_base, &val)) {
		/* Error in read */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to read CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Check if we were successful */
	if (cpld_read_register(this_hwsem->status_base, &val)) {
		/* Error in read */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to read CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Cleanup the return */
	val = val & 0x3;

	if (val == this_hwsem->locked) {
		/* Got the sema4 */
		ret = 0;
		goto exit_or_err;
	}

	if (val == this_hwsem->unavailable) {
		/* Locked by other processor */
		ret = -EBUSY;
		goto exit_or_err;
	}

	/* oterwise it is in invalid or unknown state */
	ret = -EFAULT;

exit_or_err:
	spin_unlock_irqrestore(&hwsem_lock, flags);
	return ret;
}

/*
 * silkworm_hwsem_unlock: unlock the HW semaphore
 *
 * This routine attempts to unlock a hardware smeaphore.
 *
 * Input	: semaphore number
 * Output	: none
 * Return 	: 0 on success -error otherwise
 * Error 	: -ENOSYS - BD not supported
 *       	: -EINVAL - Semaphore number is out of bound
 *       	: -EBUSY  - Semaphore is owned by other CPU
 *       	: -EFAULT - H/W Error occurred while reading semaphore
 */
int silkworm_hwsem_unlock(unsigned char semid)
{
	unsigned char  val;
	unsigned long flags;
	int ret = 0;

	if (hwsem_bd_index < 0) {
		printk(KERN_ERR "SILKWORM_HWSEM: UNLock: HW semaphore is not supported\n");
		return -ENOSYS;
	}

	if (semid >= SILKWORM_HWSEM_MAX)
		return -EINVAL;

	spin_lock_irqsave(&hwsem_lock, flags);

	/* Select the Semaphore number register */

	if (cpld_write_register(this_hwsem->index_base, semid)) {
		/* Error in write */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to write CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Verify if it is really owned by us */
	if (cpld_read_register(this_hwsem->status_base, &val)) {
		/* Error in read */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to read CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Cleanup the return */
	val = val & 0x3;

	if (val != this_hwsem->locked) {
		/* We do not own it or some error or got reset */
		printk(KERN_ERR "SILKWORM_HWSEM:: Not locked by this cpu %d\n", semid);
		ret = -EINVAL;
		goto exit_or_err;
	}

	/* Try to release the semaphore */
	if (cpld_read_register(this_hwsem->unlock_base, &val)) {
		/* Error in read */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to read CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Check if we were successful */
	if (cpld_read_register(this_hwsem->status_base, &val)) {
		/* Error in read */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to read CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Cleanup the return */
	val = val & 0x3;

	/* Check if it is still locked */
	if (val == this_hwsem->locked) {
		/* Still locked ? */
		printk(KERN_ERR "SILKWORM_HWSEM:: unable to unlock the semaphore %d\n", semid);
		ret = -EFAULT;
		goto exit_or_err;
	}

	ret = 0;

exit_or_err:
	spin_unlock_irqrestore(&hwsem_lock, flags);
	return ret;
}

/*
 * silkworm_hwsem_get_status: Get the current status of semaphore
 *
 * This routine reads and returns the status of a given semaphore
 *
 * Input	: semaphore number
 * Output	: status of the semaphore
 * Return 	: 0 on success -error otherwise
 * Error 	: -ENOSYS - BD not supported
 *       	: -EINVAL - Semaphore number is out of bound
 *       	: -EBUSY  - Semaphore is owned by other CPU
 *       	: -EFAULT - H/W Error occurred while reading semaphore
 */
int silkworm_hwsem_get_status(unsigned char semid, int *status)
{
	unsigned char  val;
	unsigned long flags;
	int ret = -ENOSYS;

	if (hwsem_bd_index < 0) {
		printk(KERN_ERR "SILKWORM_HWSEM: Lock: HW semaphore is not supported\n");
		return -ENOSYS;
	}

	if (semid >= SILKWORM_HWSEM_MAX)
		return -EINVAL;

	if (!status)
		return -EINVAL;

	spin_lock_irqsave(&hwsem_lock, flags);

	/* Select the Semaphore number register */

	if (cpld_write_register(this_hwsem->index_base, semid)) {
		/* Error in write */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to write CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Read the status */
	if (cpld_read_register(this_hwsem->status_base, &val)) {
		/* Error in read */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to read CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Cleanup the return */
	val = val & 0x3;

	ret = 0;

	if (val == this_hwsem->locked) {
		/* Locked by this processor */
		*status = SILKWORM_HWSEM_STATE_LOCKED_BY_THIS_CPU;
	}

	if (val == this_hwsem->unavailable) {
		/* Locked by other processor */
		*status = SILKWORM_HWSEM_STATE_LOCKED_BY_OTHER_CPU;
	}

	if (val == this_hwsem->available) {
		/* Not Locked by any processor */
		*status = SILKWORM_HWSEM_STATE_NOT_LOCKED;
	}

	if (val == this_hwsem->invalid) {
		/* Not Locked by any processor */
		*status = SILKWORM_HWSEM_STATE_INVALID;
	}

exit_or_err:
	spin_unlock_irqrestore(&hwsem_lock, flags);
	return ret;
}

/*
 * silkworm_hwsem_reset: Try to bust the HW semaphore
 *
 * This routine resets the semaphore
 *
 * Input	: semaphore number
 * Output	: none
 * Return 	: 0 on success -error otherwise
 * Error 	: -ENOSYS - BD not supported
 *       	: -EINVAL - Semaphore number is out of bound
 *       	: -EBUSY  - Semaphore is owned by other CPU
 *       	: -EFAULT - H/W Error occurred while reading semaphore
 *       	: -EPERM  - This CPU is not allowed to reset the semaphore
 */
int silkworm_hwsem_reset(unsigned char semid)
{
	unsigned char  val;
	unsigned long flags;
	int ret = -ENOSYS;

	if (hwsem_bd_index < 0) {
		printk(KERN_ERR "SILKWORM_HWSEM: Lock: HW semaphore is not supported\n");
		return -ENOSYS;
	}

	if (semid >= SILKWORM_HWSEM_MAX)
		return -EINVAL;

	if (!(this_hwsem->capable & SILKWORM_HWSEM_RESET_CAPABLE)) {
		printk(KERN_ERR "SILKWORM_HWSEM:: This CPU is not authorised to reset semaphore\n");
		return -EPERM;
	}

	spin_lock_irqsave(&hwsem_lock, flags);

	/* Select the Semaphore number register */

	if (cpld_write_register(this_hwsem->index_base, semid)) {
		/* Error in write */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to write CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* Try to reset the semaphore */
	if (cpld_read_register(this_hwsem->reset_base, &val)) {
		/* Error in read */
		printk(KERN_ERR "SILKWORM_HWSEM:: Unable to read CPLD HWSEM reg\n");
		ret = -EFAULT;
		goto exit_or_err;
	}

	/* We should not check the status as it is useless to know, other cpu might have acquired it again */

	ret = 0;

exit_or_err:
	spin_unlock_irqrestore(&hwsem_lock, flags);
	return ret;
}

EXPORT_SYMBOL(silkworm_hwsem_trylock);
EXPORT_SYMBOL(silkworm_hwsem_unlock);
EXPORT_SYMBOL(silkworm_hwsem_get_status);
EXPORT_SYMBOL(silkworm_hwsem_reset);
