/*
 * drivers/brcdecc/brcdecc_440epx.c
 *
 * Brocade's ECC reporting driver for AMCC/IBM 440 EPX
 *
 * Author: Jeff Haran (jharan@brocade.com)
 *
 * Copyright (c) 2009 Brocade
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/mm.h>
#include <linux/device.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/dma-mapping.h>
#include <linux/crc32.h>
#include <linux/proc_fs.h>
#include <asm/reg.h>
#include <asm/ibm44x.h>
#include <brocade/platform.h>
#include <brocade/silkworm.h>

/* in case DEBUG_PRINTKS gets defined in above includes */
#ifdef DEBUG_PRINTKS
#undef DEBUG_PRINTKS
#endif /* DEBUG_PRINTKS */

/*
 * remove DEBUG_PRINTKS define below to compile out debugging printks
 * in proc routines
 */
#define DEBUG_PRINTKS

/* DDR SDRAM controller register offsets and bit defines */

#define DDR0_00_OFFSET 0
#define DDR0_00_INTACK_MSK 0x7f000000
#define DDR0_00_INTSTA_MSK 0x007f0000
#define DDR0_00_INTSTA_MCE 0x00080000
#define DDR0_00_INTSTA_SCE 0x00040000

#define DDR0_22_OFFSET 22
#define DDR0_22_CTLRAW_MASK 0x03000000
#define DDR0_22_CTLRAW_ECC_ON 0x03000000

#define DDR0_23_OFFSET 23
#define DDR0_23_FWC 0x00000001  /* used to generate ECC errors */

#define DDR0_31_OFFSET 31
#define DDR0_31_XORCB_MSK 0x0000ffff /* used to generate ECC errors */

#define DDR0_38_OFFSET 38	/* lower 32 bits of address of correctable error */
/*
 * note there is also the msb of the 33 bit correctable address in
 * DDR0_39, but this code assumes that bit is always 0 and can thus
 * be ignored.
 */

/* module parameters */

static uint debug = 0;

/*
 * not a lot of agreement in existing kernel sources as to what you should
 * return if a proc entry create fails. I found examples using -ENOMEM,
 * -ENODEV and -EIO. -ENOENT seems the most descriptive to me.
 */
#define PROC_CREATE_FAIL (-ENOENT)

/* proc entry names and associated pointers */
#define ECC_PROC_DIR_NAME "driver/ecc"
static struct proc_dir_entry *ecc_dir = NULL;

#define ECC_PROC_MODEL_NAME "model"
static struct proc_dir_entry *proc_model = NULL;

#define ECC_PROC_DDR0_00_INTSTA_NAME "intsta"
static struct proc_dir_entry *proc_ddr0_00_intsta = NULL;

#define ECC_PROC_DDR0_00_INTSTA_CLEAR_NAME "intsta_clear"
static struct proc_dir_entry *proc_ddr0_00_intsta_clear = NULL;

#define ECC_PROC_ECC_ERR_INJECT_NAME "ecc_err_inject"
static struct proc_dir_entry *proc_ecc_err_inject = NULL;

/*
 * returns the value read from the DDR SDRAM controller register
 * indexed by the given offset.
 */

static u32 ddr_sdram_read(u8 offset)
{
	u32 value;

	/*
	 * FIXME: note there are no locks around the access to these DCRs.
	 * at the time of the writing of this module, the only place in the
	 * kernel sources where these registers were accessed was during
	 * platform initialization to determine where the end of memory is,
	 * which will be long over with by the time this module gets loaded. 
	 * so long as we are running single processor with kernel pre-emption
	 * disabled, this should be sufficient.
	 */

	mtdcr(DCRN_SDRAM0_CFGADDR, offset);
	value = mfdcr(DCRN_SDRAM0_CFGDATA);
	return value;
}

/*
 * writes value to the DDR SDRAM controller register indexed by offset
 */

static void ddr_sdram_write(u8 offset, u32 value)
{
	mtdcr(DCRN_SDRAM0_CFGADDR, offset);
	mtdcr(DCRN_SDRAM0_CFGDATA, value);
}

/*
 * The read side of the procfs interface is screwy.
 * Read the comment block titled:
 * "How to be a proc read function" in fs/proc/generic.c:proc_file_read()
 * for some hint as to why the read functions below do what they do.
 *
 * Note that these implementations make some assumptions on the behavior
 * of the user space code reading them:
 *
 * 1) They read sequentially from the beginning of the file, no lseek()s.
 * 2) Their reads are big enough to get all the data in one read() call.
 *
 * These will be valid assumptions for the daemon written to read
 * these and cat, but perhaps not for other apps. User beware.
 */

/*
 * copy the lesser of d_len and s_len bytes from p_src to p_dst.
 * if s_len bytes are copied, indicate an end of file condition at *eof.
 *
 * return the number of bytes copied.
 */

static int proc_read_copy(
	char *p_dst, int d_len, char *p_src, int s_len, int *eof)
{
	int len = 0;

	while (d_len > 0) {
		if (s_len <= 0) {
			if (eof) {
				*eof = 1;
			}
			break;
		}
		*p_dst++ = *p_src++;
		d_len--;
		s_len--;
		len++;
	}
	return len;
}

#define MODEL_NAME "AMCC_PPC440EPX"

/* return the processor model to the caller */

static int proc_model_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_model_read(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	len = proc_read_copy(page, count, MODEL_NAME, strlen(MODEL_NAME) + 1, eof);
	return len;
}

/* return the contents of the DDR0_00 register INTSTA field to the caller */

static int proc_ddr0_00_intsta_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	u32 ddr0_00;
	u32 intsta;
	char buf[16];

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"proc_ddr0_00_intsta_read(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	/* read the DDR0_00 register */
	ddr0_00 = ddr_sdram_read(DDR0_00_OFFSET);
	/* mask off all but the INTSTA field bits */
	intsta = ddr0_00 & DDR0_00_INTSTA_MSK;

	snprintf(buf, sizeof(buf), "%x", intsta);
	len = proc_read_copy(page, count, buf, strlen(buf) + 1, eof);
	return len;
}

void edac_mc_scrub_block(unsigned long page, unsigned long offset, u32 size);

/*
 * the DDR controller will correct the data that it reads and presents on
 * the bus, but it wont write the corrected data back out the the DIMM.
 * that's what this function is for.
 * borrows heavily from open source EDAC module.
 */

static void handle_sbe(void)
{
	u32 capture_address;
	unsigned long page_frame_number;
	unsigned long offset;

	/* read the physical address of the data with the single bit error */
	capture_address = ddr_sdram_read(DDR0_38_OFFSET);

	/* figure out what page frame number and offset into that page its at */
	page_frame_number = capture_address >> PAGE_SHIFT;
	offset = capture_address & (~PAGE_MASK);

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG
			"handle_sbe() address=0x%x, 0x%lx, 0x%lx\n",
				capture_address, page_frame_number, offset);
	}
#endif /* DEBUG_PRINTKS */

	/* read and write back the 8 bytes starting at that page and offset */
	edac_mc_scrub_block(page_frame_number, offset, 8);
}

/*
 * return the contents of the DDR0_00 register INTSTA field to the caller
 * and clear the contents of the register
 */

static int proc_ddr0_00_intsta_clear_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	u32 ddr0_00;
	u32 intsta;
	char buf[16];

#ifdef DEBUG_PRINTKS
	if (debug > 1) {
		printk(KERN_DEBUG "proc_ddr0_00_intsta_clear_read"
			"(0x%p, 0x%p, 0x%lx, %d, 0x%x, 0x%p)\n",
				page, start ? *start : (char *) 0xdeadbeef, off,
				count, eof ? *eof : 0xdeadbeef, data);
	}
#endif /* DEBUG_PRINTKS */

	/* if this is not the first read, signal EOF to caller */
	if (*start || off) {
		*eof = 1;
		return 0;
	}

	/* read the DDR0_00 register */
	ddr0_00 = ddr_sdram_read(DDR0_00_OFFSET);
	/* mask off all but the INTSTA field bits */
	intsta = ddr0_00 & DDR0_00_INTSTA_MSK;
	if (intsta) {
		/* if any correctable errors occurred, scrub them */
		if (intsta & (DDR0_00_INTSTA_MCE | DDR0_00_INTSTA_SCE)) {
			handle_sbe();
		}
		/*
		 * shift contents of INTSTA field over to INTACK field
		 * and write back out to clear.
		 */
		ddr_sdram_write(DDR0_00_OFFSET, (intsta << 8));
	}
	snprintf(buf, sizeof(buf), "%x", intsta);
	len = proc_read_copy(page, count, buf, strlen(buf) + 1, eof);
	return len;
}

/*
 * scan up to the first 8 chars of buf into their hex representation
 * and return the result.
 */

static u32 scan_u32(char *buf)
{
	u32 val = 0;
	int len = 0;
	char *p = buf;

	for ( ; len < 8; ++p, ++len ) {
		if ((*p >= '0') && (*p <= '9')) {
			val = (val << 4) + *p - '0';
			continue;
		}
		if ((*p >= 'a') && (*p <= 'f')) {
			val = (val << 4) + *p + 10 - 'a';
			continue;
		}
		if ((*p >= 'A') && (*p <= 'F')) {
			val = (val << 4) + *p + 10 - 'A';
			continue;
		}
		break;
	}
	return val;
}

/*
 * convert the user written string of hex characters to binary
 * and write to the DDR0_31 register XORCB field followed by a
 * write of 1 to the DDR0_23 FWC field. this causes the DDR
 * controller to inject an ECC error into the next write to DRAM.
 *
 * to use this, reference the table of syndrome codes in table 19-11
 * of the EPX manual.
 */

static int proc_ecc_err_inject_write(
	struct file *file, const char *buffer, unsigned long count, void *data)
{
	int len;
	u32 ecc_err_inject;
	u32 ddr0_23;
	char my_buf[16];

	if (count > (sizeof(my_buf) - 1)) {
		len = sizeof(my_buf) - 1;
	} else {
		len = count;
	}

	if (copy_from_user(my_buf, buffer, len)) {
		return -EFAULT;
	}

	my_buf[len] = 0;

	/* scan the desired syndrome bit codes from user input */
	ecc_err_inject = scan_u32(my_buf);

	/* write those codes to lower 16 bits of DDR0_31 */
	ddr_sdram_write(DDR0_31_OFFSET, (ecc_err_inject & DDR0_31_XORCB_MSK));

	/*
	 * write a 1 to DDR0_23 field FWC to force ECC to occur on
	 * the next write to DRAM. Note the manual says this bit is
	 * self clearing, so no need to set it back to 0.
	 */
	ddr0_23 = ddr_sdram_read(DDR0_23_OFFSET);
	ddr0_23 |= DDR0_23_FWC;
	ddr_sdram_write(DDR0_23_OFFSET, ddr0_23);

	return len;
}

/* deallocate/destroy stuff created in init routine */

static void brcd_ecc_cleanup(void)
{
	if (proc_ecc_err_inject) {
		remove_proc_entry(ECC_PROC_ECC_ERR_INJECT_NAME, ecc_dir);
		proc_ecc_err_inject = NULL;
	}

	if (proc_ddr0_00_intsta_clear) {
		remove_proc_entry(ECC_PROC_DDR0_00_INTSTA_CLEAR_NAME, ecc_dir);
		proc_ddr0_00_intsta_clear = NULL;
	}

	if (proc_ddr0_00_intsta) {
		remove_proc_entry(ECC_PROC_DDR0_00_INTSTA_NAME, ecc_dir);
		proc_ddr0_00_intsta = NULL;
	}

	if (proc_model) {
		remove_proc_entry(ECC_PROC_MODEL_NAME, ecc_dir);
		proc_model = NULL;
	}

	if (ecc_dir) {
		remove_proc_entry(ECC_PROC_DIR_NAME, NULL);
		ecc_dir = NULL;
	}
}


static int __init brcdecc_init_440epx(void)
{
	int rc = 0;
	u32 reg;

	printk(KERN_INFO "brcdecc_init_440epx() %u\n", debug);

	/* we break out of the following one pass "do" on errors */
	do {
		/* make sure the PROM enabled ECC */
		reg = ddr_sdram_read(DDR0_22_OFFSET);
		if ((reg & DDR0_22_CTLRAW_MASK) != DDR0_22_CTLRAW_ECC_ON) {
			printk(KERN_ERR "brcdecc_init_440epx() ECC not enabled 0x%x\n",
				reg);
			rc = -ENODEV;
			break;
		}

		/* create /proc/driver/ecc directory */
		ecc_dir = proc_mkdir(ECC_PROC_DIR_NAME, NULL);
		if (!ecc_dir) {
			printk(KERN_ERR "Failed to create ecc proc dir\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		ecc_dir->owner = THIS_MODULE;

		/* create ecc/model, readonly to everybody */
		proc_model = create_proc_entry(ECC_PROC_MODEL_NAME, 0444, ecc_dir);
		if (!proc_model) {
			printk(KERN_ERR "Failed to create ecc proc model\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_model->data = NULL;
		proc_model->owner = THIS_MODULE;
		proc_model->read_proc = proc_model_read;
		proc_model->write_proc = NULL;

		/* create ecc/intsta, readonly to everybody */
		proc_ddr0_00_intsta =
			create_proc_entry(ECC_PROC_DDR0_00_INTSTA_NAME, 0444, ecc_dir);
		if (!proc_ddr0_00_intsta) {
			printk(KERN_ERR "Failed to create ecc proc intsta\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_ddr0_00_intsta->data = NULL;
		proc_ddr0_00_intsta->owner = THIS_MODULE;
		proc_ddr0_00_intsta->read_proc = proc_ddr0_00_intsta_read;
		proc_ddr0_00_intsta->write_proc = NULL;

		/* create ecc/intsta_clear, readonly to root only */
		proc_ddr0_00_intsta_clear =
			create_proc_entry(ECC_PROC_DDR0_00_INTSTA_CLEAR_NAME,
				0400, ecc_dir);
		if (!proc_ddr0_00_intsta_clear) {
			printk(KERN_ERR "Failed to create ecc proc intsta_clear\n");
			rc = PROC_CREATE_FAIL;
			break;
		}
		proc_ddr0_00_intsta_clear->data = NULL;
		proc_ddr0_00_intsta_clear->owner = THIS_MODULE;
		proc_ddr0_00_intsta_clear->read_proc = proc_ddr0_00_intsta_clear_read;
		proc_ddr0_00_intsta_clear->write_proc = NULL;

		if (debug) {
			/* create ecc/ecc_err_inject, write-only to root */
			proc_ecc_err_inject =
				create_proc_entry(ECC_PROC_ECC_ERR_INJECT_NAME, 0200, ecc_dir);
			if (!proc_ecc_err_inject) {
				printk(KERN_ERR "Failed to create ecc proc ecc_err_inject\n");
				rc = PROC_CREATE_FAIL;
				break;
			}
			proc_ecc_err_inject->data = NULL;
			proc_ecc_err_inject->owner = THIS_MODULE;
			proc_ecc_err_inject->read_proc = NULL;
			proc_ecc_err_inject->write_proc = proc_ecc_err_inject_write;
		}
	} while (0);

	/* on any error, cleanup the mess */
	if (rc) {
		brcd_ecc_cleanup();
	}

	return rc;
}

static void __exit brcdecc_exit_440epx(void)
{
	brcd_ecc_cleanup();
}

module_init(brcdecc_init_440epx);
module_exit(brcdecc_exit_440epx);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Brocade");
module_param(debug, uint, 0644);
MODULE_PARM_DESC(debug, "Set to non-zero to enable some debugging stuff");

