/*
 * FILE NAME ibm_ocp_gpio.c
 *
 * BRIEF MODULE DESCRIPTION
 *  API for IBM PowerPC 4xx GPIO device.
 *  Driver for IBM PowerPC 4xx GPIO device.
 *
 *  Armin Kuster akuster@pacbell.net
 *  Sept, 2001
 *
 *  Orignial driver
 *  Author: MontaVista Software, Inc.  <source@mvista.com>
 *          Frank Rowand <frank_rowand@mvista.com>
 *          Debbie Chu   <debbie_chu@mvista.com>
 *
 * Copyright 2000,2001,2002 MontaVista Software Inc.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE	LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	TODO: devfs
 *
 *	Version: 02/01/12 - Armin
 *			 converted to ocp and using ioremap
 *
 *	1.2 02/21/01 - Armin
 *		minor compiler warning fixes
 *
 *	1.3 02/22/01 - Armin
 *		added apm
 *
 *	1.4 05/07/02 - Armin/David Mueller
 *		coverted to core_ocp[];
 *
 *	1.5 05/25/02 - Armin
 *	 name change from *_driver to *_dev
 *
 *	1.6 06/04/02 - Matt Porter
 *	ioremap paddr. Comment as 4xx generic driver.
 *	Fix header to be userland safe and locate in
 *	an accessible area.  Add ioctl to configure
 *	multiplexed GPIO pins.
 *
 *	1.7 07/25/02 - Armin
 *	added CPM to enable/disable in init/exit
 *
 *	1.8 06/12/06 - Bill Z.
 *	Linux 2.6 port
 */

#define VUFX "06.12.06"

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <linux/pm.h>
#include <linux/ibm_ocp_gpio.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/ocp.h>
#include <linux/pci.h>

struct miscdevice ibm_gpio_miscdev;
static struct gpio_regs *gpio_ptrs[2];

static int ibm_pci_alloc(struct ibm_pci_ioctl_data *st_data);
static int ibm_pci_free(struct ibm_pci_ioctl_data *st_data);
static int ibm_pci_read(struct ibm_pci_ioctl_data *st_data);
static int ibm_pci_write(struct ibm_pci_ioctl_data *st_data);

#ifdef CONFIG_PM
static struct pm_dev *pm_gpio;

static int
ibm_gpio_suspend(u32 state)
{
	mtdcr(DCRN_CPMFR, mfdcr(DCRN_CPMFR) | state);
	return 0;
}

static int
ibm_gpio_resume(u32 state)
{
	mtdcr(DCRN_CPMFR, mfdcr(DCRN_CPMFR) & ~state);
	return 0;
}
#endif

static struct gpio_regs  *device_to_regs(__u32 device)
{
#ifdef CONFIG_440EPX
	if ((device != 0) && (device != 1))
		return NULL;
#else
	if (device != 0)
		return NULL;
#endif
	return gpio_ptrs[device];
}

int
ibm_gpio_config(__u32 device, __u32 mask, __u32 data)
{
	u32 cfg_reg;

	struct gpio_regs *gpiop;

	gpiop = device_to_regs(device);
	if (gpiop == NULL)
		return -ENXIO;
#ifdef CONFIG_PM
	pm_access(pm_gpio);
#endif

#ifdef CONFIG_40x
	/*
	 * PPC405 uses CPC0_CR0 to select multiplexed GPIO pins.
	 */
	cfg_reg = mfdcr(DCRN_CHCR0);
	cfg_reg = (cfg_reg & ~mask) | (data & mask);
	mtdcr(DCRN_CHCR0, cfg_reg);
#elif CONFIG_440
#ifdef CONFIG_440GX
	/* 
	 * PPC440GX use DCRN_SDR_PFC0 to select multiplexed GPIO pins.
	 */
	mtdcr(DCRN_SDR_CONFIG_ADDR, DCRN_SDR_PFC0);
	cfg_reg = mfdcr(DCRN_SDR_CONFIG_DATA);
	cfg_reg = (cfg_reg & ~mask) | (data & mask);
	mtdcr(DCRN_SDR_CONFIG_DATA, cfg_reg);
#else
	/*
	 * PPC440 uses CPC0_GPIO to select multiplexed GPIO pins.
	 */
	cfg_reg = mfdcr(DCRN_CPC0_GPIO);
	cfg_reg = (cfg_reg & ~mask) | (data & mask);
	mtdcr(DCRN_CPC0_GPIO, cfg_reg);
#endif
#elif defined (CONFIG_440EPX)
	/* Add 440EPX specific GPIO ops : TODO */
#else
#error This driver is only supported on PPC40x and PPC440 CPUs
#endif

	return 0;
}

int
ibm_gpio_tristate(__u32 device, __u32 mask, __u32 data)
{
	u32 test;
	struct gpio_regs *gpiop;

	gpiop = device_to_regs(device);
	if (gpiop == NULL)
		return -ENXIO;
#ifdef CONFIG_PM
	pm_access(pm_gpio);
#endif
	gpiop->tcr = (gpiop->tcr & ~mask) | (data & mask);
	test = (gpiop->tcr & ~mask) | (data & mask);
	return 0;
}

int
ibm_gpio_open_drain(__u32 device, __u32 mask, __u32 data)
{
	struct gpio_regs *gpiop;

	gpiop = device_to_regs(device);
	if (gpiop == NULL)
		return -ENXIO;
#ifdef CONFIG_PM
	pm_access(pm_gpio);
#endif
	gpiop->odr = (gpiop->odr & ~mask) | (data & mask);

	return 0;
}

int
ibm_gpio_in(__u32 device, __u32 mask, volatile __u32 * data)
{
	struct gpio_regs *gpiop;

	gpiop = device_to_regs(device);
	if (gpiop == NULL)
		return -ENXIO;
#ifdef CONFIG_PM
	pm_access(pm_gpio);
#endif
	gpiop->tcr = gpiop->tcr & ~mask;
	eieio();

	/*
	   ** If the previous state was OUT, and gpiop->ir is read once, then the
	   ** data that was being OUTput will be read.  One way to get the right
	   ** data is to read gpiop->ir twice.
	 */

	*data = gpiop->ir;
	*data = gpiop->ir & mask;
	eieio();
	return 0;
}

int
ibm_gpio_out(__u32 device, __u32 mask, __u32 data)
{
	struct gpio_regs *gpiop;

	gpiop = device_to_regs(device);
	if (gpiop == NULL)
		return -ENXIO;
#ifdef CONFIG_PM
	pm_access(pm_gpio);
#endif
	gpiop->or = (gpiop->or & ~mask) | (data & mask);
	eieio();
	gpiop->tcr = gpiop->tcr | mask;
	eieio();
	return 0;
}

static int
ibm_gpio_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int
ibm_gpio_release(struct inode *inode, struct file *file)
{
	return 0;
}

static int
ibm_gpio_ioctl(struct inode *inode, struct file *file,
	       unsigned int cmd, unsigned long arg)
{
	struct ibm_gpio_ioctl_data ioctl_data;
	int status;

	if (copy_from_user(&ioctl_data,
		   (struct ibm_gpio_ioctl_data *) arg,
		   sizeof (ioctl_data))) {
		return -EFAULT;
	}

	switch (cmd) {
	case IBMGPIO_IN:

		status = ibm_gpio_in(ioctl_data.device,
				     ioctl_data.mask, &ioctl_data.data);
		if (status != 0)
			return status;

		if (copy_to_user((struct ibm_gpio_ioctl_data *) arg,
				 &ioctl_data, sizeof (ioctl_data))) {
			return -EFAULT;
		}

		break;

	case IBMGPIO_OUT:

		return ibm_gpio_out(ioctl_data.device,
				    ioctl_data.mask, ioctl_data.data);

		break;

	case IBMGPIO_OPEN_DRAIN:

		return ibm_gpio_open_drain(ioctl_data.device,
					   ioctl_data.mask, ioctl_data.data);

		break;

	case IBMGPIO_TRISTATE:

		return ibm_gpio_tristate(ioctl_data.device,
					 ioctl_data.mask, ioctl_data.data);

		break;

	case IBMGPIO_CFG:

		return ibm_gpio_config(ioctl_data.device,
				ioctl_data.mask, ioctl_data.data);

		break;
	case IBMPCI_MALLOC:
		return ibm_pci_alloc((struct ibm_pci_ioctl_data *)arg);
		break; 

	case IBMPCI_FREE:
		return ibm_pci_free((struct ibm_pci_ioctl_data *)arg);
		break; 

	case IBMPCI_READ:
		return ibm_pci_read((struct ibm_pci_ioctl_data *)arg);
		break; 

	case IBMPCI_WRITE:
		return ibm_pci_write((struct ibm_pci_ioctl_data *)arg);
		break; 

	default:
		return -ENOIOCTLCMD;

	}
	return 0;
}

static int
ibm_pci_alloc(struct ibm_pci_ioctl_data *st_data)
{
	__u32 bus, dev, size;
	u_int devfn;
	struct pci_dev *dp;
	dma_addr_t dma_handle;
	void *cpu_addr;

	bus = st_data->device;
	dev = st_data->mask;
	size = st_data->data;
	devfn = PCI_DEVFN(dev, 0);
	dp = pci_find_slot(bus, devfn);

	if (!(dp)) {
		return (-ENODEV);
	}
	cpu_addr = pci_alloc_consistent(dp, size, &dma_handle);
	st_data->addr = dma_handle;
	return ((int)cpu_addr);
}

static int
ibm_pci_free( struct ibm_pci_ioctl_data *st_data)
{
	__u32 bus, dev, size;
	u_int devfn;
	struct pci_dev *dp;
	dma_addr_t dma_handle;

	bus = st_data->device;
	dev = st_data->mask;
	size = st_data->data;
	devfn = PCI_DEVFN(dev, 0);
	dp = pci_find_slot(bus, devfn);

	if (!(dp)) {
		return (-ENODEV);
	}
	pci_free_consistent(dp, size, (void *)st_data->addr, dma_handle);
	return (0);
}

static int
ibm_pci_read( struct ibm_pci_ioctl_data *st_data)
{
	int *pAddress;

	pAddress = (int *)st_data->addr;
	st_data->data = *pAddress;
	return (0);
}

static int
ibm_pci_write( struct ibm_pci_ioctl_data *st_data)
{
	int *pAddress;

	pAddress = (int *)st_data->addr;
	*pAddress = st_data->data;
	return (0);
}

static struct file_operations ibm_gpio_fops = {
	owner:THIS_MODULE,
	ioctl:ibm_gpio_ioctl,
	open:ibm_gpio_open,
	release:ibm_gpio_release,
};

static int __devinit ibm_gpio_probe(struct ocp_device *ocp)
{
	ibm_gpio_miscdev.minor = 185;   /*GPIO_MINOR; */
	ibm_gpio_miscdev.name = "gpio";
	ibm_gpio_miscdev.fops = &ibm_gpio_fops;
	misc_register(&ibm_gpio_miscdev);

#ifdef CONFIG_440EPX
/* FIXME:
 * Since there are many platform map GPIO1 to device 0,
 * We have to map GPIO0 to device 1
 */
	if (ocp->def->index == 1) {
  	    gpio_ptrs[0] = (struct gpio_regs *) ioremap(ocp->def->paddr,
			sizeof (struct gpio_regs));
	} else {
  	    gpio_ptrs[1] = (struct gpio_regs *) ioremap(ocp->def->paddr,
			sizeof (struct gpio_regs));
	}
#else
	gpio_ptrs[0] = (struct gpio_regs *) ioremap(ocp->def->paddr,
			sizeof (struct gpio_regs));
#endif
	return 0;
}

static void __devexit ibm_gpio_remove(struct ocp_device *ocp)
{
	misc_deregister(&ibm_gpio_miscdev);
}

static struct ocp_device_id ibm_gpio_ids[] __devinitdata =
{
	{ .vendor = OCP_VENDOR_IBM, .function = OCP_FUNC_GPIO },
	{ .vendor = OCP_VENDOR_INVALID }
};

MODULE_DEVICE_TABLE(ocp, ibm_gpio_ids);

static struct ocp_driver ibm_gpio_driver =
{
	.name		= "gpio",
	.id_table	= ibm_gpio_ids,
	.probe		= ibm_gpio_probe,
	.remove		= __devexit_p(ibm_gpio_remove),
#if defined(CONFIG_PM)
	.suspend	= ibm_gpio_suspend,
	.resume		= ibm_gpio_resume,
#endif
};

static int __init
ibm_gpio_init(void)
{
	printk(KERN_NOTICE "IBM gpio driver version %s\n", VUFX);
	return ocp_register_driver(&ibm_gpio_driver);
}

static void __exit
ibm_gpio_exit(void)
{
	ocp_unregister_driver(&ibm_gpio_driver);
}

module_init(ibm_gpio_init);
module_exit(ibm_gpio_exit);

EXPORT_SYMBOL(ibm_gpio_config);
EXPORT_SYMBOL(ibm_gpio_tristate);
EXPORT_SYMBOL(ibm_gpio_open_drain);
EXPORT_SYMBOL(ibm_gpio_in);
EXPORT_SYMBOL(ibm_gpio_out);

MODULE_LICENSE("GPL");
