/*
 * drivers/char/watchdog/booke_wdt.c
 *
 * Watchdog timer for PowerPC Book-E systems
 *
 * Author: Matthew McClintock
 * Maintainer: Kumar Gala <kumar.gala@freescale.com>
 *
 * Copyright 2005 Freescale Semiconductor Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/notifier.h>
#include <linux/watchdog.h>

#include <asm/reg_booke.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>

/* If the kernel parameter wdt_enable=1, the watchdog will be enabled at boot.
 * Also, the wdt_period sets the watchdog timer period timeout.
 * For E500 cpus the wdt_period sets which bit changing from 0->1 will
 * trigger a watchog timeout. This watchdog timeout will occur 3 times, the
 * first time nothing will happen, the second time a watchdog exception will
 * occur, and the final time the board will reset.
 */

#ifdef	CONFIG_FSL_BOOKE
#define WDT_PERIOD_DEFAULT 61	/* max possible */
#else
#define WDT_PERIOD_DEFAULT 4	/* Refer to the PPC40x and PPC4xx manuals */
#endif				/* for timing information */

u32 booke_wdt_enabled = 0;
u32 booke_wdt_period = WDT_PERIOD_DEFAULT;

#ifdef	CONFIG_FSL_BOOKE
#define WDTP(x)		((((63-x)&0x3)<<30)|(((63-x)&0x3c)<<15))
#else
#define WDTP(x)		(TCR_WP(x))
#endif

static __inline__ void booke_wdt_ping(void);

#if defined(CONFIG_SILKWORM) || defined(CONFIG_SILKWORM_COMMON)

/*
 * BROCADE implementation notes:
 *
 * We always set the h/w watchdog at a known 83.5 millisec value.
 * Which allows us top catch interrupts locked overflow in 168 millisec.
 * To ensure that user space got a chance to run once in 30 sec we implement
 * A softwatchdog which is based on jiffies or HZ. A global count is decremented
 * every tick and the h/w is tickled every HZ. If the user space has not sent
 * a single keep-alive in last sw_wdt_period then siftware watchdog triggers.
 */

#ifdef CONFIG_WATCHDOG_INTERRUPT
#ifdef  DEBUG_WDT
#define REFRESHES_PER_SEC	(1)	/* refreshed every 2^29 clocks */
#else 
#define REFRESHES_PER_SEC	(3)	/* refreshed every 2^25 clocks */
#endif /* DEBUG_WDT */
#else
#define REFRESHES_PER_SEC	(HZ)	/* refreshed each timer tick */
#endif	/* CONFIG_WATCHDOG_INTERRUPT */

#define HW_WDT_MILLISEC_PERIOD_DEFAULT	80	/* 80 millisec, 160millisec panic */

int sw_wdt_enable;
unsigned long sw_wdt_period;
atomic_t sw_wdt_heartbeat_count;
int sw_wdt_default = 1;

/* The user space wtd daemon passes value in micorsec */
#define SW_WDT_DEFAULT_PERIOD	120	/* system default 2 minutes */
#define	MAXONEHOUR		3600UL	/* Max timeout period 60 minutes */
#define	TENMSBASE		10000UL	/* 10 ms */
#define MICROSECBASE		1000000

static inline void
sw_update_timer(void)
{
	atomic_set(&sw_wdt_heartbeat_count, sw_wdt_period);
}

/* 
 * ppc_booke_sw_wdt_heartbeat: detects intr locked.
 *
 * This is invoked from timer IRQ. The system is allowed to be locked 
 * for few millisec (?) otherwise the watchdog will kickin.
 * Otherwise if system is too busy, yet not locked the soft-watchdog 
 * timer is trigrred.
 */
void ppc_booke_sw_wdt_heartbeat(void)
{
	if (sw_wdt_default) {
		/* default used until wdt inits */
		atomic_set(&sw_wdt_heartbeat_count, sw_wdt_period);
		sw_wdt_default = 0;
	}	

	if ((atomic_read(&sw_wdt_heartbeat_count) > 0) || ( !sw_wdt_enable )) {
		if (atomic_read(&sw_wdt_heartbeat_count) > 0)
			atomic_dec(&sw_wdt_heartbeat_count);
			booke_wdt_ping();
        } else {
		printk("Watchdog Timer Timed out, system reset!");
		atomic_set(&sw_wdt_heartbeat_count, sw_wdt_period);
		dump_stack();
		BUG();
	}
	ppc_md.heartbeat_count = 0;
}

/*
 * hw_booke_calibrate_timeout : calibrate timout
 * 
 * This routine makes the watchdog timeout independent of machine/system
 * clock cycle by using a calibrated value. The kernel time-keeping has already
 * done heavy-lifting and we base our timeout on that. 
 * The kernel needs to know by how much TimerBase changes in a TICK to install
 * the timer for HZ. (tb_ticks_per_jiffy)
 */
static __inline__ unsigned int  hw_booke_calibrate_timeout(unsigned int millisec_timeout)
{
	extern unsigned tb_ticks_per_jiffy;
	unsigned int millisec_per_jiffy, val;
	unsigned int jiffy_per_timeout;
	unsigned tb_ticks_per_timeout;

	/* Normalize to the system HZ */

	/* How many millisec in one Jiffy */
	millisec_per_jiffy = 1000/HZ;

	/* How many jiffies in our timeout */
	jiffy_per_timeout = millisec_timeout/millisec_per_jiffy;

	/* So how much will be the TimerBase value */
	tb_ticks_per_timeout = jiffy_per_timeout * tb_ticks_per_jiffy;

	/* Now find the MSB that can be used to progam the watchdog */
	val =  (unsigned int) fls(tb_ticks_per_timeout);

	printk (KERN_INFO "PowerPC Book-E Watchdog - Calibrated %d millisec to Bit value %d\n",
				millisec_timeout, val);
	return (val);

}

void kick_watchdog(void)
{
	booke_wdt_ping();
}

void shut_watchdog(void)
{
	sw_wdt_enable = 0;
	printk (KERN_INFO "PowerPC Book-E Watchdog Shutdown soft timer\n");
}

EXPORT_SYMBOL(shut_watchdog);
EXPORT_SYMBOL(kick_watchdog);

#endif /* SILKWORM - Brocade specific */

/*
 * booke_wdt_enable:
 */
static __inline__ void booke_wdt_enable(void)
{
	u32 val;

	val = mfspr(SPRN_TCR);
	val |= (TCR_WIE|TCR_WRC(WRC_CHIP)|WDTP(booke_wdt_period));

	mtspr(SPRN_TCR, val);
}

/*
 * booke_wdt_ping:
 */
static __inline__ void booke_wdt_ping(void)
{
	mtspr(SPRN_TSR, TSR_ENW|TSR_WIS);
}

/*
 * booke_wdt_write:
 */
static ssize_t booke_wdt_write (struct file *file, const char *buf,
				size_t count, loff_t *ppos)
{
	booke_wdt_ping();
	return count;
}


static struct watchdog_info ident = {
  .options = WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING,
  .firmware_version = 0,
  .identity = "PowerPC Book-E Watchdog",
};

/*
 * booke_wdt_ioctl:
 */
static int booke_wdt_ioctl (struct inode *inode, struct file *file,
			    unsigned int cmd, unsigned long arg)
{
	u32 tmp = 0;
	unsigned long period;
	int state;

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		if (copy_to_user ((struct watchdog_info *) arg, &ident,
				sizeof(struct watchdog_info)))
			return -EFAULT;
	case WDIOC_GETSTATUS:
		return put_user(ident.options, (u32 *) arg);
	case WDIOC_GETBOOTSTATUS:
		/* XXX: something is clearing TSR */
		tmp = mfspr(SPRN_TSR) & TSR_WRS(3);
		/* returns 1 if last reset was caused by the WDT */
		return (tmp ? 1 : 0);
	case WDIOC_KEEPALIVE:
#if defined(CONFIG_SILKWORM) || defined(CONFIG_SILKWORM_COMMON)
		sw_update_timer();
#else
		booke_wdt_ping();
#endif
		return 0;
	case WDIOC_SETTIMEOUT:
		if (get_user(booke_wdt_period, (u32 *) arg))
			return -EFAULT;
		mtspr(SPRN_TCR, (mfspr(SPRN_TCR)&~WDTP(0))|WDTP(booke_wdt_period));
		return 0;
	case WDIOC_GETTIMEOUT:
		return put_user(booke_wdt_period, (u32 *) arg);
	case WDIOC_SETOPTIONS:
#if defined(CONFIG_SILKWORM) || defined(CONFIG_SILKWORM_COMMON)
		if(copy_from_user(&state, (int*) arg, sizeof(int)))
			return -EFAULT;
		if (state & WDIOS_DISABLECARD) {
			sw_wdt_enable = 0;
			printk(KERN_NOTICE "Soft watchdog timer is disabled\n");
			break;
		}
		if (state & WDIOS_ENABLECARD) {
			sw_update_timer();
			sw_wdt_enable = 1;
			booke_wdt_ping();
			printk(KERN_NOTICE "Soft watchdog timer is enabled\n");
			break;
		}
		return 0;
#else
		if (get_user(tmp, (u32 *) arg))
			return -EINVAL;
		if (tmp == WDIOS_ENABLECARD) {
			booke_wdt_ping();
			break;
		} else
			return -EINVAL;
#endif
#if defined(CONFIG_SILKWORM) || defined(CONFIG_SILKWORM_COMMON)
	case WDIOC_GETPERIOD:
		/* return watchdog period (units = microseconds) */
		period = sw_wdt_period / REFRESHES_PER_SEC * MICROSECBASE;
		if (copy_to_user((unsigned long *)arg, &period, 
		    sizeof(period))) {
			return -EFAULT;
		}
		break;
	case WDIOC_SETPERIOD:
		/*
		** set watchdog period (units = microseconds)
		** value of zero means maximum
		**
		** Don't set a watchdog period to a value less than
		** the requested value (period will be rounded up to
		** next available interval the watchdog supports).
		**
		** The software watchdog will expire at some point in
		** the range of (rounded up period) ..
		** (rounded up period + 1 jiffie).  If interrupts are
		** disabled so that the software watchdog is unable to
		** reset the system, then the hardware watchdog will
		** eventually reset the system.
		*/
		if (copy_from_user(&period, (unsigned long *)arg,
		    sizeof(period))) {
			return -EFAULT;
		}
		if (period == 0)
			period = MAXONEHOUR * MICROSECBASE;        
		sw_wdt_period = period / MICROSECBASE * REFRESHES_PER_SEC;
		sw_update_timer();
		printk(KERN_NOTICE"PowerPC Book-E Watchdog: SW New period set to %d millisec\n", sw_wdt_period);
#endif
		break;
	default:
		return -ENOIOCTLCMD;
	}

	return 0;
}
/*
 * booke_wdt_open:
 */
static int booke_wdt_open (struct inode *inode, struct file *file)
{
#if defined(CONFIG_SILKWORM) || defined(CONFIG_SILKWORM_COMMON)
	if (sw_wdt_period == 0)
		sw_wdt_period = SW_WDT_DEFAULT_PERIOD * REFRESHES_PER_SEC;
	sw_update_timer();
	sw_wdt_enable = 1;
	booke_wdt_period = hw_booke_calibrate_timeout(HW_WDT_MILLISEC_PERIOD_DEFAULT);
#endif

	if (booke_wdt_enabled == 0) {
		booke_wdt_enabled = 1;
		booke_wdt_enable();
		printk (KERN_INFO "PowerPC Book-E Watchdog Timer Enabled\n",
				booke_wdt_period);
	}

	return 0;
}

static struct file_operations booke_wdt_fops = {
  .owner = THIS_MODULE,
  .llseek = no_llseek,
  .write = booke_wdt_write,
  .ioctl = booke_wdt_ioctl,
  .open = booke_wdt_open,
};

static struct miscdevice booke_wdt_miscdev = {
  .minor = WATCHDOG_MINOR,
  .name = "watchdog",
  .fops = &booke_wdt_fops,
};

static void __exit booke_wdt_exit(void)
{
	misc_deregister(&booke_wdt_miscdev);
}

/*
 * booke_wdt_init:
 */
static int __init booke_wdt_init(void)
{
	int ret = 0;

	printk (KERN_INFO "PowerPC Book-E Watchdog Timer Loaded\n");
	ident.firmware_version = cpu_specs[0].pvr_value;

	ret = misc_register(&booke_wdt_miscdev);
	if (ret) {
		printk (KERN_CRIT "Cannot register miscdev on minor=%d (err=%d)\n",
				WATCHDOG_MINOR, ret);
		return ret;
	}

	if (booke_wdt_enabled == 1) {

#if defined(CONFIG_SILKWORM) || defined(CONFIG_SILKWORM_COMMON)
		if (sw_wdt_period == 0)
			sw_wdt_period = SW_WDT_DEFAULT_PERIOD * REFRESHES_PER_SEC;
		sw_update_timer();
		booke_wdt_period = hw_booke_calibrate_timeout(HW_WDT_MILLISEC_PERIOD_DEFAULT);
#endif

		printk("already enabled\n");
		printk (KERN_INFO "PowerPC Book-E Watchdog Timer Enabled (wdt_period=%d)\n",
				booke_wdt_period);
		booke_wdt_enable();
	}

	return ret;
}
device_initcall(booke_wdt_init);

