/*
 *
 *    Copyright (c) 2008 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *      Intel/DEC 21554 non-transparent PCI-to-PCI bridge (aka Drawbridge)
 *      device driver for 8548-based Directors
 *
 */
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/version.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <asm/processor.h>


#include <brocade/cpld.h>
#include <brocade/swbd.h>

#include <platforms/4xx/silkworm_sin.h>	/* for sin_get_swbd() */


/*
 * Please include the right header for the right platform.
 */


#include <asm/mpc85xx.h>

#include "../include/db.h"


/* Preprocessor Defines */


/* Bit masks for the 8548 GPIO pins that control primary side DB reset */
#define GPIO0_BIT_28		0x00000008
#define GPIO0_BIT_29		0x00000004
#define GPIO0_BIT_30		0x00000002

#define FS_GPIOCR			0xE0030
#define FS_GPOUTDR			0xE0040

/* Function Macros */

#ifdef DEBUG
#define DBGPRINT(x...) printk(KERN_ERR x)
#else
#define DBGPRINT(x...)
#endif

static int db_neptune_primary_reset(void);

/*
 * Platform configuration instances.  The following contain the platform
 * specific parameters that the DB module uses in controlling and configuring
 * the drawbridges. 
 */

extern db_known_dev_t common_known_devs[];

/* Pointer to current platform configuration instance.  */
extern db_platform_cfg_t *cur_cfg_ptr;

static const uint32_t neptune_base_addrs[] = {
	0x80000000, 				/* drawbridge 1 */
	0xA8000000,  				/* drawbridge 2 */
	0xD0000000   				/* drawbridge 3 */
};

static db_platform_cfg_t neptune_cfg = {		/* Neptune */
    swbd: SWBD_NEPTUNE,
    num_db: 3,						/* num drawbridges */
    busses_per_slot: 10,			/* busses per chassis slot */
    root_bus_incr: 0x40,			/* root bus increment */
    irq_num: 26,					/* internal IRQ num */
    cp_subordinate_bus: 1,			/* Default doesn't matter */
    physical_pci_base: NEPTUNE_PCI_PHYS_BASE, /* Not actually used */
    physical_db_base: 0,
    device_alloc_sz: 0,				/* N/A */
    blade_alloc_sz: DB_NEPTUNE_BLADE_ALLOTMENT, /* Allow 128MB per blade slot */

	/* REVISIT: Not sure what this will be for Neptune. Used by bridge fix-up code */

    fwd_win_offset: 0x08000000,			/* forwarding window offset */
	gpio_base: NULL,				/* to be initialized */
    base_addr: neptune_base_addrs, 		/* drawbridge base addresses */
    db_fixup_method: NULL,			/* No fix-up routine needed */
    config_bases_method: db_configure_bases,	/* Always configure BARs */
    read_bases_method: db_read_bases,		/* Read cur config of BARs */

    setup_dma_method: NULL,			/* No ASIC DMA setup */
    reset_method: db_neptune_primary_reset,	/* ptr to reset func */
    alloc_method: db_dynamic_address_allocate, 	/* ptr to addr. alloc method */
	ntb_lt_primary: 0x10,			/* allow up to n transactions */
	ntb_lt_secondary: 0x10,			/* allow up to n transactions */
	tb_lt_primary: 0x10,			/* allow up to n transactions */
	tb_lt_secondary: 0x10,			/* allow up to n transactions */
	enable_secondary_mem: 0,		/* disable DB secondary side mem */
	known_devs: common_known_devs,	/* list of devices known apriori.*/
									/* Only needed for Scimitar blades */

};
static db_platform_cfg_t pluto_cfg = {		/* Pluto */
    swbd: SWBD_PLUTO,
    num_db: 3,						/* num drawbridges */
    busses_per_slot: 10,			/* busses per chassis slot */
    root_bus_incr: 0x40,			/* root bus increment */
    irq_num: 26,					/* internal IRQ num */
    cp_subordinate_bus: 1,			/* Default doesn't matter */
    physical_pci_base: NEPTUNE_PCI_PHYS_BASE, /* Not actually used */
    physical_db_base: 0,
    device_alloc_sz: 0,				/* N/A */
    blade_alloc_sz: DB_NEPTUNE_BLADE_ALLOTMENT, /* Allow 128MB per blade slot */

    fwd_win_offset: 0x08000000,			/* forwarding window offset */
	gpio_base: NULL,				/* to be initialized */
    base_addr: neptune_base_addrs, 		/* drawbridge base addresses */
    db_fixup_method: NULL,			/* No fix-up routine needed */
    config_bases_method: db_configure_bases,	/* Always configure BARs */
    read_bases_method: db_read_bases,		/* Read cur config of BARs */

    setup_dma_method: NULL,			/* No ASIC DMA setup */
    reset_method: db_neptune_primary_reset,	/* ptr to reset func */
    alloc_method: db_dynamic_address_allocate, 	/* ptr to addr. alloc method */
	ntb_lt_primary: 0x10,			/* allow up to n transactions */
	ntb_lt_secondary: 0x10,			/* allow up to n transactions */
	tb_lt_primary: 0x10,			/* allow up to n transactions */
	tb_lt_secondary: 0x10,			/* allow up to n transactions */
	enable_secondary_mem: 0,		/* disable DB secondary side mem */
	known_devs: common_known_devs,	/* list of devices known apriori.*/
									/* Only needed for Scimitar blades */

};


/* Internal configuration table. */
db_platform_cfg_t *(platform_cfg_tbl[]) = {

	&neptune_cfg,
	&pluto_cfg

};

int db_cfg_tbl_sz = (sizeof (platform_cfg_tbl) / sizeof (void *));

/*
 *
 * int db_primary_reset_init()
 *
 * Description:
 *
 *  Encapsulation of the processor-specific initialization in
 *  support of the DB primary bus interface reset.  The reset
 *  control function should only *control* the interface, not
 *  initialize it: including mapping/unmapping resources.
 *
 *  A subsequent designer/developer may wish to generalize this
 *  to support all platform/processor-specific initialization
 *
 * Input(s):
 *  None.
 *
 * Output(s):
 *  global gpio_base member of DB control structure is initialized
 *  with virtual memory base address
 *
 * Returns:
 *  0 - Success
 *  -ENODEV - no device control structure found
 *  -ENOMEM - i/o subsystem could not map resources
 *
 */
int
db_primary_reset_init(void)
{
  /* Map the GPIO control register space. */
	if (cur_cfg_ptr == NULL) {
		return (-ENODEV);
	}

	cur_cfg_ptr->gpio_base =
		ioremap_nocache(CCSRBAR, MPC85xx_CCSRBAR_SIZE);

	return ((cur_cfg_ptr->gpio_base != NULL) ? 0 : -ENOMEM);
}

/*
 *
 * int db_neptune_primary_reset()
 *
 * Description:
 *
 *  Bits 28, 29, and 30 of the 8548 GPIO register are used to toggle
 *  the 21555 primary reset line on Neptune.
 *
 *  The GPIO bits are connected as:
 *    Bit 30 resets 21555 device 1
 *    Bit 29 resets 21555 device 2
 *    Bit 28 resets 21555 device 3
 *
 * NOTE: The configuration registers of all bridges should
 *       be saved prior to calling this routine.
 *
 *       This routine is CPU-specific because it uses CPU
 *       GPIO registers.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  N/A
 *
 * Returns:
 *  0 - Success
 *  Non-0 - Failed.
 *
 */
int
db_neptune_primary_reset(void)
{
	u_long gpio_or, gpio_mask = (GPIO0_BIT_28 | GPIO0_BIT_29 | GPIO0_BIT_30);
	u_long flags;  /* DO NOT DELETE. USED BY SYS_INT_SAVE() */
	int error = 0;
	void *gpio_base;

	DBGPRINT("PCI DrawBridge Primary Reset GPIO Word 0x%lx\n", gpio_mask);

	if ((cur_cfg_ptr == NULL) || (cur_cfg_ptr->gpio_base == NULL)) {
		return (-ENODEV);
	}

	gpio_base = (void *)cur_cfg_ptr->gpio_base;
	gpio_or = in_be32((volatile __u32 *)(gpio_base + FS_GPOUTDR));

	/* Force a bridge reset via the processor gpio0 register. Clears */
	/* all config registers, and re-reads strapping pins. */
	/* RMW GPIO output register to force the appropriate bits */
	/* low, then force the bits back high */

	SYS_INT_SAVE();

	out_be32((volatile __u32 *)(gpio_base + FS_GPOUTDR),
			 (gpio_or & ~gpio_mask));

	/*
	 * Don't allow anything else to run. The Drawbridges
	 * are unusable at this point and we could hang if
	 * anyone tries a PCI access.
	 * But don't let the delay to exceed 1ms.
	 */
	udelay(1000);

	out_be32((volatile __u32 *)(gpio_base + FS_GPOUTDR),
			 (gpio_or | gpio_mask));

	SYS_INT_RESTORE();

  return (error);
}

