/*
 * (C) Copyright 2003-2004
 * Humboldt Solutions Ltd, adrian@humboldt.co.uk.

 * This is a combined i2c adapter and algorithm driver for the
 * MPC107/Tsi107 PowerPC northbridge and processors that include
 * the same I2C unit (8240, 8245, 85xx).
 *
 * Release 0.8
 *
 * This file is licensed under the terms of the GNU General Public
 * License version 2. This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/time.h>
#include <linux/fsl_devices.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/sysctl.h>
#include <asm/uaccess.h>
#include <brocade/cpld.h>
#include "silkworm_i2c_stats.h"

#define MPC_I2C_ADDR  0x00
#define MPC_I2C_FDR 	0x04
#define MPC_I2C_CR	0x08
#define MPC_I2C_SR	0x0c
#define MPC_I2C_DR	0x10
#define MPC_I2C_DFSRR 0x14
#define MPC_I2C_REGION 0x20

#define CCR_MEN  0x80
#define CCR_MIEN 0x40
#define CCR_MSTA 0x20
#define CCR_MTX  0x10
#define CCR_TXAK 0x08
#define CCR_RSTA 0x04

#define CSR_MCF  0x80
#define CSR_MAAS 0x40
#define CSR_MBB  0x20
#define CSR_MAL  0x10
#define CSR_SRW  0x04
#define CSR_MIF  0x02
#define CSR_RXAK 0x01

#define IIC_TIMEOUT		(HZ/4)		/* 250 msec */
#define IIC_RETRY		2

static int core_i2c_debug = 0;

// #if 0
int i2cTurboCnfg = 0;
int i2cTurboFlag = 0;
static unsigned one_time_on = 0;
static unsigned one_time_off = 0;
static ctl_table i2c_tab[]={
	{ 18, "i2cTurboFlag", &i2cTurboFlag,
		sizeof(int), 0644, NULL, &proc_dointvec},
	{ 19, "i2cTurboCnfg", &i2cTurboCnfg,
		sizeof(int), 0644, NULL, &proc_dointvec},
	{ 20, "i2c_debug", &core_i2c_debug, sizeof(int), 0644, NULL,
&proc_dointvec},
	{0}
};

static ctl_table i2c_root_tab[]={
	{CTL_DEBUG, "debug", NULL, 0, 0555, i2c_tab},
	{0}
};
// #endif

/* Types to handle at interrupt context */
enum { OTHER, BASIC_READ, BASIC_WRITE };

struct mpc_i2c {
	void __iomem *base;
	u32 interrupt;
	wait_queue_head_t queue;
	struct i2c_adapter adap;
	int irq;
	u32 flags;
	u8 *data;
	u8 cmd;		/* OTHER, BASIC_READ, BASIC_WRITE */
	u16 count;	/* bytes to read */
};

static __inline__ void writeccr(struct mpc_i2c *i2c, u32 x)
{
	writeb(x, i2c->base + MPC_I2C_CR);
}

/* process a byte read from the interrupt service routine */
static void byteread(struct mpc_i2c *i2c)
{
	int l=i2c->count;
	if (l <= 1) {	/* should only happen with l == 1, but lets be defensive */
		i2c->cmd = OTHER;
		/* switch off master bit, causes stop to be sent */
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_TXAK);
	}
	if (l == 2) {
		/* switch on txak bit, do not send ack bit after getting next byte */
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_TXAK);
	}
	*i2c->data++ = readb(i2c->base + MPC_I2C_DR);
	i2c->count = --l;
	if (l <= 0) {		/* should only happen with l == 0 */
		wake_up_interruptible(&i2c->queue);
		return;
	}
}

#if 0
static void bytewrite(struct mpc_i2c *i2c)
{
	u32 l=i2c->count--;
	u32 x=i2c->interrupt;
	if ((x & CSR_MCF) || !(x & CSR_MAL) || (l == 0)) {
		wake_up_interruptible(&i2c->queue);
		return;
	}
	writeb(*i2c->data++, i2c->base + MPC_I2C_DR);
}
#endif

static irqreturn_t mpc_i2c_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct mpc_i2c *i2c = dev_id;
	u32 x;
	if (readb(i2c->base + MPC_I2C_SR) & CSR_MIF) {
		/*
		 * Read again to allow register to stabilise
		 * Tracy has a theory that there is a race condition here
		 * during access of the i2c->interrupt value
		 */

		i2c->interrupt = x = readb(i2c->base + MPC_I2C_SR);
		writeb(0, i2c->base + MPC_I2C_SR);
		switch (i2c->cmd) {
		default: BUG();
		case OTHER: wake_up_interruptible(&i2c->queue); break;
		case BASIC_READ:
			if ((x & (CSR_MCF|CSR_MAL)) != CSR_MCF)
				wake_up_interruptible(&i2c->queue);
			else 
				byteread(i2c);
			break;
		/* case BASIC_WRITE: bytewrite(i2c); break; */
		}

	}
	return IRQ_HANDLED;
}

/*
 * wait for I2C event
 * After this routine returns i2c->interrupt is 0
 */

static int i2c_wait(struct mpc_i2c *i2c, unsigned timeout, int writing)
{
	unsigned long orig_jiffies = jiffies;
	u32 x;
	int result = 0;
	struct timeval	perf_begin;

	do_gettimeofday(&perf_begin);

	if (i2c->irq == 0)
	{
		while (!(readb(i2c->base + MPC_I2C_SR) & CSR_MIF)) {
		/* We don't expect this code to be used as we expect to have an irq */
			schedule();
			if (time_after(jiffies, orig_jiffies + timeout)) {
				pr_debug("I2C: timeout\n");
				result = -EIO;
				break;
			}
		}
		x = readb(i2c->base + MPC_I2C_SR);
		writeb(0, i2c->base + MPC_I2C_SR);
	} else {
		/*
		 * Interrupt mode. Timeout must be well under 1 sec.
		 * Is there is a race here with the i2c->interrupt variable
		 * wait_event_interruptible_timeout is a macro, which includes a loop
		 * that tests the condition
		 */

		result = wait_event_interruptible_timeout(i2c->queue,
			(i2c->interrupt & CSR_MIF), timeout);

		if (unlikely(result < 0))
			pr_debug("I2C: wait interrupted\n");
		else if (unlikely(!(i2c->interrupt & CSR_MIF))) {
			pr_debug("I2C: wait timeout\n");
			result = -ETIMEDOUT;
		}

		x = i2c->interrupt;
		i2c->interrupt = 0;
	}

	if (result < 0) {
		goto end;
	}

	if (!(x & CSR_MCF)) {
		pr_debug("I2C: unfinished\n");
		result = -EIO;
		goto end;
	}

	if (x & CSR_MAL) {
		pr_debug("I2C: MAL\n");
		result = -EIO;
		goto end;
	}

	if (writing && (x & CSR_RXAK)) {
		pr_debug("I2C: No RXAK\n");
		/* generate stop */
		writeccr(i2c, CCR_MEN);
		result = -EIO;
		goto end;
	}

 end:

	perf_stats_xfer_end(&perf_stats->wait_times, &perf_begin);

	return (result);
}

static void mpc_i2c_setclock(struct mpc_i2c *i2c)
{
	/* Set clock and filters */
	if (i2c->flags & FSL_I2C_DEV_SEPARATE_DFSRR) {
		writeb(0x2f, i2c->base + MPC_I2C_FDR);
		writeb(0x10, i2c->base + MPC_I2C_DFSRR);
	} else if (i2c->flags & FSL_I2C_DEV_CLOCK_5200)
		writeb(0x3f, i2c->base + MPC_I2C_FDR);
	else
		writel(0x1031, i2c->base + MPC_I2C_FDR);
}
/*
 * This takes the I2C controller out of soft reset.
 * It does NOT issue START command on the I2C bus.
 */
static void mpc_i2c_start(struct mpc_i2c *i2c)
{
	/* Reset the device in case it thinks the bus is busy */
	writeccr(i2c, 0);
	udelay(20);

	/* Clear arbitration */
	writeb(0, i2c->base + MPC_I2C_SR);

	/* Start with MEN */
	writeccr(i2c, CCR_MEN);
	i2c->cmd = OTHER;	/* reset the interrupt processing */
	i2c->interrupt = 0;
}
/*
 * This code is manually asserting STOP command on the bus in part because
 * the controller provides no way to do it automatically after writes.
 * The code MUST ensure correct timing or else the STOP may not be
 * recognized by other devices. This is required by MSA spec for SFP
 * devices.
 */
static void mpc_i2c_stop(struct mpc_i2c *i2c)
{
	udelay(20); /* Ensure STOP is not asserted too soon */
	writeccr(i2c, CCR_MEN);
	udelay(20); /* Ensure STOP remains asserted long enough */
}
/*
 *
 */
static void mpc_access_debug(struct mpc_i2c *i2c, int target,
		    const u8 * data, int length, int restart, int read)
{
	int i;
	unsigned char cpld_data;

	printk("-----------------------------\n");
	printk("\n%s(): %s target 0x%x i2c->base 0x%x"
		   "length %d restart 0x%x",
		   __FUNCTION__, ((read)?"READ":"WRITE"),
		   target, (int)i2c->base, length, restart);

	for (i = 0; i < length; i++) {

		if (!(i % 0x10)) {
			printk("\n0x%4.4x: ", i);
		}
		printk("%02x ", data[i]);

	}
	printk("\n");
	i = cpld_read_register(MET_I2C_SELECT, &cpld_data);
	printk("\n%s():cpld_data: 0x%x  ret: %d\n",
		   __FUNCTION__, cpld_data, i);
	printk("-----------------------------\n");
}

static int mpc_write(struct mpc_i2c *i2c, int target,
		     const u8 * data, int length, int restart)
{
	int i, result = 0;
	unsigned timeout = i2c->adap.timeout;
	u32 flags = restart ? CCR_RSTA : 0;
	struct timeval	perf_begin;

	perf_stats_xfer_begin(&perf_stats->write_xfers,
						  &perf_stats->write_bytes,
						  &perf_begin, length);

	/* Start with MEN */
	if (!restart) {
		writeccr(i2c, CCR_MEN);
	}

	/* Start as master */
	writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_MTX | flags);
	/* Write target byte */
	writeb((target << 1), i2c->base + MPC_I2C_DR);

	if (i2c_wait(i2c, timeout, 1) < 0) {
		result = -1;
		goto end;
	}

	if (core_i2c_debug > 0) {

		mpc_access_debug(i2c, target, data, length, restart, 0);

	}
	for (i = 0; i < length; i++) {
		/* Write data byte */
		writeb(data[i], i2c->base + MPC_I2C_DR);

		if (i2c_wait(i2c, timeout, 1) < 0) {
			result = -1;
			goto end;
		}
	}

 end:

	perf_stats_xfer_end(&perf_stats->write_xfers, &perf_begin);

	return (result);
}

static int mpc_read(struct mpc_i2c *i2c, int target,
		    u8 * data, int length, int restart)
{
	unsigned timeout = i2c->adap.timeout;
	int i, result = length;
	u32 flags = restart ? CCR_RSTA : 0;
	struct timeval	perf_begin;

	perf_stats_xfer_begin(&perf_stats->read_xfers,
						  &perf_stats->read_bytes,
						  &perf_begin, length);

	/* Start with MEN */
	if (!restart) {
		writeccr(i2c, CCR_MEN);
	}

	i2c->interrupt = 0;
	/* Switch to read - restart */
	writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_MTX | flags);
	/* Write target address byte - this time with the read flag set */
	writeb((target << 1) | 1, i2c->base + MPC_I2C_DR);

	if (i2c_wait(i2c, timeout, 1) < 0) {
		result = -1;
		goto end;
	}

	/*
	 * This delay is added to comply with SFP MSA spec that requires
	 * a minimum of 20us between end of an address cycle and beginning of
	 * the data cycle. There is no way to tell if an SFP is being
	 * accessed, so we have to always do this.
	 */

	udelay(20);

	if (length == 0) goto maybedebug;

	if (length == 1)
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_TXAK);
	else
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA);

// #if 0
	/*
	 * Add debug/Trace code here to print to console one-time
	 * whenever the i2cTurboCnfg is going from:
	 *  0 to non-zero (1 = I2C_TURBO_CNFG_ENABLE_SFP,
	 *				   2 = I2C_TURBO_CNFG_ENABLE_ALL mode)
	 *  non-zero to 0 (0 = I2C_TURBO_CNFG_DISABLED) 
	 *  This can be ifdef'd out in the GA version.
	 */
	if (i2cTurboCnfg) {
		if (!one_time_on) {
			printk("i2cTurboCnfg on=%d\n", i2cTurboCnfg);
		}
		one_time_on++;
		one_time_off = 0;
	} else {
		if (!one_time_off) {
			printk("i2cTurboCnfg off\n");
		}
		one_time_off++;
		one_time_on = 0;
	}
// #endif

	if ((i2c->irq) && (i2cTurboFlag)) {
		/* Interrupts are enabled, do this the fast way */
		i2c->data = data;
		i2c->count = length;
		i2c->cmd = BASIC_READ;
		i2c->interrupt = 0;
		/* Dummy read */
		readb(i2c->base + MPC_I2C_DR);
		if (i2c_wait(i2c, timeout * length, 0) < 0) {
			result = -1;
			goto end;
		}
	} else {
		/* Dummy read */
		readb(i2c->base + MPC_I2C_DR);


		for (i = 0; i < length; i++) {
			if (i2c_wait(i2c, timeout, 0) < 0) {
				result = -1;
				goto end;
			}

			/* Generate txack on next to last byte */
			if (i == length - 2)
				writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_TXAK);
			/* Generate stop on last byte */
			if (i == length - 1)
				writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_TXAK);
			data[i] = readb(i2c->base + MPC_I2C_DR);
		}
	}
maybedebug:

	if (core_i2c_debug > 0) {

		mpc_access_debug(i2c, target, data, length, restart, 1);

	}

 end:

	perf_stats_xfer_end(&perf_stats->read_xfers, &perf_begin);

	return (result);
}

static int mpc_xfer(struct i2c_adapter *adap, struct i2c_msg *msgs, int num)
{
	struct i2c_msg *pmsg;
	int i;
	int ret = 0;
	unsigned long orig_jiffies = jiffies;
	struct mpc_i2c *i2c = i2c_get_adapdata(adap);

	mpc_i2c_start(i2c);

	/* Allow bus up to 1s to become not busy */
	while (readb(i2c->base + MPC_I2C_SR) & CSR_MBB) {
		if (signal_pending(current)) {
			pr_debug("I2C: Interrupted\n");
			return -EINTR;
		}
		if (time_after(jiffies, (orig_jiffies + IIC_TIMEOUT))) {
			printk("\n%s %d: I2C: timeout\n", __FUNCTION__, __LINE__);

			return -EIO;
		}
		schedule();
	}

	for (i = 0; ret >= 0 && i < num; i++) {
		pmsg = &msgs[i];
		pr_debug("Doing %s %d bytes to 0x%02x - %d of %d messages\n",
			 pmsg->flags & I2C_M_RD ? "read" : "write",
			 pmsg->len, pmsg->addr, i + 1, num);
		if (pmsg->flags & I2C_M_RD)
			ret =
			    mpc_read(i2c, pmsg->addr, pmsg->buf, pmsg->len, i);
		else
			ret =
			    mpc_write(i2c, pmsg->addr, pmsg->buf, pmsg->len, i);
	}
	mpc_i2c_stop(i2c);
	return (ret < 0) ? ret : num;
}

static u32 mpc_functionality(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static struct i2c_algorithm mpc_algo = {
	.master_xfer = mpc_xfer,
	.functionality = mpc_functionality,
	.fixup_i2c		= 0,        // Set in board initialization
};

static struct i2c_adapter mpc_ops = {
	.owner = THIS_MODULE,
	.name = "MPC adapter",
	.id = I2C_HW_MPC107,
	.algo = &mpc_algo,
	.class = I2C_CLASS_HWMON,
	.timeout = IIC_TIMEOUT,
	.retries = IIC_RETRY /* Doesn't seem to be used - yet */
};

static int fsl_i2c_probe(struct device *device)
{
	int result = 0;
	struct mpc_i2c *i2c;
	struct platform_device *pdev = to_platform_device(device);
	struct fsl_i2c_platform_data *pdata;
	struct resource *r = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	pdata = (struct fsl_i2c_platform_data *) pdev->dev.platform_data;

	if (!(i2c = kmalloc(sizeof(*i2c), GFP_KERNEL))) {
		return -ENOMEM;
	}
	memset(i2c, 0, sizeof(*i2c));

	i2c->cmd = OTHER;	/* Not strictly rqd as OTHER=0 & we just set it to 0 */
	i2c->irq = platform_get_irq(pdev, 0);
	i2c->flags = pdata->device_flags;
	init_waitqueue_head(&i2c->queue);

	i2c->base = ioremap((phys_addr_t)r->start, MPC_I2C_REGION);

	if (!i2c->base) {
		printk(KERN_ERR "i2c-mpc - failed to map controller\n");
		result = -ENOMEM;
		goto fail_map;
	}

	if (i2c->irq != 0)
		if ((result = request_irq(i2c->irq, mpc_i2c_isr,
					  SA_SHIRQ, "i2c-mpc", i2c)) < 0) {
			printk(KERN_ERR
			       "i2c-mpc - failed to attach interrupt\n");
			goto fail_irq;
		}

	mpc_i2c_setclock(i2c);
	dev_set_drvdata(device, i2c);

	i2c->adap = mpc_ops;
	i2c_set_adapdata(&i2c->adap, i2c);
	i2c->adap.dev.parent = &pdev->dev;

	if ((result = i2c_add_adapter(&i2c->adap)) < 0) {
		printk(KERN_ERR "i2c-mpc - failed to add adapter\n");
		goto fail_add;
	}

	return result;

      fail_add:
	if (i2c->irq != 0)
		free_irq(i2c->irq, NULL);
      fail_irq:
	iounmap(i2c->base);
      fail_map:
	kfree(i2c);
	return result;
};

static int fsl_i2c_remove(struct device *device)
{
	struct mpc_i2c *i2c = dev_get_drvdata(device);

	i2c_del_adapter(&i2c->adap);
	dev_set_drvdata(device, NULL);

	if (i2c->irq != 0)
		free_irq(i2c->irq, i2c);

	iounmap(i2c->base);
	kfree(i2c);
	return 0;
};

/* Structure for a device driver */
static struct device_driver fsl_i2c_driver = {
	.name = "fsl-i2c",
	.bus = &platform_bus_type,
	.probe = fsl_i2c_probe,
	.remove = fsl_i2c_remove,
};



static int read_bus_i2c_debug(char *buf, char **start, off_t offset, int len, int *eof,
                 void *private)
{
	int nr = 0;

	// MOD_INC_USE_COUNT;
	nr += sprintf(buf+nr, "%d\n", core_i2c_debug);
	// MOD_DEC_USE_COUNT;

	return nr;
}

static int write_bus_i2c_debug(struct file *file, const char *buffer,
			     unsigned long count, void *data)
{
	char ptr[PAGE_SIZE];

	// MOD_INC_USE_COUNT;

	if(count > PAGE_SIZE)
		count = PAGE_SIZE;

	if(copy_from_user(ptr, buffer, count)) {
		// MOD_DEC_USE_COUNT;
		printk(KERN_ERR "%s: copy_from_user failed.\n", __FUNCTION__);
		return -EFAULT;
	}

	if(*ptr) {
		if (*ptr >= '0' && *ptr <= '9') {
			core_i2c_debug = *ptr - '0';
			printk(KERN_INFO "%s :Setting "
					   "core_i2c_debug to %d \n", __FUNCTION__,
				   core_i2c_debug);
		} else {
			printk(KERN_ERR "Invalid debug level.\n");
		}
	}

	// MOD_DEC_USE_COUNT;

	return count;
}
/*
 *
 */
static void iic_debug_create_proc(void)
{

	struct proc_dir_entry *proc_dir_i2c;
	struct proc_dir_entry *proc_debug_i2c;

	proc_dir_i2c = proc_mkdir("driver/iic", NULL);

	if (proc_dir_i2c == NULL) {

		printk(KERN_ERR "\n\n%s: Could not create directory driver/iic/\n",
			   __FUNCTION__);
		return;

	}
	proc_debug_i2c = create_proc_entry("i2c_debug", 0, proc_dir_i2c);

	if (!proc_debug_i2c) {
		printk(KERN_ERR "%s: Could not create /driver/iic/i2c_debug",
			   __FUNCTION__);
		//		i2cproc_cleanup();
		return;
 	}
	proc_debug_i2c->read_proc = &read_bus_i2c_debug;
	proc_debug_i2c->write_proc = &write_bus_i2c_debug;

// #if 0
	// much simpler to use sysctl these days
	register_sysctl_table(i2c_root_tab, 0);
// #endif

	if (!perf_stats) {
		perf_stats_init();
	}

}
static int __init fsl_i2c_init(void)
{

	iic_debug_create_proc();

	return driver_register(&fsl_i2c_driver);
}

static void __exit fsl_i2c_exit(void)
{
	driver_unregister(&fsl_i2c_driver);
}

module_init(fsl_i2c_init);
module_exit(fsl_i2c_exit);
EXPORT_SYMBOL(i2cTurboCnfg);
EXPORT_SYMBOL(i2cTurboFlag);

MODULE_AUTHOR("Adrian Cox <adrian@humboldt.co.uk>");
MODULE_DESCRIPTION
    ("I2C-Bus adapter for MPC107 bridge and MPC824x/85xx/52xx processors");
MODULE_LICENSE("GPL");
