/* 
 * Polled-mode ATA dumper used to write kernel crash dump
 * data to a pre-configured area on a harddrive.
 *
 * The polled-mode dump code bypasses the linux ATA driver 
 * and is intended to be called only from the panic() function 
 * with interrupts disabled.  No functions in the dump path 
 * will schedule.  The char device is used only to configure
 * the dumper, and may be invoked at anytime.
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/proc_fs.h>
#include <asm/processor.h>
#include <asm/io.h>
#include <linux/ata_dumper.h>
#include <linux/vmalloc.h>		/* testing only */
#include <linux/ata_dumper.h>
#include <linux/hdreg.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include <linux/hdreg.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/utsname.h>
#include <linux/rwsem.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/ide.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/seq_file.h>
#include <brocade/platform.h>
#include <brocade/silkworm.h>

extern sw_bd silkworm;

#if 0
/* consider whether better to use readb/writeb in the future as
 * silkworm_ide.c does.
 */

#define inb ide_mm_inb
#define outb ide_mm_outb

static inline u8 ide_mm_inb (unsigned long port)
{
        return (u8) readb((void __iomem *) (port + (ulong) isa_io_base));
}
static inline void ide_mm_outb (u8 value, unsigned long port)
{
        writeb(value, (void __iomem *) (port + (ulong) isa_io_base));
}
#endif

#define SECTOR_SIZE	512		/* LBA sector size */
#define WDT_TIMEOUT	(120*HZ)	/* allow the wdt to expire after 
					 * this much time has passed.
					 */
#ifndef min
#define min(x,y) ( (x)<(y)?(x):(y) )
#endif

#define DPRINTK(args...)	
//#define DPRINTK(args...)	kick_watchdog(); printk(## args)

/* block maps */
/* It would take too much space to allocate a large array of unsigned longs
 * so try to group the blocks into contiguous sections.
 */
struct blkmap {
	unsigned long lba;	/* starting lba number */
	unsigned long num;	/* how many contiguious blocks */
};

/* data that is particular to a single file mapping */
/* This data is initialized at boot time, and must be intact for the
 * dumper to work.  Therefore we protect all this data with a checksum.
 */
typedef struct dumper_priv_struct {
	unsigned int map_id;
	char *write_buf;	/* sector buffer */
	dev_t dumpdev;
	struct blkmap *blkmap;	/* pointer to block maps array */
	int num_blkmaps;	/* number of blkmaps in the array */
	int sec_p_blk;		/* number of LBA sectors / block */
	unsigned long total_lba;/* total number of LBA sectors we can write */
	unsigned long io_ports[IDE_NR_PORTS];	/* task file registers */
	int master;		/* master/slave drive */
	int armed;
	int dumper_is_open;
} dumper_priv_type;

/********************************************************************
	there's a state machine of sorts associated with each instance of
	dumper_priv_type. it is driven through it's states via the ioctls.
	the state machine is simple enough that no explicit state variable
	is used, rather the state is implied from the contents of the
	various fields.

	the normal state transitions and the ioctls that drive them are
	described below. any IOCTLs received in a given state for which
	explicit action isn't described below are considered an error
	condition.

	Unallocated State (this is the initial state)
		all fields contain zeroes.
		valid IOCTLs:
			ATA_DUMPER_IOC_ALLOC => Allocated State
				map_id <- passed in map ID
				write_buf <- kmalloc()'ed sector buffer

	Allocated State
		map_id has allocated map ID.
		write_buf has kmalloc()'ed sector buffer.
		other fields contain zeroes
		valid IOCTLs:
			ATA_DUMPER_IOC_FREE => Unallocated State
				map_id <- 0
				write_buf freed then <- 0
			ATA_DUMPER_IOC_ADD_MAP => Mapped State
				dumpdev <- device ID
				blkmap <- kmalloc()'ed block map array
				num_blkmaps <- # of elements in block map array
				sec_p_blk <- # of sectors per superblock
			ATA_DUMPER_IOC_CLEAR_MAP => no-op
			ATA_DUMPER_IOC_DISARM => no-op
			ATA_DUMPER_IOC_ALLOC => returns error with
				EADDRINUSE in errno (see usage below).

	Mapped State
		map_id has allocated map ID.
		write_buf has kmalloc()'ed sector buffer.
		blkmap != NULL, num_blkmaps != 0, dumpdev and sec_p_blk initialized.
		other fields contain zeroes
		valid IOCTLs:
			ATA_DUMPER_IOC_CLEAR_MAP => Allocated State
				blkmap freed then <- 0
				num_blkmaps <- 0
				dumpdev and sec_p_blk <- 0
			ATA_DUMPER_IOC_ARM => Armed State
				armed <- 1
				total_lba <- total number of LBA sectors we can write
				io_ports <- array of task file registers
				master <- master/slave drive (always 1 for now)
			ATA_DUMPER_IOC_DISARM => no-op

	Armed State
		map_id has allocated map ID.
		write_buf has kmalloc()'ed sector buffer.
		blkmap != NULL, num_blkmaps != 0, dumpdev and sec_p_blk initialized.
		armed == 1, total_lba, io_ports and master initialized.
		valid IOCTLs:
			ATA_DUMPER_IOC_DISARM => Mapped State
				armed <- 0
				total_lba <- 0
				io_ports cleared
				master <- 0
			ATA_DUMPER_IOC_ARM => no-op

	A fifth sort of state is entered during a panic when the kernel
	interface is entered via a call to ata_dump_open(), which gets
	indicated by setting dumper_is_open to 1, but by then the switch
	is going down and all that is going to happen is an open, writes,
	and a close, so its really a different state machine by then.

	So how can this code tell what state a given dumper_priv_type is in?

		if (map_id == 0) => Unallocated State
		else if (armed != 0) => Armed State
		else if (num_blkmaps != 0) => Mapped State
		else => Allocated State

	at least that's how the code in the ioctl handler looks at it.

	So how is an application supposed to use this?

	Expected order of ioctls for a given map ID is:
		ATA_DUMPER_IOC_ALLOC => Allocated State
		ATA_DUMPER_IOC_ADD_MAP => Mapped State
		ATA_DUMPER_IOC_ARM => Armed State

	To change a mapping that the application knows is already Armed:
		ATA_DUMPER_IOC_DISARM => Mapped State
		ATA_DUMPER_IOC_CLEARMAP => Allocated State
		ATA_DUMPER_IOC_ADD_MAP => Mapped State
		ATA_DUMPER_IOC_ARM => Armed State

	How is application to deal with unknown map ID state (like might
		happen if application dies after above sequence and then
		gets restarted)?

		ATA_DUMPER_IOC_ALLOC
			returns error with EADDRINUSE in errno, so the ID
				is already allocated but state is otherwise unknown, so...
		ATA_DUMPER_IOC_DISARM
		ATA_DUMPER_IOC_CLEARMAP

		the DISARM and CLEARMAP will succeed or fail gracefully depending
		on state. at this point no matter what state the map ID started in
		its now back in Allocated State and application can remap and rearm.
		
 ********************************************************************/

#define NUM_MAPPINGS 10

static dumper_priv_type dumper_priv[NUM_MAPPINGS];

/* data that is not particular to a single file mapping */
static struct {
	unsigned ticks_jiffy;	/* how many timebase ticks per jiffy */
	unsigned ticks_usec;	/* how many timebase ticks per usec */
	int wdt_kicked;		/* keep track of the first time we kick the wdt */
} dumper_priv_global;

/* This data is created during the dump, so they don't have to be 
 * protected by checksum
 */
static unsigned long curr_lba[NUM_MAPPINGS];
	/* our current LBA block number */
static unsigned long lba_counter[NUM_MAPPINGS];
	/* count of which logical sector we are writing to */
static struct blkmap *curr_blkmap[NUM_MAPPINGS];
	/* ptr to our current blkmap entry */
static int write_count[NUM_MAPPINGS];
	/* # of bytes in sector buffer (dumper_priv[x].write_buf) */

extern unsigned tb_ticks_per_jiffy;
#if 0
static int ata_dumper_read_sector(char dump[512]);
#endif

/*
	searches dump_priv[] array for element with map_id equal to
	passed in parameter.

	if found, returns index to matching array element
	if not found, returns -1
*/

static int ata_dumper_lookup_map_id(unsigned int map_id)
{
	int map_idx;

	for (map_idx = 0; map_idx < NUM_MAPPINGS; ++map_idx) {
		if (dumper_priv[map_idx].map_id == map_id)
			break;
	}

	if (map_idx >= NUM_MAPPINGS)
		return -1;
	else
		return map_idx;
}

/* return the next LBA sector to write, or ((unsigned long)-1) if error
 */
static unsigned long get_next_lba(int map_idx)
{
	/* bump the logical sector number we are writing to */
	lba_counter[map_idx]++;

	/* check for overrun in entire file mapping */
	if (lba_counter[map_idx] <= dumper_priv[map_idx].total_lba) {
		/* check for overrun of current blkmap structure */
		if (curr_lba[map_idx] >=
				(curr_blkmap[map_idx]->lba + curr_blkmap[map_idx]->num)) {
			/* make next blkmap structure in file mapping the current one */
			curr_blkmap[map_idx]++;
			/* make current lba the first in the new current blkmap */
			curr_lba[map_idx] = curr_blkmap[map_idx]->lba;
		}
		return curr_lba[map_idx]++;
	}
	else
		return (unsigned long) -1;	/* overrun! */
}

/************************************************************************************
 * This code partly taken from Rusty Russell's OOPSER patch.
 */

/* This is the core dumper code: x86 IDE.  It must *not* reference any
   outside code. */


/* Wait for at least N usecs (1 clock per cycle, 1GHz processor = 1000) */
static inline void ata_dumper_usec(unsigned int num_usec)
{
#if 0
	unsigned int i;
	for (i = 0; i < 1000 * num_usec; i++);
#else
	uint32_t tbl, tbu, tbu2;
	uint64_t tbase;
	static uint64_t timeout;

	/* Handle rollover during read */
	do {
		tbu = mfspr(SPRN_TBRU);
		tbl = mfspr(SPRN_TBRL);
		tbu2 = mfspr(SPRN_TBRU);
	} while(tbu != tbu2);
	tbase = (((uint64_t)tbu) << 32) + ((uint64_t)tbl);

	timeout = tbase + ((uint64_t)num_usec * dumper_priv_global.ticks_usec);

	while(tbase < timeout) {
		do {
			tbu = mfspr(SPRN_TBRU);
			tbl = mfspr(SPRN_TBRL);
			tbu2 = mfspr(SPRN_TBRU);
		} while(tbu != tbu2);
		tbase = (((uint64_t)tbu) << 32) + ((uint64_t)tbl);
	}
#endif
}


/* Wait for up to one second. */
static int wait_for(int map_idx, unsigned char flag)
{
	unsigned int i;
	unsigned char b;
	unsigned long *ioports = dumper_priv[map_idx].io_ports;

	/* Pause (at least 400ns) in case we just issued a command */
	ata_dumper_usec(1);
	for (i = 0; i < 1000; i++) {
		b = inb(ioports[IDE_STATUS_OFFSET]);
		if (!(b & BUSY_STAT)) {
			if (b & ERR_STAT) return 0;
			if (b & flag) return 1;
		}
		/* since we could be in this loop for as long as a second,
			kick the pooch */
		kick_watchdog();
		ata_dumper_usec(1000);
	}
	return 0;
}

static int wait_before_command(int map_idx)
{
	return wait_for(map_idx, READY_STAT);
}

static int wait_before_data(int map_idx)
{
	return wait_for(map_idx, DRQ_STAT);
}

#if 0
static int read_one_sector(int master, unsigned int lba,
			   unsigned char sect[512])
{
	unsigned int *ioports = dumper_priv.io_ports;

	/* Bits 24-27, 0x40=LBA, 0x10=slave */
	if (!wait_before_command()) return -EIO;
	outb(0x40 | (master ? 0 : 0x10) | ((lba >> 24) & 0x0F), ioports[IDE_SELECT_OFFSET]);

	if (!wait_before_command()) return -EIO;
	outb(1, ioports[IDE_NSECTOR_OFFSET]); /* One sector */
	/* Bits 0-7 of LBA address */
	outb(lba & 0xFF, ioports[IDE_SECTOR_OFFSET]);
	/* Bits 8-15 */
	outb((lba >> 8) & 0xFF, ioports[IDE_LCYL_OFFSET]);
	/* Bits 16-23 */
	outb((lba >> 16) & 0xFF, ioports[IDE_HCYL_OFFSET]);
	/* READ SECTOR (with retries) */
	outb(0x20, ioports[IDE_COMMAND_OFFSET]);

	/* Read one sector (256 words) */
	if (!wait_before_data()) return -EIO;
	insw(ioports[IDE_DATA_OFFSET], sect, 256);

#if 0
	printk("Contents of sector %u\n", lba);
	printk("SECTOR: %02X %02X %02X %02X %02X %02X %02X %02X ...\n",
	       sect[0], sect[1], sect[2], sect[3], 
	       sect[4], sect[5], sect[6], sect[7]);
	printk("SECTOR: ... %02X %02X %02X %02X %02X %02X %02X %02X\n",
	       sect[504], sect[505], sect[506], sect[507], 
	       sect[508], sect[509], sect[510], sect[511]);
#endif
	return 0;
}
#endif /* 0 */

static int write_one_sector(int map_idx, int master, unsigned int lba,
			    const unsigned char sect[512])
{
	unsigned long *ioports = dumper_priv[map_idx].io_ports;

	/* Bits 24-27, 0x40=LBA, 0x10=slave */
	if (!wait_before_command(map_idx)) return -EIO;
	outb(0x40 | (master ? 0 : 0x10) | ((lba >> 24) & 0x0F),
		ioports[IDE_SELECT_OFFSET]);
	
	if (!wait_before_command(map_idx)) return -EIO;
	outb(1, ioports[IDE_NSECTOR_OFFSET]); /* One sector */
	/* Bits 0-7 of LBA address */
	outb(lba & 0xFF, ioports[IDE_SECTOR_OFFSET]);
	/* Bits 8-15 */
	outb((lba >> 8) & 0xFF, ioports[IDE_LCYL_OFFSET]);
	/* Bits 16-23 */
	outb((lba >> 16) & 0xFF, ioports[IDE_HCYL_OFFSET]);
	/* WRITE SECTOR (with retries) */
	outb(0x30, ioports[IDE_COMMAND_OFFSET]);

	/* Write one sector (256 words) */
	if (!wait_before_data(map_idx)) return -EIO;
	outsw(ioports[IDE_DATA_OFFSET], sect, 256);
	return 0;
}

#if 0
/* Called with interrupts off */
static int ata_dumper_read_sector(char dump[512])
{
	unsigned int *ioports = dumper_priv.io_ports;
	unsigned lba = get_next_lba();

	/* don't read past the end */
	if(lba == (unsigned long) -1) 
		return -1;		/* EOF */
	
	/* Wait for non-busy: if not, reset anyway */
	wait_before_command();
	/* Soft reset of drive (set nIEN as well) */
	outb(0x0e, ioports[IDE_CONTROL_OFFSET]);
	ata_dumper_usec(1); /* 400ns according to spec */
	outb(0x0a, ioports[IDE_CONTROL_OFFSET]);

	return read_one_sector(dumper_priv.master, lba, dump);
}
#endif /* 0 */

static int ata_dumper_write_sector(int map_idx, const char dump[512])
{
	unsigned lba = get_next_lba(map_idx);

	/* don't write past the end */
	if(lba == (unsigned long) -1) 
		return -1;		/* EOF */

	return write_one_sector(map_idx, dumper_priv[map_idx].master, lba, dump);
}

/* ata_dump_open()
 * This simple open will just rewind to the start of the file.
 */
int ata_dump_open(unsigned int map_id)
{
	int map_idx;

	if ((map_idx = ata_dumper_lookup_map_id(map_id)) < 0) {
		return -1;
	}

	if (dumper_priv[map_idx].dumper_is_open) {
		printk(KERN_ERR "dump already opened\n");
		return -1;
	}

	if (!dumper_priv[map_idx].armed) {
		printk(KERN_ERR "Cannot dump to ATA - not yet armed.\n");
		return -1;
	}

	/* dumper_is_open needs to be protected since this fn initializes some 
	 * critical data for the dumper
	 */
	dumper_priv[map_idx].dumper_is_open = 1;

	curr_blkmap[map_idx] = dumper_priv[map_idx].blkmap;
	curr_lba[map_idx] = curr_blkmap[map_idx]->lba;
	lba_counter[map_idx] = 0;
	write_count[map_idx] = 0;

	/* return index to caller, which caller uses as a handle to
		write and close routines */
	return map_idx;
}

/* ata_dump_close() - This simple close will just close the device.
	return 0 => success, != otherwise
 */
int ata_dump_close(int map_idx)
{
	int ret = 0;

	/* sanity check index */
	if ((map_idx < 0) || (map_idx >= NUM_MAPPINGS))
	{
		printk(KERN_ERR "ata_dump_close(%d): invalid index\n", map_idx);
		return -1;
	}

	if (!dumper_priv[map_idx].dumper_is_open) {
		printk(KERN_ERR "ata_dump_close(%d): dump not open\n", map_idx);
		return -1;
	}

	/* if nothing in the sector buffer, we done */
	if (write_count[map_idx]) {

		kick_watchdog();

		/* zero the space remaining in the sector buffer */
		memset(dumper_priv[map_idx].write_buf + write_count[map_idx], 0,
			SECTOR_SIZE - write_count[map_idx]);

		/* write it to compact flash */
		ret = ata_dumper_write_sector(map_idx, dumper_priv[map_idx].write_buf);
	}

	/* dumper_is_open needs to be protected since ata_dump_open initializes some 
	 * critical data for the dumper
	 */
	dumper_priv[map_idx].dumper_is_open = 0;

	return ret;
}

/* This is our exported function */
int ata_dump_write(int map_idx, char *buf, int size)
{
	int ret;

	/* sanity check index */
	if ((map_idx < 0) || (map_idx >= NUM_MAPPINGS))
	{
		printk(KERN_ERR "ata_dump_write(%d): invalid index\n", map_idx);
		return -1;
	}

	if (!dumper_priv[map_idx].dumper_is_open) {
		printk(KERN_ERR "Dump device not opened yet.\n");
		return -1;
	}

	cli();	/* just to be sure - we don't want or expect interrupts */

	/* while we have data left to write */
	while (size > 0) {
		int copy_len;

		kick_watchdog();

		/* figure out how much data to copy to sector buffer */
		copy_len = min(size, (SECTOR_SIZE - write_count[map_idx]));

		/* copy that much data to the sector buffer */
		memcpy(dumper_priv[map_idx].write_buf +
			write_count[map_idx], buf, copy_len);

		buf += copy_len;
		write_count[map_idx] += copy_len;
		size -= copy_len;

		/* if the sector buffer is now full */
		if (write_count[map_idx] == SECTOR_SIZE) {

			/* write it to compact flash */
			if ((ret = ata_dumper_write_sector(map_idx,
					dumper_priv[map_idx].write_buf))) {
				return ret;
			}

			/* zero the write count so we can buffer some more data */
			write_count[map_idx] = 0;
		}
	}

	return 0;	/* success */
}

/* This is the only part of the in-kernel oops dumper which can call
   into the rest of the kernel. */

/* Protect open and close */
static DECLARE_MUTEX(ata_dumper_lock);

static int ata_dumper_major;

static int ata_dumper_fops_open(struct inode *inode, struct file *filp)
{
	/* nothing to do here really */
	return 0;
}

static int ata_dumper_fops_close(struct inode *inode, struct file *filp)
{
	return 0;
}

/*
	returns the state of the dumper_priv[] array element indexed by map_idx

	assumes that the element with the desired map ID has already
		been located via a search of the array and that map_idx
		indexes that element.

	see the big comment block at the beginning of this file for a
	description of these states.
*/

#define ATA_DUMPER_ALLOCATED_STATE 1
#define ATA_DUMPER_MAPPED_STATE 2
#define ATA_DUMPER_ARMED_STATE 3

static int ata_dumper_get_state(int map_idx)
{
	if (dumper_priv[map_idx].armed != 0)
		return ATA_DUMPER_ARMED_STATE;
	if (dumper_priv[map_idx].num_blkmaps != 0)
		return ATA_DUMPER_MAPPED_STATE;
	return ATA_DUMPER_ALLOCATED_STATE;
}

extern ide_drive_t *ide_disk_to_drive(struct gendisk *);

static int ata_dumper_fops_ioctl(struct inode *inode, struct file *filp,
				unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	unsigned int map_id;
	int map_idx;
	int state;

	if (down_interruptible(&ata_dumper_lock) != 0)
		return -EINTR;

	switch (cmd) {

	case ATA_DUMPER_IOC_ALLOC:
	{
		/* the application puts the desired map ID in arg */
		map_id = (unsigned int) arg;

		/* valid map IDs are non-zero */
		if (map_id == 0) {
			ret = -EFAULT;
			break;
		}

		/* make sure this ID is not already mapped */
		if (ata_dumper_lookup_map_id(map_id) >= 0) {
			ret = -EADDRINUSE;
			break;
		}

		/* now find an unused dumper_priv[] array element */
		if ((map_idx = ata_dumper_lookup_map_id(0)) < 0) {
			ret = -ENOSPC;
			break;
		}

		/* allocate a sector buffer for the mapping */
		if ((dumper_priv[map_idx].write_buf =
				kmalloc(SECTOR_SIZE, GFP_KERNEL)) == NULL) {
			ret = -ENOMEM;
			break;
		}

		/* mark this element as allocated to this mapping */
		dumper_priv[map_idx].map_id = map_id;

		break;
	}

	case ATA_DUMPER_IOC_FREE:
	{
		/* the application puts the desired map ID in arg */
		map_id = (unsigned int) arg;

		/* valid map IDs are non-zero */
		if (map_id == 0) {
			ret = -EFAULT;
			break;
		}

		/* find the array element allocated to this ID  */
		if ((map_idx = ata_dumper_lookup_map_id(map_id)) < 0) {
			ret = -ENXIO;
			break;
		}

		/* we only allow a free in Allocated State */
		if (ata_dumper_get_state(map_idx) != ATA_DUMPER_ALLOCATED_STATE) {
			ret = -EPERM;
			break;
		}

		/* free the sector buffer */
		if (dumper_priv[map_idx].write_buf) {
			kfree(dumper_priv[map_idx].write_buf);
			dumper_priv[map_idx].write_buf = NULL;
		} else {
			/* this shouldn't happen */
			printk(KERN_ERR "ata_dumper_ioctl(), write_buf NULL\n");
		}

		/* this dump_priv[] element is now unallocated */
		dumper_priv[map_idx].map_id = 0;

		break;
	}

	case ATA_DUMPER_IOC_CLEAR_MAP: 
	{
		/* the application puts the desired map ID in arg */
		map_id = (unsigned int) arg;

		/* valid map IDs are non-zero */
		if (map_id == 0) {
			ret = -EFAULT;
			break;
		}

		/* find the array element allocated to this ID  */
		if ((map_idx = ata_dumper_lookup_map_id(map_id)) < 0) {
			ret = -ENXIO;
			break;
		}

		state = ata_dumper_get_state(map_idx);

		/* a CLEAR_MAP in Allocated State is a no-op, so just break */
		if (state == ATA_DUMPER_ALLOCATED_STATE) {
			break;
		}

		/* otherwise the only valid state for a CLEAR_MAP is Mapped */
		if (state != ATA_DUMPER_MAPPED_STATE) {
			ret = -EPERM;
			break;
		}

		/* delete the entire block map */
		if (dumper_priv[map_idx].blkmap) {
			kfree(dumper_priv[map_idx].blkmap); 	
			dumper_priv[map_idx].blkmap = NULL;
		} else {
			/* this shouldn't happen */
			printk(KERN_ERR "ata_dumper_ioctl(), blkmap NULL\n");
		}
		dumper_priv[map_idx].num_blkmaps = 0;

		break;
	}

	case ATA_DUMPER_IOC_ADD_MAP:
	{
		int i, j, size, discontig, num_fs_blks;
		unsigned long start, spb, start_lba;
		int sm = 1; /* sectors multiplier, for xfs */
		struct blkmap *lbamap;
		struct ata_dumper_blkmap uspace_blkmap;
		unsigned long *fsblocks;
		struct block_device *bdev;
		dev_t dumpdev;

		/* application passes pointer to ata_dumper_blkmap in arg */

		/* FIXME, is there some way to validate arg before passing
			it to copy_from_user(), or will copy_from_user() fail
			gracefully if arg is bogus? */

		/* get local copy of mapping structure */
		if (copy_from_user((void*)&uspace_blkmap,
				(void*)arg, sizeof(uspace_blkmap))) {
			ret = -EFAULT;
			break;
		}

		map_id = uspace_blkmap.map_id;

		/* valid map IDs are non-zero */
		if (map_id == 0) {
			ret = -EFAULT;
			break;
		}

		/* find the array element allocated to this ID  */
		if ((map_idx = ata_dumper_lookup_map_id(map_id)) < 0) {
			ret = -ENXIO;
			break;
		}

		state = ata_dumper_get_state(map_idx);

		/* an ADD_MAP is only allowed in Allocated State */
		if (state != ATA_DUMPER_ALLOCATED_STATE) {
			ret = -EPERM;
			break;
		}

		/* allocate a local buffer for the application block map */
		if (!(fsblocks = vmalloc(uspace_blkmap.num_blocks *
				sizeof(unsigned long)))) {
			ret = -ENOMEM;
			break;
		}

		/* make local copy of application block map */
		if (copy_from_user(fsblocks, (void*)uspace_blkmap.blks, 
				uspace_blkmap.num_blocks * sizeof(unsigned long))) {
			vfree(fsblocks);
			ret = -EFAULT;
			break;
		}

		/* figure out how many discontiguous blocks we have,
		 * this determines how large our blkmap array has to be.
		 */
		discontig = 1;
		for(i=0; i<uspace_blkmap.num_blocks-1; i++) {
			if(fsblocks[i+1] != (fsblocks[i] + 1))
				discontig++;
		}
		
		/* allocate space for our LBA blkmap */
		size = sizeof(struct blkmap) * discontig;
		if((lbamap = kmalloc(size, GFP_KERNEL)) == NULL) {
			vfree(fsblocks);
			ret = -ENOMEM;
			break;
		}
		memset(lbamap, 0, size);

		/* Get the block device pointer */
		dumpdev = old_decode_dev(uspace_blkmap.device);
		if (!(bdev = bdget(dumpdev))) {
			vfree(fsblocks);
			kfree(lbamap);
			ret = -ENODEV;
			break;
		}

		/* if XFS based platform sector multiplier is 8 */
		/* in future find the true file system type, we */
		/* cannot here, because this is the raw block device */
		if ((silkworm.board==SWBD21) 
			|| (silkworm.board==SWBD12)
			|| (silkworm.board==SWBD22)) sm=8;
		if (sm==8) printk(KERN_DEBUG "ATA_DUMPER using XFS sectors map %i\n",map_id);
		if (!blkdev_get(bdev, FMODE_READ, 0)) {
			spb = sm * bdev->bd_block_size / SECTOR_SIZE;
			start = get_start_sect(bdev);
			/*
			 * get_super() and drop_super()
			 * must be called in pair to release
			 * the "unmount" semaphore
			 */
			blkdev_put(bdev);
		} else {
			/*
			 * we cannot get the super block
			 * of the file system, return here
			 */
			vfree(fsblocks);
			kfree(lbamap);
			ret = -EFAULT;
			break;
		}

		/* fill in our LBA blkmap with real LBA sector numbers */
		i = 0;	/* index into uspace_blkmap.blks[] */
		j = 0;	/* index into LBA blkmap[] */
		while(i < uspace_blkmap.num_blocks) {

			if(j >= discontig)	
				BUG();	/* If this happens, we have a logic error somewhere */

			/* starting LBA of this FS block */
			start_lba = (fsblocks[i] * spb) + start;

			/* Number of contiguous FS blocks */
			num_fs_blks = 0;
			do {
				num_fs_blks++;
				i++;
			} while(i < uspace_blkmap.num_blocks && 
				uspace_blkmap.blks[i] == uspace_blkmap.blks[i-1] + 1) ;

			lbamap[j].lba = start_lba;
			lbamap[j].num = num_fs_blks * spb;
			j++;
		}

		/* delay updating the dumper_priv element fields until we know
			this function will succeed */
		dumper_priv[map_idx].dumpdev = dumpdev;
		dumper_priv[map_idx].num_blkmaps = j;
		dumper_priv[map_idx].blkmap = lbamap;
		dumper_priv[map_idx].sec_p_blk = spb;

		vfree(fsblocks);
		break;
	}

	case ATA_DUMPER_IOC_ARM:
	{
		int i;
		ide_drive_t *drive;
		struct block_device *bdev;

		/* the application puts the desired map ID in arg */
		map_id = (unsigned int) arg;

		/* valid map IDs are non-zero */
		if (map_id == 0) {
			ret = -EFAULT;
			break;
		}

		/* find the array element allocated to this ID  */
		if ((map_idx = ata_dumper_lookup_map_id(map_id)) < 0) {
			ret = -ENXIO;
			break;
		}

		state = ata_dumper_get_state(map_idx);

		/* an ARM in Armed State is a no-op, so just break */
		if (state == ATA_DUMPER_ARMED_STATE) {
			break;
		}

		/* otherwise the only valid state for an ARM is Mapped */
		if (state != ATA_DUMPER_MAPPED_STATE) {
			ret = -EPERM;
			break;
		}

		/* mark the map as complete and arm for a dump */
		if (!(bdev = bdget(dumper_priv[map_idx].dumpdev))) {
			ret = -ENODEV;
			break;
		}
		drive = ide_disk_to_drive(bdev->bd_disk);
		if (!drive || !HWIF(drive)) {
			printk("ARM: error to get HWIF\n");
			ret = -ENXIO;
			break;
		}

		/* Grab the I/O addresses for the drive's task file registers */
		memcpy(dumper_priv[map_idx].io_ports,
			HWIF(drive)->io_ports, sizeof(dumper_priv[map_idx].io_ports));
		dumper_priv[map_idx].total_lba = 0;
		for (i = 0; i < dumper_priv[map_idx].num_blkmaps; i++)
			dumper_priv[map_idx].total_lba +=
				 dumper_priv[map_idx].blkmap[i].num;

		dumper_priv[map_idx].master = 1;		//FIXME
		dumper_priv[map_idx].armed = 1;

		break;
	}

	case ATA_DUMPER_IOC_DISARM:
	{
		/* the application puts the desired map ID in arg */
		map_id = (unsigned int) arg;

		/* valid map IDs are non-zero */
		if (map_id == 0) {
			ret = -EFAULT;
			break;
		}

		/* find the array element allocated to this ID  */
		if ((map_idx = ata_dumper_lookup_map_id(map_id)) < 0) {
			ret = -ENXIO;
			break;
		}

		state = ata_dumper_get_state(map_idx);

		/* a DISARM in any state except Armed is a no-op, so just break */
		if (state != ATA_DUMPER_ARMED_STATE) {
			break;
		}

		/* flag that this mapping is unarmed */
		dumper_priv[map_idx].armed = 0;

		/* clearing the following fields just for completeness */
		memset(dumper_priv[map_idx].io_ports, 0,
			sizeof(dumper_priv[map_idx].io_ports));
		dumper_priv[map_idx].total_lba = 0;
		dumper_priv[map_idx].master = 0;

		break;
	}
	case ATA_DUMPER_IOC_TEST_EXCEPTION:
		goto test_exception;
	case ATA_DUMPER_IOC_TEST_DIE:
		goto test_die;
	default:
		printk("error: default case\n");
		ret = -EINVAL;
		break;
	}
	up(&ata_dumper_lock);
	return ret;

 test_exception:
	up(&ata_dumper_lock);
	*((char *)0) = 0;

 test_die:
	up(&ata_dumper_lock);
	//die("ATA dumper testing die()");
	return 0;	/* never get here */
}

static struct file_operations ata_dumper_fops = {
	open:		ata_dumper_fops_open,
	release:	ata_dumper_fops_close,
	ioctl:		ata_dumper_fops_ioctl,
};


static void *ata_dumper_status_read_seq_start(struct seq_file *m, loff_t *pos)
{
	loff_t npos = *pos; /* logical integer position in virtual file */
	dumper_priv_type *iterator;

	/* 0 thru (NUM_MAPPINGS - 1) implies print that dumper_priv element,
		NUM_MAPPINGS implies print the dumper_priv_global data,
		so if beginning position is greater than NUM_MAPPINGS, then
		its past the end of the virtual file.
		the check for negative is paranoia */
	if ((npos > NUM_MAPPINGS) || (npos < 0)) {
		return NULL;
	}

	/* our iterator is just the pointer to the dumper_priv[] array element */
	iterator = dumper_priv + (int) npos;
	return iterator;
}

static void *ata_dumper_status_read_seq_next(struct seq_file *m,
	void *v, loff_t *pos)
{
	dumper_priv_type *iterator = v;

	(*pos)++; /* increment the logical position in the virtual file */

	/* if that puts us past logical end of virtual file, return NULL */
	if (*pos > NUM_MAPPINGS) {
		return NULL;
	}

	/* bump our iterator and return it */
	return iterator + 1;
}

static void ata_dumper_status_read_seq_stop(struct seq_file *m, void *v)
{
	/* nothing for us to free up */
}

static int ata_dumper_status_read_seq_show(struct seq_file *m, void *v)
{
	dumper_priv_type *p = v;
	int i;
	struct block_device *bdev;
	unsigned long size = 0;

	/* sanity check the iterator we got passed */
	if ((p < dumper_priv) || (p > dumper_priv + NUM_MAPPINGS)) {
		return -1; /* something stinks in finland, bail */
	}

	/* print the global data */
	if (p == dumper_priv + NUM_MAPPINGS) {

		seq_printf(m, "TMB ticks/jiffy: %d\n", dumper_priv_global.ticks_jiffy);
		seq_printf(m, "TMB ticks/usec:  %d\n", dumper_priv_global.ticks_usec);
		return 0;
	}

	/* print the dumper_priv[] array element addressed by iterator */

	/* skip unallocated array elements */
	if (p->map_id == 0) {
		return 0;
	}
	
	/* print array element index */
	seq_printf(m, "ATA polled-mode dumper[%d] ID: %d (%p)\n",
		p - dumper_priv, p->map_id, p);

	seq_printf(m, "Armed, ready to dump: %s\n", (p->armed ? "Yes" : "No"));

	if ((bdev = bdget(p->dumpdev))) {
		seq_printf(m, "Dump device: %d\n", p->dumpdev);
	}

	/* compute and print size of mapped file */
	size = 0;
	for (i = 0; i < p->num_blkmaps; i++)
		size += p->blkmap[i].num;
	size *= SECTOR_SIZE;

	seq_printf(m, "Size of dump file: %lu bytes\n", size);

	seq_printf(m, "Number of blkmap entries: %d\n", p->num_blkmaps);

	seq_printf(m, "total_lba: %ld\n", p->total_lba);
	seq_printf(m, "blkmap: %p\n", p->blkmap);
	if (p->blkmap) {
		seq_printf(m, "First LBA block: %lu\n", p->blkmap[0].lba);
		seq_printf(m, "Last LBA block: %lu\n", 
			p->blkmap[p->num_blkmaps - 1].lba + 
			p->blkmap[p->num_blkmaps - 1].num - 1);
		seq_printf(m, "Size of blkmap array: %lu bytes\n", 
			(unsigned long) &p->blkmap[p->num_blkmaps] -
				(unsigned long) &p->blkmap[0]);
	}

	seq_printf(m, "Filesystem block size: %u bytes\n", 
			p->sec_p_blk * SECTOR_SIZE);
	seq_printf(m, "ATA registers located at:\n");
	for (i = 0; i < IDE_NR_PORTS; i++) {
		seq_printf(m, "0x%lx\n", p->io_ports[i]);
	}

	return 0;
}

static struct seq_operations ata_dumper_status_read_devices_op = {
	start:	ata_dumper_status_read_seq_start,
	next:	ata_dumper_status_read_seq_next,
	stop:	ata_dumper_status_read_seq_stop,
	show:	ata_dumper_status_read_seq_show
};

static int ata_dumper_status_read_dev_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &ata_dumper_status_read_devices_op);
}

static struct file_operations ata_dumper_status_read_ops = {
	open:		ata_dumper_status_read_dev_open,
	read:		seq_read,
	llseek:		seq_lseek,
	release:	seq_release,
};

/* FIXME: write-protect the module pages */
static int __init init(void)
{
	struct proc_dir_entry *ata_dumper_dir;
	struct proc_dir_entry *entry;

	/* clear the whole mapping array */
	memset(&dumper_priv, 0, sizeof(dumper_priv));

	/* clear the "global" data */
	memset(&dumper_priv_global, 0, sizeof(dumper_priv_global));

	if(!tb_ticks_per_jiffy) {
		printk(KERN_ERR "dumper_init: timebase frequency not initialized\n");
		return -1;
	}
	dumper_priv_global.ticks_jiffy = tb_ticks_per_jiffy;
	dumper_priv_global.ticks_usec =
		((unsigned long)tb_ticks_per_jiffy * HZ) / 1000000;

	ata_dumper_major = register_chrdev(0, "ata_dumper", &ata_dumper_fops);

	ata_dumper_dir = proc_mkdir("driver/ata_dumper", NULL);
	if (!ata_dumper_dir) {
		printk(KERN_ERR "dumper_init: proc_mkdir() failed\n");
		return -ENOMEM;
	}

	entry = create_proc_entry("status", 0, ata_dumper_dir);
	if (!entry) {
		printk(KERN_ERR "dumper_init: create_proc_entry(status) failed\n");
		return -ENOMEM;
	}

	entry->proc_fops = &ata_dumper_status_read_ops;

	printk("ATA polled-mode panic dumper on char-major-%d.\n", ata_dumper_major);
	return ((ata_dumper_major >= 0) ? 0 : ata_dumper_major);
}

static void __exit fini(void)
{
	unregister_chrdev(ata_dumper_major, "ata_dumper");
}

module_init(init);
module_exit(fini);

EXPORT_SYMBOL(ata_dump_open);
EXPORT_SYMBOL(ata_dump_close);
EXPORT_SYMBOL(ata_dump_write);

