/*
 *    Copyright (c) 2004-2010 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *	Note: We should combine the platform-specific IDE routines
 *		in silkworm.c and silkworm440.c and move them
 *		here so we need to maintain one set of routines
 *		only.
 */

#include <linux/types.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/delay.h>

#include <linux/ide.h>
#include <brocade/cpld.h>
#include <brocade/platform.h>
#include <brocade/silkworm.h>

/*
 * WARNING!! WARNING!!
 *
 * When we add a platform, we need to implement a function to
 * hard reset the compactFlash card. This is to work around a
 * Toshiba compactFlash card hang problem.
 *
 * For a platform whose IDE interface is implemented using CPLD,
 * The hard reset function is cpld_ide_hard_reset(). However,
 * we will need to make sure the CompactFlash reset enable
 * and control registers are defined properly for that platform.
 * They should be defined in linux/include/brocade/cpld.h.
 */

static	uint8_t *cpld_base = 0;
static	int	rst_ctrl_reg = 0;
static	int 	cf_rst_ctrl_offset = 0;
static	int	rst_enable_reg = 0;
static	int	cf_rst_enable_offset = 0;		

extern sw_bd silkworm;

extern	unsigned int cmd64x_reset(ide_drive_t *drive);
extern	unsigned int cmd64x_reset_20ms(ide_drive_t *drive);
int	poll_start_time;

static unsigned int
cpld_can_reset(void)
{
        static uint8_t cpld_data;
        int can_reset = 0;

	/* Write to Reset enable register */
        cpld_data = readb(cpld_base + rst_enable_reg);
        cpld_data |= cf_rst_enable_offset;
        writeb(cpld_data, cpld_base + rst_enable_reg);

	/* Read from Reset enable register */
        cpld_data = readb(cpld_base + rst_enable_reg);
	can_reset = cpld_data & cf_rst_enable_offset;

        /* Disable CF reset */
        cpld_data = readb(cpld_base + rst_enable_reg);
        cpld_data &= ~cf_rst_enable_offset;
        writeb(cpld_data, cpld_base + rst_enable_reg);

        return (can_reset);
}

static unsigned int
cpld_ide_hard_reset(ide_drive_t *drive)
{
        static unsigned long flags;
        static uint8_t cpld_data;
        int timeout;


        /* Enable CF reset */
        cpld_data = readb(cpld_base + rst_enable_reg);
        cpld_data |= cf_rst_enable_offset;
        writeb(cpld_data, cpld_base + rst_enable_reg);

        /* Assert CF reset line  */
        cpld_data = readb(cpld_base + rst_ctrl_reg);
        cpld_data |= cf_rst_ctrl_offset;
        writeb(cpld_data, cpld_base + rst_ctrl_reg);

        /* Wait for 100 ms */
        local_irq_save(flags);   /* local CPU only */
        local_irq_enable();      /* local CPU only */

        timeout = jiffies + 10;             /* timeout is 100 ms */

        /* Wait for 100 ms */
        while (time_after(timeout, jiffies));

        local_irq_restore(flags); /* local CPU only */

        /* De-assert CF reset line */
        cpld_data = readb(cpld_base + rst_ctrl_reg);
        cpld_data &= ~cf_rst_ctrl_offset;
        writeb(cpld_data, cpld_base + rst_ctrl_reg);

        /* Disable CF reset */
        cpld_data = readb(cpld_base + rst_enable_reg);
        cpld_data &= ~cf_rst_enable_offset;
        writeb(cpld_data, cpld_base + rst_enable_reg);


#if 0
	/* Defect 57170, remove the messages as they worry a customer */
	printk("IDE CPLD reset done!\n");
#endif

        return (0);
}

static unsigned int
saturn_ide_hard_reset(ide_drive_t *drive)
{
        static unsigned long flags;
        static uint8_t cpld_data;
        int timeout;

        /* Assert CF reset line  */
        cpld_data = readb(cpld_base + rst_ctrl_reg);
        cpld_data |= cf_rst_ctrl_offset;
        writeb(cpld_data, cpld_base + rst_ctrl_reg);

        /* Wait for 100 ms */
        local_irq_save(flags);   /* local CPU only */
        local_irq_enable();      /* local CPU only */

        timeout = jiffies + 2;             /* timeout is 100 ms */

        /* Wait for 100 ms */
        while (time_after(timeout, jiffies));

        local_irq_restore(flags); /* local CPU only */

        /* De-assert CF reset line */
        cpld_data = readb(cpld_base + rst_ctrl_reg);
        cpld_data &= ~cf_rst_ctrl_offset;
        writeb(cpld_data, cpld_base + rst_ctrl_reg);

#if 0
	/* Defect 57170, remove the messages as they worry a customer */
	printk("IDE CPLD reset done!\n");
#endif

        return (0);
}

void silkworm_ide_init(void)
{
	switch (silkworm.board) {
		case SWBD22:	/* Blazer */
		case SWBD26:	/* Dazzler 16 port */
		case SWBD27:	/* Dazzler 8 port */
		case SWBD33:	/* Gracier */
        		if ((cpld_base = (void*) ioremap(CPLD_BASE_ADDRESS, CPLD_MAX_REGISTERS)) == NULL) {
                		printk(KERN_WARNING "cpld_ide_hard_reset: failed ioremap of cpld\n");
				break;
        		}
			rst_ctrl_reg = CPLD_SYS_RESET_REG;
			cf_rst_ctrl_offset = CPLD_CF_RESET;
			rst_enable_reg = CPLD_RESET_ENABLE_REG;
			cf_rst_enable_offset = CPLD_CF_RESET_ENABLE;		

			/* CPLD doesn't support hard reset */
			if (!cpld_can_reset()) {
				printk("CPLD doesn't support IDE hard reset.\n");
				break;
			}
			ide_hard_reset = cpld_ide_hard_reset;
			poll_start_time = HZ/20;
			break;
			
		case SWBD29:		/* Icecube */
		case SWBD32:		/* Pulsar32 */
		case SWBD23:		/* Pulsar */
		case SWBD34:		/* Stealth */
		case SWBD36:		/* Marathon */
		case SWBD37:		/* Blazer2 */
		case SWBD43:		/* Icecube2 */
		case SWBD44:		/* Viking */
		case SWBD45:		/* Glacier2 */
		case SWBD51:		/* Dragon */
		case SWBD53:		/* Triathlon */
		case SWBD55:		/* Scimitar */
		case SWBD58:		/* Pulsar2 */
		case SWBD61:		/* Superhawk */
		case SWBD62:		/* Neptune */
		case SWBD63:		/* Neptune CO-Processor */
		case SWBD64:		/* Thor */
		case SWBD66:		/* Spirit(440EPX) */
		case SWBD67:		/* MACE */
		case SWBD68:		/* Lance */
		case SWBD70:		/* LittleBear 8G */
		case SWBD71:		/* Tomahawk(440EPX) */
		case SWBD72:        /* Bulova(440EPX) */
		case SWBD73:        /* Blazer3(Dirk) */
		case SWBD75:		/* Superhawk2  */
		case SWBD76:		/* ELARA */
		case SWBD77:		/* Pluto */
		case SWBD78:		/* Leda */
		case SWBD79:		/* Spear CP */
		case SWBD82:        /* Utah(440EPX) */
		case SWBD83:		/* Spike */
		case SWBD85:		/* Spike-BP */
		case SWBD86:		/* Falcon2 */
		case SWBD87:		/* Kestrel2 */
		case SWBD88:		/* Europa */
		case SWBD89:		/* Striker */
		case SWBD92:		/* Gladius(440EPX) */
		case SWBD109:		/* Stinger(440EPX) */
		case SWBD124:		/* Dragon2 */
		case SWBD125:           /* Dragon2T */ 
		case SWBD134:           /* Qilin */        
		case SWBD117:		/* Pharos(440EPX) */
		case SWBD118:		/* Tomtoo(440EPX) */
		case SWBD129:		/* Bulova2(440EPX) */
		case SWBD130:		/* Superhawk3(440EPX) */
		case SWBD149:            /* Kestrel3 */
		case SWBD150:		/* Falcon3(440EPX) */
		//case SWBD133:		/* Odin */
			if ((cpld_base = (void*) ioremap(PULSAR_CPLD_BASE_ADDR, CPLD_MAX_REGISTERS)) == NULL) {
				printk(KERN_WARNING "cpld_ide_hard_reset: failed ioremap of cpld\n");
				break;
			}
			rst_ctrl_reg = PULSAR_INDIVIDUAL_RESET_CNTRL_REG;
			cf_rst_ctrl_offset = PULSAR_CF_RESET;
			rst_enable_reg = PULSAR_INDIVIDUAL_RESET_ENABLE_REG;
			cf_rst_enable_offset = PULSAR_CF_RESET_ENABLE;

			/* CPLD doesn't support hard reset */
			if (!cpld_can_reset()) {
				printk("CPLD doesn't support IDE hard reset.\n");
				break;
			}
			ide_hard_reset = cpld_ide_hard_reset;
			poll_start_time = HZ/20;
			break;
		case SWBD42:		/* Saturn */
		case SWBD46:		/* Sprint */
		case SWBD133:
        		if ((cpld_base = (void*) ioremap(PULSAR_CPLD_BASE_ADDR, CPLD_MAX_REGISTERS)) == NULL) {
                		printk(KERN_WARNING "cpld_ide_hard_reset: failed ioremap of cpld\n");
				break;
        		}

			rst_ctrl_reg = SATURN_BOOTROM_CTRL_REG;
			cf_rst_ctrl_offset = SATURN_REV_ATA_RESET;

			ide_hard_reset = saturn_ide_hard_reset;
			poll_start_time = 80;
			break;
		case SWBD12:
			ide_hard_reset = cmd64x_reset;
			poll_start_time = HZ/20;
			break;
		case SWBD21:
			ide_hard_reset = cmd64x_reset_20ms;
			/*
		 	 * On Meteor platform, CPLD asserts the reset line. Wait for 800 ms
		 	 * before polling IDE status
		 	 */
			poll_start_time = 80;
			break;
		case SWBD10:
			/*
		 	 * On Ulysses platform, socket reset takes more than 100ms.  Wait 
			 * for 200ms before polling IDE status
			 * Ulysses ide_hard_reset function is registered in pcmcia module
		 	 */
			poll_start_time = 20;
			break;
		default:
			printk("Warning! IDE hard reset routine has not been implemented!\n");
			break;
	}

	return;
}
