/* 
 * Polled-mode flash dumper used to write kernel crash dump
 * data to the Intel Strataflash.
 *
 * Both Ulysses and Terminator use Strataflash in a 2 x 16
 * configuration.
 *
 * This code bypasses the MTD driver and is intended to be
 * called only from the panic() function with interrupts
 * disabled.  No functions here will block or schedule.
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/proc_fs.h>
#include <asm/processor.h>
#include <asm/io.h>
#include <linux/flash_dumper.h>
#include <linux/vmalloc.h>		/* testing only */
#include <linux/interrupt.h>

#define CMD(x)	(dumper_priv.width == 2 ? x : \
		(dumper_priv.width == 4 ? \
		(((unsigned long)x) | (((unsigned long)x)<<16)) : x))

#ifndef min
#define min(x,y) ( (x)<(y)?(x):(y) )
#endif

#define DPRINTK(args...)	
//#define DPRINTK(args...)	kick_watchdog(); printk(## args)

/* This data is initialized at boot time, and must be intact for the
 * dumper to work.  Therefore we protect all this data with a checksum.
 */
static struct {
	void *base;		/* base address of the flash partition */
	unsigned long size;	/* size of the partition */
	int sector_size; 	/* size of the erase-unit */
	int width;		/* width of the flash */
	uint8_t *cpld_addr;	/* for controling Vpp */
	uint8_t vpp_offset;	/* cpld register offset */
	uint8_t vpp_mask;	/* mask bits */
	struct proc_dir_entry *proc_entry;
	struct proc_dir_entry *proc_entry2;
	unsigned ticks_jiffy;	/* how many timebase ticks per jiffy */
} dumper_priv;

/* This data is created during the dump, so they don't have to be 
 * protected by checksum
 */
static int dumper_fptr;
static int dumper_is_open;

extern unsigned tb_ticks_per_jiffy;

static __inline__ void ENABLE_VPP(void)
{
	uint8_t *reg = dumper_priv.cpld_addr + dumper_priv.vpp_offset;
	*reg |= dumper_priv.vpp_mask;
	eieio();
}

static __inline__ void DISABLE_VPP(void)
{
	uint8_t *reg = dumper_priv.cpld_addr + dumper_priv.vpp_offset;
	*reg &= ~dumper_priv.vpp_mask;
	eieio();
}

#define cfi_read(offset, data) {		\
	if (dumper_priv.width == 4) {	\
		*(data) = (*(unsigned long*)((offset) + dumper_priv.base));	\
	} else if (dumper_priv.width == 2) {				\
		*(data) = (*(unsigned short*)((offset) + dumper_priv.base));\
	}				\
}

#define cfi_write(data, offset)	{ 	\
	if (dumper_priv.width == 4) {	\
		*(unsigned long*)((offset) + dumper_priv.base) = (data);	\
	} else if (dumper_priv.width == 2) {					\
		*(unsigned short*)((offset) + dumper_priv.base) = (data);	\
	}				\
	eieio();			\
}					

/* flash_write_word() - write a 32 bit word to flash.  According to the intel datasheet, a word write
 * can take as long as 630usec.  The fastest watchdog frequency is 167msec.  This means we
 * should refresh the watchdog at least every 250 words.
 */
static int flash_write_word(unsigned long offset, int data)
{
	int ret = -1;
	unsigned long status, status_OK;
	unsigned long timeout = 1000000; //TBD

	DPRINTK("write_word: 0x%08x at 0x%08lx\n", data, offset);
	DPRINTK("CMD(0x40) = 0x%08lx\n", CMD(0x40));

	ENABLE_VPP();
	cfi_write(CMD(0x40), offset);
	cfi_write(data, offset);
	status_OK = CMD(0x80);
	while(timeout-- != 0)
	{
		cfi_read(offset, &status);
		if((status & status_OK) == status_OK)
		{
			DPRINTK("write_word() finished with timeout = %lu\n", timeout);
			ret = 0;
			break;
		}

	}
	if(ret != 0)
	{
		DPRINTK(KERN_ERR "write_word() timed out at offset 0x%08lx. "
				"status = 0x%08lx.\n", offset, status);
		ret = -1;
	}

	DISABLE_VPP();
	if (status & CMD(0x02)) {
		/* clear status */
		cfi_write(CMD(0x50), offset);
		/* put back into read status register mode */
		cfi_write(CMD(0x70), offset);
	}

	/* put device back into read mode */
	cfi_write(CMD(0xFF), offset);
	return ret;
}

/* flash_write_hword() - write a 16 bit word to flash.  According to the intel datasheet, a word write
 * can take as long as 630usec.  The fastest watchdog frequency is 167msec.  This means we
 * should refresh the watchdog at least every 250 words.
 */
static int flash_write_hword(unsigned long offset, short data)
{
	int ret = -1;
	unsigned long status, status_OK;
	unsigned long timeout = 1000000; //TBD

	DPRINTK("write_hword: 0x%08x at 0x%08lx\n", data, offset);
	DPRINTK("CMD(0x40) = 0x%08lx\n", CMD(0x40));

	ENABLE_VPP();
	cfi_write(CMD(0x40), offset);
	cfi_write(data, offset);
	status_OK = CMD(0x80);
	while(timeout-- != 0)
	{
		cfi_read(offset, &status);
		if((status & status_OK) == status_OK)
		{
			DPRINTK("write_hword() finished with timeout = %lu\n", timeout);
			ret = 0;
			break;
		}

	}
	if(ret != 0)
	{
		DPRINTK(KERN_ERR "write_hword() timed out at offset 0x%08lx. "
				"status = 0x%08lx.\n", offset, status);
		ret = -1;
	}

	DISABLE_VPP();
	if (status & CMD(0x02)) {
		/* clear status */
		cfi_write(CMD(0x50), offset);
		/* put back into read status register mode */
		cfi_write(CMD(0x70), offset);
	}

	/* put device back into read mode */
	cfi_write(CMD(0xFF), offset);
	return ret;
}

/* flash_write_buffer() - write up to 16 words at once.  According to the intel datasheet,
 * a buffer write can take as long as 654usec.  The fastest watchdog frequency is 167msec.  This means we
 * should refresh the watchdog at least every 250 writes.
 */
static int flash_write_buffer(unsigned long offset, int *data, int count)
{
	int i;
	int ret = -1;
	unsigned long status, status_OK;
	unsigned long timeout = 1000000; //TBD
	short *hwptr;

	DPRINTK("CMD(0xE8) = 0x%08lx\n", CMD(0xE8));

	if (dumper_priv.width == 2) {
		hwptr = (short *)data;
		count = (count << 1);
	}

	DPRINTK("write_buffer: 0x%08x words at 0x%08lx\n", count, offset);
	ENABLE_VPP();
	status_OK = CMD(0x80);
	while(timeout-- != 0)
	{
		/* Issue write to buffer command */
		cfi_write(CMD(0xE8), offset);

		cfi_read(offset, &status);
		if((status & status_OK) == status_OK)
		{
			ret = 0;
			break;
		}

	}
	if(ret != 0)
	{
		DPRINTK(KERN_ERR "write_buffer() (part 1) timed out at offset 0x%08x. "
				"status = 0x%08lx.\n", offset, status);
	}

	/* Issue word count and data, then confirm cycle */
	cfi_write(CMD(count-1), offset);
	for(i=0; i<count; i++) {
		if (dumper_priv.width == 2) {					
			cfi_write(hwptr[i], offset+(i*sizeof(uint16_t)));
		} else {
			cfi_write(data[i], offset+(i*sizeof(uint32_t)));
		}
	}
	cfi_write(CMD(0xD0), offset);

	ret = -1;
	timeout = 10000000;
	while(timeout-- != 0)
	{
		cfi_read(offset, &status);
		if((status & status_OK) == status_OK)
		{
			DPRINTK("write_buffer() finished with timeout = %lu\n", timeout);
			ret = 0;
			break;
		}
	}
	if(ret != 0)
	{
		DPRINTK(KERN_ERR "write_buffer() timed out at offset 0x%08x. "
				"status = 0x%08x.\n", offset, status);
		ret = -1;
	}

	DISABLE_VPP();
	if (status & CMD(0x32)) {
		/* clear status */
		cfi_write(CMD(0x50), offset);
		/* put back into read status register mode */
		cfi_write(CMD(0x70), offset);
	}

	/* put device back into read mode */
	cfi_write(CMD(0xFF), offset);
	return ret;
}

/*
 * erase_sector() - erase a complete sector at the given offset.
 * According to the intel strataflash datasheet, worst case sector erase time is 5.0 second.
 */
static int erase_sector(unsigned long offset)
{
	int ret = -1;
	unsigned long status, status_OK;
	unsigned long timeout = 10000000;  /* appr 10 seconds */

	ENABLE_VPP();

	/* Clear the status register first */
	cfi_write(CMD(0x50), offset);

	/* Now erase */
	cfi_write(CMD(0x20), offset);
	cfi_write(CMD(0xD0), offset);
	status_OK = CMD(0x80);

	while(timeout-- != 0)
	{
		/* kick watchdog periodically */
		if((timeout % 100) == 0)
			kick_watchdog();

		cfi_read(offset, &status);
		if((status & status_OK) == status_OK)
		{
			DPRINTK("erase_sector() finished with timeout = %d\n", 
				timeout);
			ret = 0;
			break;
		}
	}
	if(ret != 0)
	{
		DPRINTK(KERN_ERR "erase_sector() timed out at offset 0x%08x. "
				"status = 0x%08x.\n", offset, status);
		ret = -1;
	}


	/* clear status */
	cfi_write(CMD(0x50), offset);
	/* put back into read status register mode */
	cfi_write(CMD(0x70), offset);

	DISABLE_VPP();

	/* put device back into read mode */
	cfi_write(CMD(0xFF), offset);
	kick_watchdog();
	return ret;
}

static int dumper_proc_read(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len = 0, i;
	char *out = page;
	char buffer[8], *buf = buffer;
	uint32_t dataval;

	out += sprintf(out, "Crash dump logger\n");
	out += sprintf(out, "Flash partition base addr:       0x%p [0x%p phys]\n", 
			dumper_priv.base, (void*)virt_to_phys(dumper_priv.base));
	out += sprintf(out, "Flash partition size:            0x%08lx\n", 
			dumper_priv.size);
	out += sprintf(out, "Flash sector size:               0x%08x\n", 
			dumper_priv.sector_size);
	out += sprintf(out, "Flash width:                     %d bytes\n", 
			dumper_priv.width);
	out += sprintf(out, "CPLD address:                    0x%p\n",
			dumper_priv.cpld_addr);
	out += sprintf(out, "CPLD VPP register offset:        0x%08hx\n",
			dumper_priv.vpp_offset);
	out += sprintf(out, "CPLD VPP register mask:          0x%08hx\n",
			dumper_priv.vpp_mask);
	out += sprintf(out, "Timebase ticks/jiffy:            %u\n",
			dumper_priv.ticks_jiffy);

	cfi_read(0, &dataval);
	memcpy(buf, &dataval, sizeof(dataval));
	buf += sizeof(dataval);
	cfi_read(sizeof(dataval), &dataval);
	memcpy(buf, &dataval, sizeof(dataval));
	buf += sizeof(dataval);

	out += sprintf(out, "Flash contents:                  [");
	for(i=0; i<8; i++)
		out += sprintf(out, "%c", buffer[i]);
	out += sprintf(out, "]\n");

	len = out - page;
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	return len;
}

static int dumper_proc_write(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	printk("Writing %ld bytes to flash\n", count);

	flash_dump_open();
	flash_dump_write((void*)buffer, count);
	flash_dump_close();
	return count;
}

static int dumper_proc_read2(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len = 0, i;
	char *out = page;
	char *ptr;
	unsigned long starttime, size;

	size = 1024*1024;
	if((ptr = vmalloc(size)) == NULL)
	{
		printk("Couldn't allocate buffer\n");
		return 0;
	}
	for(i=0; i<size; i++)
		ptr[i] = (i % 0x4A) + 0x30;	/* numbers and letters */

	starttime = jiffies;
	flash_dump_open();
	flash_dump_write(ptr, size);
	flash_dump_close();
	out += sprintf(out, "Completed %ld bytes in %lu jiffies.\n", size, (jiffies-starttime));

	starttime = jiffies;
	flash_dump_open();
	flash_dump_write(ptr, (size/64) - 5);
	flash_dump_close();
	out += sprintf(out, "Completed %ld bytes in %lu jiffies.\n", (size/64)-5, (jiffies-starttime));

	vfree(ptr);

	len = out - page;
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	return len;
}


/* dumper_init() - call this early in the boot process to 
 * initialize the dumper code.
 */
int flash_dump_init(void *base, unsigned long size, int width, int sector_size, 
		char *mtdname, uint8_t *cpld_addr, uint8_t vpp_offset, uint8_t vpp_mask )
{
	if(sector_size % 4096)
	{
		printk(KERN_ERR "dumper_init: bad sector size [0x%08x]\n", 
				sector_size);
		return -1;
	}
	if(width != 2 && width != 4)
	{
		printk(KERN_ERR "dumper_init: bad flash width [%d]\n", 
				width);
		return -1;
	}
	if(!tb_ticks_per_jiffy)
	{
		printk(KERN_ERR "dumper_init: timebase frequency not initialized\n");
		return -1;
	}

	memset(&dumper_priv, 0, sizeof(dumper_priv));
	dumper_priv.base = base;
	dumper_priv.size = size;
	dumper_priv.width = width;
	dumper_priv.sector_size = sector_size;
	dumper_priv.cpld_addr = cpld_addr;
	dumper_priv.vpp_offset = vpp_offset;
	dumper_priv.vpp_mask = vpp_mask;
	dumper_priv.ticks_jiffy = tb_ticks_per_jiffy;

	/* for testing only */
	if (mtdname) {
		dumper_priv.proc_entry = 
			create_proc_entry("flashdumper", S_IFREG|S_IRUGO, NULL);
		dumper_priv.proc_entry->read_proc = dumper_proc_read;
		dumper_priv.proc_entry->write_proc = dumper_proc_write;
		dumper_priv.proc_entry2 = 
			create_proc_entry("flashwrite", S_IFREG|S_IRUGO, NULL);
		dumper_priv.proc_entry2->read_proc = dumper_proc_read2;
	}

	if (mtdname) {
		printk(KERN_INFO "Silkworm crash dump logger using %s\n", mtdname);
	}
	return 0;
}

/* flash_dump_open() - Only one user can open this dump device at a time.
 * This simple open will just rewind to the start of the flash partition.
 */
int flash_dump_open(void)
{
	if(dumper_is_open)
	{
		printk(KERN_ERR "flash dump already opened\n");
		return -1;
	}
	dumper_is_open = 1;
	dumper_fptr = 0;
	return 0;
}

/* flash_dump_close() - This simple close will just close the device.
 */
int flash_dump_close(void)
{
	dumper_is_open = 0;
	dumper_fptr = 0;
	return 0;
}


#if 1

/* write_block() - write 16-word blocks to flash.  
 * This uses the write_buffer method in the strataflash which should give
 * better performance.
 *
 * Requirements: dumper_fptr must be 16-word aligned.  data ptr must be word aligned.
 */
static int write_block(unsigned long offset, uint32_t *data)
{
	int ret = 0;
	
	if(offset % (16*sizeof(uint32_t)))
	{	
		DPRINTK("calling write_block() with non-aligned flash pointer\n");
		return -1;
	}
	DPRINTK("write_block() at word offset 0x%08x = 0x%08x\n", offset, data[0]);

	/*
	 * If flash data path is 16-bit wide, we can only write 16 
	 * half-words at a time (8 words). Thus, we will need to call 
	 * flash_write_buffer() twice to write the 16-word block. 
	 */ 
	if (dumper_priv.width == 2) {
		ret = flash_write_buffer(offset, data, 8);
		if (ret == 0) {
			return flash_write_buffer(offset+32, &data[8], 8);
		} else {
			return (ret);
		}
	} else {
		return flash_write_buffer(offset, data, 16);
	}
}

/* write_word() - write a 32 bit word to flash.  According to the intel datasheet, a word write
 * can take as long as 630usec.  The fastest watchdog frequency is 167msec.  This means we
 * should refresh the watchdog at least every 250 words.
 *
 * Requirements: dumper_fptr must be word aligned.
 */
static int write_word(unsigned long offset, uint32_t data)
{
	int status;
	if(offset % sizeof(uint32_t))
	{	
		DPRINTK("calling write_word() with non-aligned flash pointer\n");
		return -1;
	}
	DPRINTK("write_word() at word offset 0x%08x = 0x%08x\n", offset, data);

	/*
	 * If flash data path is 16 bit wide, we can only 
	 * write 1 half-word at a time. Thus, we will need to call 
	 * flash_write_hword() twice to write 1 word. 
	 */ 
	if (dumper_priv.width == 2) {
		status = flash_write_hword(offset, (data >> 16));
		if (status == 0) {
			return flash_write_hword(offset + 2, (data & 0xffff)); 
		} else {
			return (status);
		}
	}
	return flash_write_word(offset, data);	/* rewrite merged data */
}

/* write_byte() - write 8 bits to flash.  Merge with the existing data in the other bytes of
 * the word.
 *
 * Requirements: dumper_fptr and data ptr have no alignment requirements.
 */
static int write_byte(unsigned long offset, uint8_t data)
{
	uint32_t merge_data;
	uint16_t merge_data_short;
	uint8_t *byte;
	unsigned long flash_offset, remainder;

	if (dumper_priv.width == 2) {
		remainder = (offset % sizeof(uint16_t));
		flash_offset = offset - remainder;

		/* read existing word and merge new data into it */
		cfi_read(flash_offset, &merge_data_short);
		byte = (void*)&merge_data_short;
		byte[remainder] = data;

		DPRINTK("write_byte() byte %d of word %d = 0x%x [word = 0x%08x]\n", 
			offset - flash_offset, flash_offset/sizeof(uint16_t), data, merge_data_short);

		return flash_write_hword(flash_offset, merge_data_short);/* rewrite merged data */
	} else { 
		remainder = (offset % sizeof(uint32_t));
		flash_offset = offset - remainder;

		/* read existing word and merge new data into it */
		cfi_read(flash_offset, &merge_data);
		byte = (void*)&merge_data;
		byte[remainder] = data;

		DPRINTK("write_byte() byte %d of word %d = 0x%x [word = 0x%08x]\n", 
			offset - flash_offset, flash_offset/sizeof(uint32_t), data, merge_data);

		return flash_write_word(flash_offset, merge_data);	/* rewrite merged data */
	}
}

static int erase_if_necessary(void)
{
	int ret = 0;

	/* When crossing into a new sector, erase it first */
	if((dumper_fptr % dumper_priv.sector_size) == 0)
	{
		DPRINTK(KERN_INFO "Erasing sector at word offset 0x%08x\n",
			dumper_fptr/sizeof(uint32_t));
		ret = erase_sector(dumper_fptr);
	}
	return ret;
}

/* 
 * flash_dump_write() - will write 'size' bytes to the flash, erasing flash
 * and kicking the watchdog as necessary.  Our flash memory only handles
 * 32-bit aligned writes.  The flash data pointer and the input buffer
 * might NOT be 32-bit aligned.  The code handles it either way.
 * May not be called prior to calling flash_dump_init().
 *
 */
int flash_dump_write(char *buf, unsigned long size)
{
	uint8_t *byte;
	uint32_t *wptr;
	int blocks = 0;
	unsigned long flags;

	/* Prevent overwriting */
	if((dumper_fptr + size) > dumper_priv.size)
	{
		size = dumper_priv.size - dumper_fptr;
		printk("Attempt to write beyond end of partition. "
			"Truncating this write to %ld bytes\n", size);
	}

	local_irq_save(flags); local_irq_disable(); /* just to be sure */
	kick_watchdog();

	byte = (void*)buf;		/* data buffer pointer */

	/* First write any non-word-aligned bytes into flash.  */
	while(size > 0 && (dumper_fptr % sizeof(uint32_t)))
	{
		erase_if_necessary();
		write_byte(dumper_fptr++, *byte++);
		size--;
	}

	/* Now our flash pointer is word aligned.  Write complete words until we hit
	 * a 16-word aligned condition 
	 */
	wptr = (void*)byte;		/* data buffer pointer */

	while(size >= sizeof(uint32_t) && (dumper_fptr % (16*sizeof(uint32_t))))
	{
		int step = sizeof(uint32_t);

		erase_if_necessary();
		write_word(dumper_fptr, *wptr++);
		dumper_fptr += step;
		size -= step;
	}

	/* Now our flash pointer is 16-word aligned.  Write as many 16-word blocks as we can.
	 */
	while(size >= (16*sizeof(uint32_t)))
	{
		int step = 16 * sizeof(uint32_t);

		erase_if_necessary();
		write_block(dumper_fptr, wptr);
		wptr += 16;
		dumper_fptr += step;
		size -= step;

		/* kick the watchdog every 200 blocks.
		 * 200 * 654usec = 130msec.
		 * If flash data path is 16-bit wide, it
		 * requires two write-to-buffer calls to write
		 * 1 block. Thus, we should kick watchdog
		 * every 100 blocks.
		 */
		if (dumper_priv.width == 4) {
			if((++blocks % 200) == 0)
				kick_watchdog();
		} else {
			if((++blocks % 100) == 0)
				kick_watchdog();
		}
	}

	/* Write any remaining complete words */
	while(size >= sizeof(uint32_t))
	{
		int step = sizeof(uint32_t);

		erase_if_necessary();
		write_word(dumper_fptr, *wptr++);
		dumper_fptr += step;
		size -= step;
	}

	/* Write any remaining bytes */
	byte = (void*)wptr;		/* data pointer */
	while(size > 0)
	{
		erase_if_necessary();
		write_byte(dumper_fptr++, *byte++);
		size--;
	}
	kick_watchdog();

	local_irq_restore(flags);
	return 0;
}

#endif

EXPORT_SYMBOL(flash_dump_open);
EXPORT_SYMBOL(flash_dump_close);
EXPORT_SYMBOL(flash_dump_write);
