/*
 * $Id: mtdchar.c,v 1.66 2005/01/05 18:05:11 dwmw2 Exp $
 *
 * Character-device access to raw MTD devices.
 * Pure 2.4 version - compatibility cruft removed to mtdchar-compat.c
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/compatmac.h>
#include <linux/mtd/map.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>

static void mtd_notify_add(struct mtd_info* mtd)
{
	if (!mtd)
		return;

	devfs_mk_cdev(MKDEV(MTD_CHAR_MAJOR, mtd->index*2),
		      S_IFCHR | S_IRUGO | S_IWUGO, "mtd/%d", mtd->index);
		
	devfs_mk_cdev(MKDEV(MTD_CHAR_MAJOR, mtd->index*2+1),
		      S_IFCHR | S_IRUGO, "mtd/%dro", mtd->index);
}

static void mtd_notify_remove(struct mtd_info* mtd)
{
	if (!mtd)
		return;
	devfs_remove("mtd/%d", mtd->index);
	devfs_remove("mtd/%dro", mtd->index);
}

static struct mtd_notifier notifier = {
	.add	= mtd_notify_add,
	.remove	= mtd_notify_remove,
};

static inline void mtdchar_devfs_init(void)
{
	devfs_mk_dir("mtd");
	register_mtd_user(&notifier);
}

static inline void mtdchar_devfs_exit(void)
{
	unregister_mtd_user(&notifier);
	devfs_remove("mtd");
}
#else /* !DEVFS */
#define mtdchar_devfs_init() do { } while(0)
#define mtdchar_devfs_exit() do { } while(0)
#endif

#ifdef CONFIG_MTDCHAR_CACHE
/*
 * Cache stuff... (adapted from mtdblock.c)
 * 
 * To allow the userspace apps to treat mtd as just a regular file, we implement 
 * read-modify-write on flash sectors for each write request.  To avoid
 * over-erasing flash sectors and to speed things up, we locally cache a whole
 * flash sector while it is being written to until a different sector is required.
 * 
 * The cached data is written to flash under the following circumstances:
 * 1) write() is called with a multiple of the sector size
 * 2) another sector is written
 * 3) the device is closed 
 * 4) a MEMSYNC ioctl is sent to the device
 */

static struct mtd_cache {
	struct mtd_info *mtd; /* Locked */
	int count;
	struct semaphore cache_sem;
	unsigned char *cache_data;
	unsigned long cache_offset;
	unsigned int cache_size;
	enum { STATE_EMPTY, STATE_CLEAN, STATE_DIRTY } cache_state;
} *mtdcaches[MAX_MTD_DEVICES];

static spinlock_t mtdcaches_lock;

static void erase_callback(struct erase_info *done)
{
	wait_queue_head_t *wait_q = (wait_queue_head_t *)done->priv;
	wake_up(wait_q);
}

static int erase_write (struct mtd_info *mtd, unsigned long pos, 
			int len, size_t *retlen, const char *buf)
{
	struct erase_info erase;
	DECLARE_WAITQUEUE(wait, current);
	wait_queue_head_t wait_q;
	int ret;

	/*
	 * First, let's erase the flash block.
	 */

	init_waitqueue_head(&wait_q);
	erase.mtd = mtd;
	erase.callback = erase_callback;
	erase.addr = pos;
	erase.len = len;
	erase.priv = (u_long)&wait_q;

	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&wait_q, &wait);

	ret = MTD_ERASE(mtd, &erase);
	if (ret) {
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&wait_q, &wait);
		printk (KERN_WARNING "mtdchar: erase of region [0x%lx, 0x%x] "
				     "on \"%s\" failed\n",
			pos, len, mtd->name);
		return ret;
	}

	schedule();  /* Wait for erase to finish. */
	remove_wait_queue(&wait_q, &wait);

	/*
	 * Next, writhe data to flash.
	 */

	ret = MTD_WRITE (mtd, pos, len, retlen, buf);
	if (ret)
		return ret;
	if (*retlen != len)
		return -EIO;
	return 0;
}


static int write_cached_data (struct mtd_cache *mtdcache)
{
	struct mtd_info *mtd = mtdcache->mtd;
	int ret;
	size_t retlen;

	if (mtdcache->cache_state != STATE_DIRTY)
		return 0;

	DEBUG(MTD_DEBUG_LEVEL2, "mtdchar: writing cached data for \"%s\" "
			"at 0x%lx, size 0x%x\n", mtd->name, 
			mtdcache->cache_offset, mtdcache->cache_size);
	
	ret = erase_write (mtd, mtdcache->cache_offset, 
			   mtdcache->cache_size, &retlen, mtdcache->cache_data);
	if (ret)
		return ret;

	/*
	 * Here we could argably set the cache state to STATE_CLEAN.
	 * However this could lead to inconsistency since we will not 
	 * be notified if this content is altered on the flash by other 
	 * means.  Let's declare it empty and leave buffering tasks to
	 * the buffer cache instead.
	 */
	mtdcache->cache_state = STATE_EMPTY;
	return 0;
}


static int do_cached_write (struct mtd_cache *mtdcache, unsigned long pos, 
			    int len, size_t *total_retlen, const char *buf)
{
	struct mtd_info *mtd = mtdcache->mtd;
	unsigned int sect_size = mtdcache->cache_size;
	size_t retlen = 0;
	int ret;

	DEBUG(MTD_DEBUG_LEVEL2, "mtdchar: write on \"%s\" at 0x%lx, size 0x%x\n",
		mtd->name, pos, len);
	
	*total_retlen = 0;
	if (!sect_size)
		return MTD_WRITE (mtd, pos, len, total_retlen, buf);

	while (len > 0) {
		unsigned long sect_start = (pos/sect_size)*sect_size;
		unsigned int offset = pos - sect_start;
		unsigned int size = sect_size - offset;
		if( size > len ) 
			size = len;

		if (size == sect_size) {
			/* 
			 * We are covering a whole sector.  Thus there is no
			 * need to bother with the cache while it may still be
			 * useful for other partial writes.
			 */
			ret = erase_write (mtd, pos, size, &retlen, buf);
			*total_retlen += retlen;
			if (ret)
				return ret;
		} else {
			/* Partial sector: need to use the cache */

			if (mtdcache->cache_state == STATE_DIRTY &&
			    mtdcache->cache_offset != sect_start) {
				ret = write_cached_data(mtdcache);
				if (ret) 
					return ret;
			}

			if (mtdcache->cache_state == STATE_EMPTY ||
			    mtdcache->cache_offset != sect_start) {
				/* fill the cache with the current sector */
				mtdcache->cache_state = STATE_EMPTY;
				ret = MTD_READ(mtd, sect_start, sect_size, &retlen, mtdcache->cache_data);
				if (ret)
					return ret;
				if (retlen != sect_size)
					return -EIO;

				mtdcache->cache_offset = sect_start;
				mtdcache->cache_size = sect_size;
				mtdcache->cache_state = STATE_CLEAN;
			}

			/* write data to our local cache */
			memcpy (mtdcache->cache_data + offset, buf, size);
			mtdcache->cache_state = STATE_DIRTY;
			*total_retlen += size;
		}

		buf += size;
		pos += size;
		len -= size;
	}

	return 0;
}


static int do_cached_read (struct mtd_cache *mtdcache, unsigned long pos, 
			   int len, size_t *total_retlen, char *buf)
{
	struct mtd_info *mtd = mtdcache->mtd;
	unsigned int sect_size = mtdcache->cache_size;
	size_t retlen = 0;
	int ret;

	DEBUG(MTD_DEBUG_LEVEL2, "mtdchar: read on \"%s\" at 0x%lx, size 0x%x\n", 
			mtd->name, pos, len);

	*total_retlen = 0;
	if (!sect_size)
		return MTD_READ (mtd, pos, len, total_retlen, buf);

	while (len > 0) {
		unsigned long sect_start = (pos/sect_size)*sect_size;
		unsigned int offset = pos - sect_start;
		unsigned int size = sect_size - offset;
		if (size > len) 
			size = len;

		/*
		 * Check if the requested data is already cached
		 * Read the requested amount of data from our internal cache if it
		 * contains what we want, otherwise we read the data directly
		 * from flash.
		 */
		if (mtdcache->cache_state != STATE_EMPTY &&
		    mtdcache->cache_offset == sect_start) {
			memcpy (buf, mtdcache->cache_data + offset, size);
			*total_retlen += size;
		} else {
			ret = MTD_READ (mtd, pos, size, &retlen, buf);
			*total_retlen += retlen;
			if (ret)
				return ret;
			if (retlen != size)
				return -EIO;
		}

		buf += size;
		pos += size;
		len -= size;
	}

	return 0;
}
#endif /* CONFIG_MTDCHAR_CACHE */

static loff_t mtd_lseek (struct file *file, loff_t offset, int orig)
{
	struct mtd_info *mtd;
#ifdef CONFIG_MTDCHAR_CACHE
	struct mtd_cache *mtdcache = (struct mtd_cache *)file->private_data;
	mtd = mtdcache->mtd;
#else
	mtd = (struct mtd_info *)file->private_data;
#endif

	switch (orig) {
	case 0:
		/* SEEK_SET */
		file->f_pos = offset;
		break;
	case 1:
		/* SEEK_CUR */
		file->f_pos += offset;
		break;
	case 2:
		/* SEEK_END */
		file->f_pos =mtd->size + offset;
		break;
	default:
		return -EINVAL;
	}

	if (file->f_pos < 0)
		file->f_pos = 0;
	else if (file->f_pos >= mtd->size)
		file->f_pos = mtd->size - 1;

	return file->f_pos;
}



static int mtd_open(struct inode *inode, struct file *file)
{
	int minor = iminor(inode);
	int devnum = minor >> 1;
	struct mtd_info *mtd;
#ifdef CONFIG_MTDCHAR_CACHE
	struct mtd_cache *mtdcache;
#endif

	DEBUG(MTD_DEBUG_LEVEL0, "MTD_open\n");

	if (devnum >= MAX_MTD_DEVICES)
		return -ENODEV;

	/* You can't open the RO devices RW */
	if ((file->f_mode & 2) && (minor & 1))
		return -EACCES;

	mtd = get_mtd_device(NULL, devnum);
	
	if (!mtd)
		return -ENODEV;
	
	if (MTD_ABSENT == mtd->type) {
		put_mtd_device(mtd);
		return -ENODEV;
	}

#ifdef CONFIG_MTDCHAR_CACHE
	/* Create a cache for this device */
	if((mtdcache = kmalloc(sizeof(struct mtd_cache), GFP_KERNEL)) == NULL) {
		put_mtd_device(mtd);
		return -ENOMEM;
	}
	memset(mtdcache, 0, sizeof(struct mtd_cache));
	mtdcache->count = 1;
	mtdcache->mtd = mtd;
	init_MUTEX(&mtdcache->cache_sem);
	mtdcache->cache_state = STATE_EMPTY;
	mtdcache->cache_size = mtd->erasesize;
	mtdcache->cache_data = vmalloc(mtd->erasesize);
	if(!mtdcache->cache_data) {
		put_mtd_device(mtd);
		kfree(mtdcache);
		return -ENOMEM;
	}

	/* OK, we've created a new one. Add it to the list. */
	spin_lock(&mtdcaches_lock);
	if(mtdcaches[devnum]) {
		/* Another CPU made one at the same time as us. */
		mtdcaches[devnum]->count++;
		spin_unlock(&mtdcaches_lock);
		file->private_data = mtdcaches[devnum];
		put_mtd_device(mtd);	/* is this right?? */
		vfree(mtdcache->cache_data);
		kfree(mtdcache);
		return 0;
	}

	mtdcaches[devnum] = mtdcache;
	file->private_data = mtdcaches[devnum];
#else
	file->private_data = mtd;
#endif
		
	/* You can't open it RW if it's not a writeable device */
	if ((file->f_mode & 2) && !(mtd->flags & MTD_WRITEABLE)) {
		put_mtd_device(mtd);
		return -EACCES;
	}
		
	return 0;
} /* mtd_open */

/*====================================================================*/

static int mtd_close(struct inode *inode, struct file *file)
{
	struct mtd_info *mtd;

	DEBUG(MTD_DEBUG_LEVEL0, "MTD_close\n");

#ifdef CONFIG_MTDCHAR_CACHE
	{
		int devnum = MINOR(inode->i_rdev) >> 1;
		struct mtd_cache *mtdcache;

		mtdcache = mtdcaches[devnum];
		down(&mtdcache->cache_sem);
		write_cached_data(mtdcache);
		up(&mtdcache->cache_sem);
	
		spin_lock(&mtdcaches_lock);
		if(!--mtdcache->count) {
			/* It was the last usage. Free the device */
			mtdcaches[devnum] = NULL;
			spin_unlock(&mtdcaches_lock);
			mtd = mtdcache->mtd;
			if (mtd->sync)
				mtd->sync(mtd);
			put_mtd_device(mtd);
			vfree(mtdcache->cache_data);
			kfree(mtdcache);
		}
		else {
			spin_unlock(&mtdcaches_lock);
		}
	}
#else
	mtd = (struct mtd_info *)file->private_data;
	
	if (mtd->sync)
		mtd->sync(mtd);
	
	put_mtd_device(mtd);
#endif

	return 0;
} /* mtd_close */

/* FIXME: This _really_ needs to die. In 2.5, we should lock the
   userspace buffer down and use it directly with readv/writev.
*/
#define MAX_KMALLOC_SIZE 0x20000

static ssize_t mtd_read(struct file *file, char __user *buf, size_t count,loff_t *ppos)
{
	struct mtd_info *mtd;
	size_t retlen=0;
	size_t total_retlen=0;
	int ret=0;
	int len;
	char *kbuf;
	
#ifdef CONFIG_MTDCHAR_CACHE
	struct mtd_cache *mtdcache = (struct mtd_cache *)file->private_data;
	mtd = mtdcache->mtd;
#else
	mtd = (struct mtd_info *)file->private_data;
#endif

	DEBUG(MTD_DEBUG_LEVEL0,"MTD_read\n");

	if (*ppos + count > mtd->size)
		count = mtd->size - *ppos;

	if (!count)
		return 0;
	
	/* FIXME: Use kiovec in 2.5 to lock down the user's buffers
	   and pass them directly to the MTD functions */
	while (count) {
		if (count > MAX_KMALLOC_SIZE) 
			len = MAX_KMALLOC_SIZE;
		else
			len = count;

		kbuf=kmalloc(len,GFP_KERNEL);
		if (!kbuf)
			return -ENOMEM;

#ifdef CONFIG_MTDCHAR_CACHE
		down(&mtdcache->cache_sem);
		ret = do_cached_read(mtdcache, *ppos, len, &retlen, kbuf);
		up(&mtdcache->cache_sem);
#else
		ret = MTD_READ(mtd, *ppos, len, &retlen, kbuf);
#endif
		/* Nand returns -EBADMSG on ecc errors, but it returns
		 * the data. For our userspace tools it is important
		 * to dump areas with ecc errors ! 
		 * Userspace software which accesses NAND this way
		 * must be aware of the fact that it deals with NAND
		 */
		if (!ret || (ret == -EBADMSG)) {
			*ppos += retlen;
			if (copy_to_user(buf, kbuf, retlen)) {
			        kfree(kbuf);
				return -EFAULT;
			}
			else
				total_retlen += retlen;

			count -= retlen;
			buf += retlen;
		}
		else {
			kfree(kbuf);
			return ret;
		}
		
		kfree(kbuf);
	}
	
	return total_retlen;
} /* mtd_read */

static ssize_t mtd_write(struct file *file, const char __user *buf, size_t count,loff_t *ppos)
{
	struct mtd_info *mtd;
	char *kbuf;
	size_t retlen;
	size_t total_retlen=0;
	int ret=0;
	int len;

#ifdef CONFIG_MTDCHAR_CACHE
	struct mtd_cache *mtdcache = (struct mtd_cache *)file->private_data;
	mtd = mtdcache->mtd;
#else
	mtd = (struct mtd_info *)file->private_data;
#endif

	DEBUG(MTD_DEBUG_LEVEL0,"MTD_write\n");
	
	if (*ppos == mtd->size)
		return -ENOSPC;
	
	if (*ppos + count > mtd->size)
		count = mtd->size - *ppos;

	if (!count)
		return 0;

	while (count) {
		if (count > MAX_KMALLOC_SIZE) 
			len = MAX_KMALLOC_SIZE;
		else
			len = count;

		kbuf=kmalloc(len,GFP_KERNEL);
		if (!kbuf) {
			printk("kmalloc is null\n");
			return -ENOMEM;
		}

		if (copy_from_user(kbuf, buf, len)) {
			kfree(kbuf);
			return -EFAULT;
		}
		
#ifdef CONFIG_MTDCHAR_CACHE
		down(&mtdcache->cache_sem);
		ret = do_cached_write(mtdcache, *ppos, len, &retlen, kbuf);
		up(&mtdcache->cache_sem);
#else
	        ret = (*(mtd->write))(mtd, *ppos, len, &retlen, kbuf);
#endif
		if (!ret) {
			*ppos += retlen;
			total_retlen += retlen;
			count -= retlen;
			buf += retlen;
		}
		else {
			kfree(kbuf);
			return ret;
		}
		
		kfree(kbuf);
	}

	return total_retlen;
} /* mtd_write */

/*======================================================================

    IOCTL calls for getting device parameters.

======================================================================*/
static void mtdchar_erase_callback (struct erase_info *instr)
{
	wake_up((wait_queue_head_t *)instr->priv);
}

static int mtd_ioctl(struct inode *inode, struct file *file,
		     u_int cmd, u_long arg)
{
	struct mtd_info *mtd;
	void __user *argp = (void __user *)arg;
	int ret = 0;
	u_long size;
#ifdef CONFIG_MTDCHAR_CACHE
	struct mtd_cache *mtdcache = (struct mtd_cache *)file->private_data;
	mtd = mtdcache->mtd;
#else
	mtd = (struct mtd_info *)file->private_data;
#endif
	
	DEBUG(MTD_DEBUG_LEVEL0, "MTD_ioctl\n");

	size = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
	if (cmd & IOC_IN) {
		ret = !access_ok(VERIFY_READ, (char *)argp, size);
		if (ret) return ret;
	}
	if (cmd & IOC_OUT) {
		ret = !access_ok(VERIFY_WRITE, (char *)argp, size);
		if (ret) return ret;
	}
	
	switch (cmd) {
	case MEMGETREGIONCOUNT:
		if (copy_to_user((int *) argp, &(mtd->numeraseregions), sizeof(int)))
			return -EFAULT;
		break;

	case MEMGETREGIONINFO:
	{
		struct region_info_user ur;

		if (copy_from_user(	&ur, 
					(struct region_info_user *)argp, 
					sizeof(struct region_info_user))) {
			return -EFAULT;
		}

		if (ur.regionindex >= mtd->numeraseregions)
			return -EINVAL;
		if (copy_to_user((struct mtd_erase_region_info *) argp, 
				&(mtd->eraseregions[ur.regionindex]),
				sizeof(struct mtd_erase_region_info)))
			return -EFAULT;
		break;
	}

	case MEMGETINFO:
		if (copy_to_user((struct mtd_info *)argp, mtd,
				 sizeof(struct mtd_info_user)))
			return -EFAULT;
		break;

	case MEMERASE:
	{
		struct erase_info *erase;

		if(!(file->f_mode & 2))
			return -EPERM;

		erase=kmalloc(sizeof(struct erase_info),GFP_KERNEL);
		if (!erase)
			ret = -ENOMEM;
		else {
			wait_queue_head_t waitq;
			DECLARE_WAITQUEUE(wait, current);

			init_waitqueue_head(&waitq);

			memset (erase,0,sizeof(struct erase_info));
			if (copy_from_user(&erase->addr, (u_long *)argp,
					   2 * sizeof(u_long))) {
				kfree(erase);
				return -EFAULT;
			}
			erase->mtd = mtd;
			erase->callback = mtdchar_erase_callback;
			erase->priv = (unsigned long)&waitq;
			
			/*
			  FIXME: Allow INTERRUPTIBLE. Which means
			  not having the wait_queue head on the stack.
			  
			  If the wq_head is on the stack, and we
			  leave because we got interrupted, then the
			  wq_head is no longer there when the
			  callback routine tries to wake us up.
			*/
			ret = mtd->erase(mtd, erase);
			if (!ret) {
				set_current_state(TASK_UNINTERRUPTIBLE);
				add_wait_queue(&waitq, &wait);
				if (erase->state != MTD_ERASE_DONE &&
				    erase->state != MTD_ERASE_FAILED)
					schedule();
				remove_wait_queue(&waitq, &wait);
				set_current_state(TASK_RUNNING);

				ret = (erase->state == MTD_ERASE_FAILED)?-EIO:0;
			}
			kfree(erase);
		}
		break;
	}

	case MEMWRITEOOB:
	{
		struct mtd_oob_buf buf;
		void *databuf;
		ssize_t retlen;

		if(!(file->f_mode & 2))
			return -EPERM;

		if (copy_from_user(&buf, argp, sizeof(struct mtd_oob_buf)))
			return -EFAULT;
		
		if (buf.length > 0x4096)
			return -EINVAL;

		if (!mtd->write_oob)
			ret = -EOPNOTSUPP;
		else
			ret = !access_ok(VERIFY_READ, buf.ptr, buf.length);

		if (ret)
			return ret;

		databuf = kmalloc(buf.length, GFP_KERNEL);
		if (!databuf)
			return -ENOMEM;
		
		if (copy_from_user(databuf, buf.ptr, buf.length)) {
			kfree(databuf);
			return -EFAULT;
		}

		ret = (mtd->write_oob)(mtd, buf.start, buf.length, &retlen, databuf);

		if (copy_to_user((void *)argp + sizeof(u_int32_t), &retlen, sizeof(u_int32_t)))
			ret = -EFAULT;

		kfree(databuf);
		break;

	}

	case MEMREADOOB:
	{
		struct mtd_oob_buf buf;
		void *databuf;
		ssize_t retlen;

		if (copy_from_user(&buf, argp, sizeof(struct mtd_oob_buf)))
			return -EFAULT;
		
		if (buf.length > 0x4096)
			return -EINVAL;

		if (!mtd->read_oob)
			ret = -EOPNOTSUPP;
		else
			ret = !access_ok(VERIFY_WRITE, buf.ptr, buf.length);

		if (ret)
			return ret;

		databuf = kmalloc(buf.length, GFP_KERNEL);
		if (!databuf)
			return -ENOMEM;
		
		ret = (mtd->read_oob)(mtd, buf.start, buf.length, &retlen, databuf);

		if (put_user(retlen, (uint32_t __user *)argp))
			ret = -EFAULT;
		else if (retlen && copy_to_user(buf.ptr, databuf, retlen))
			ret = -EFAULT;
		
		kfree(databuf);
		break;
	}

	case MEMLOCK:
	{
		struct erase_info_user info;

		if (copy_from_user(&info, argp, sizeof(info)))
			return -EFAULT;

		if (!mtd->lock)
			ret = -EOPNOTSUPP;
		else
			ret = mtd->unlock(mtd, info.start, info.length);
		break;
	}

	case MEMUNLOCK:
	{
		unsigned long adrs[2];

		if (copy_from_user(adrs, (void *)argp, 2* sizeof(unsigned long)))
			return -EFAULT;

		if (!mtd->unlock)
			ret = -EOPNOTSUPP;
		else
			ret = mtd->unlock(mtd, adrs[0], adrs[1]);
		break;
	}

	case MEMSETOOBSEL:
	{
		if (copy_from_user(&mtd->oobinfo, argp, sizeof(struct nand_oobinfo)))
			return -EFAULT;
		break;
	}

	case MEMGETOOBSEL:
	{
		if (copy_to_user(argp, &(mtd->oobinfo), sizeof(struct nand_oobinfo)))
			return -EFAULT;
		break;
	}

	case MEMGETBADBLOCK:
	{
		loff_t offs;
		
		if (copy_from_user(&offs, argp, sizeof(loff_t)))
			return -EFAULT;
		if (!mtd->block_isbad)
			ret = -EOPNOTSUPP;
		else
			return mtd->block_isbad(mtd, offs);
		break;
	}

	case MEMSETBADBLOCK:
	{
		loff_t offs;

		if (copy_from_user(&offs, argp, sizeof(loff_t)))
			return -EFAULT;
		if (!mtd->block_markbad)
			ret = -EOPNOTSUPP;
		else
			return mtd->block_markbad(mtd, offs);
		break;
	}

#ifdef CONFIG_MTDCHAR_CACHE
	case MEMSYNC:
	{
		down(&mtdcache->cache_sem);
		ret = write_cached_data(mtdcache);
		up(&mtdcache->cache_sem);
		if(mtdcache->mtd->sync)
			mtdcache->mtd->sync(mtdcache->mtd);
		break;
	}
#endif
		
	case MEMGETPHYS:
	{
		struct map_info *map = (struct map_info*) mtd->priv;
		struct mtd_phys_info	user_info;
#ifdef CONFIG_MTD_PARTITIONS
		struct mtd_part *part = (struct mtd_part *)mtd;
#endif

		/* Since this was an ioremap()'ed address, must use iopa instead of virt_to_phys */
		user_info.phys = iopa((void*)map->map_priv_1);

#ifdef CONFIG_MTD_PARTITIONS
		/* We use partitioned MTD devices, this is a hack */
		user_info.phys += part->offset;
#endif

		if (copy_to_user((struct mtd_phys_info *)argp, &user_info,
				 sizeof(struct mtd_phys_info)))
			return -EFAULT;
		break;
	}

	default:
		DEBUG(MTD_DEBUG_LEVEL0, "Invalid ioctl %x (MEMGETINFO = %x)\n", cmd, MEMGETINFO);
		ret = -ENOTTY;
	}

	return ret;
} /* memory_ioctl */

static struct file_operations mtd_fops = {
	.owner		= THIS_MODULE,
	.llseek		= mtd_lseek,
	.read		= mtd_read,
	.write		= mtd_write,
	.ioctl		= mtd_ioctl,
	.open		= mtd_open,
	.release	= mtd_close,
};


static int __init init_mtdchar(void)
{
	if (register_chrdev(MTD_CHAR_MAJOR, "mtd", &mtd_fops)) {
		printk(KERN_NOTICE "Can't allocate major number %d for Memory Technology Devices.\n",
		       MTD_CHAR_MAJOR);
		return -EAGAIN;
	}

	mtdchar_devfs_init();

#ifdef CONFIG_MTDCHAR_CACHE
	printk(KERN_INFO "mtdchar: write-caching enabled\n");
#endif
	return 0;
}

static void __exit cleanup_mtdchar(void)
{
	mtdchar_devfs_exit();
}

module_init(init_mtdchar);
module_exit(cleanup_mtdchar);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Direct character-device access to MTD devices");
