/* 
 * drivers/net/gianfar_phy.c
 *
 * Gianfar Ethernet Driver -- PHY handling
 * Driver for FEC on MPC8540 and TSEC on MPC8540/MPC8560
 * Based on 8260_io/fcc_enet.c
 *
 * Author: Andy Fleming
 * Maintainer: Kumar Gala (kumar.gala@freescale.com)
 *
 * Copyright (c) 2002-2004 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/crc32.h>
#include <linux/mii.h>

#include "gianfar.h"
#include "gianfar_phy.h"
#include "sungem_phy.h"
#include "sungem.h"

static void config_genmii_advert(struct gfar_mii_info *mii_info);
static void genmii_setup_forced(struct gfar_mii_info *mii_info);
static void genmii_restart_aneg(struct gfar_mii_info *mii_info);
static int gbit_config_aneg(struct gfar_mii_info *mii_info);
static int genmii_config_aneg(struct gfar_mii_info *mii_info);
static int genmii_update_link(struct gfar_mii_info *mii_info);
static int genmii_read_status(struct gfar_mii_info *mii_info);
u16 phy_read(struct gfar_mii_info *mii_info, u16 regnum);
void phy_write(struct gfar_mii_info *mii_info, u16 regnum, u16 val);

static int bcm54xx_read_link(struct gfar_mii_info *phy);

static int reset_mii_phy(struct gfar_mii_info* phy);

/* Link modes of the BCM5400 PHY */
static const int phy_BCM5400_link_table[8][3] = {
	{ 0, 0, 0 },	/* No link */
	{ 0, 0, 0 },	/* 10BT Half Duplex */
	{ 1, 0, 0 },	/* 10BT Full Duplex */
	{ 0, 1, 0 },	/* 100BT Half Duplex */
	{ 0, 1, 0 },	/* 100BT Half Duplex */
	{ 1, 1, 0 },	/* 100BT Full Duplex*/
	{ 1, 0, 1 },	/* 1000BT */
	{ 1, 0, 1 },	/* 1000BT */
};

/* Write value to the PHY for this device to the register at regnum, */
/* waiting until the write is done before it returns.  All PHY */
/* configuration has to be done through the TSEC1 MIIM regs */
void write_phy_reg(struct net_device *dev, int mii_id, int regnum, int value)
{
	struct gfar_private *priv = netdev_priv(dev);
	struct gfar *regbase = priv->phyregs;
	u32 regval;

	while ((regval = gfar_read(&regbase->miimind)) &
			(MIIMIND_BUSY | MIIMIND_SCAN | MIIMIND_NOTVALID)) {
		if (regval & MIIMIND_SCAN) {
			printk(KERN_ERR "write_phy_reg():MIIMIND_SCAN set 0x%x 0x%x 0x%x\n",
				regval, regnum, value);
			return;
		}
		cpu_relax();
	}

	if (regnum == 4) {
		value |= 1;
	}

	/* Set the PHY address and the register address we want to write */
	gfar_write(&regbase->miimadd, (mii_id << 8) | regnum);

	/* Write out the value we want */
	gfar_write(&regbase->miimcon, value);

	/* Wait for the transaction to finish */
	while (gfar_read(&regbase->miimind) & MIIMIND_BUSY)
		cpu_relax();
}

/* Reads from register regnum in the PHY for device dev, */
/* returning the value.  Clears miimcom first.  All PHY */
/* configuration has to be done through the TSEC1 MIIM regs */
int read_phy_reg(struct net_device *dev, int mii_id, int regnum)
{
	struct gfar_private *priv = netdev_priv(dev);
	struct gfar *regbase = priv->phyregs;
	u16 value;
	u32 regval;

	while ((regval = gfar_read(&regbase->miimind)) &
			(MIIMIND_BUSY | MIIMIND_SCAN | MIIMIND_NOTVALID)) {
		if (regval & MIIMIND_SCAN) {
			printk(KERN_ERR "read_phy_reg():MIIMIND_SCAN set 0x%x 0x%x\n",
				regval, regnum);
			return 0;
		}
		cpu_relax();
	}

	/* Set the PHY address and the register address we want to read */
	gfar_write(&regbase->miimadd, (mii_id << 8) | regnum);

	/* Clear miimcom, and then initiate a read */
	gfar_write(&regbase->miimcom, 0);
	gfar_write(&regbase->miimcom, MII_READ_COMMAND);

	/* Wait for the transaction to finish */
	while (gfar_read(&regbase->miimind) & (MIIMIND_NOTVALID | MIIMIND_BUSY))
		cpu_relax();

	/* Grab the value of the register from miimstat */
	value = gfar_read(&regbase->miimstat);

#if 0
	if (regnum == 4) {
		printk(KERN_ERR "read_phy_reg(0x%x, 0x%x, 0x%x)\n",
			mii_id, regnum, value);
	}
#endif
	return value;
}

void mii_clear_phy_interrupt(struct gfar_mii_info *mii_info)
{
	if(mii_info->phyinfo->ack_interrupt)
		mii_info->phyinfo->ack_interrupt(mii_info);
}


void mii_configure_phy_interrupt(struct gfar_mii_info *mii_info, u32 interrupts)
{
	mii_info->interrupts = interrupts;
	if(mii_info->phyinfo->config_intr)
		mii_info->phyinfo->config_intr(mii_info);
}


/* Writes MII_ADVERTISE with the appropriate values, after
 * sanitizing advertise to make sure only supported features
 * are advertised 
 */
static void config_genmii_advert(struct gfar_mii_info *mii_info)
{
	u32 advertise;
	u16 adv;

	/* Only allow advertising what this PHY supports */
	mii_info->advertising &= mii_info->phyinfo->features;
	advertise = mii_info->advertising;

	/* Setup standard advertisement */
	adv = phy_read(mii_info, MII_ADVERTISE);
	adv &= ~(ADVERTISE_ALL | ADVERTISE_100BASE4);
	if (advertise & ADVERTISED_10baseT_Half)
		adv |= ADVERTISE_10HALF;
	if (advertise & ADVERTISED_10baseT_Full)
		adv |= ADVERTISE_10FULL;
	if (advertise & ADVERTISED_100baseT_Half)
		adv |= ADVERTISE_100HALF;
	if (advertise & ADVERTISED_100baseT_Full)
		adv |= ADVERTISE_100FULL;
	phy_write(mii_info, MII_ADVERTISE, adv);
}

static void genmii_setup_forced(struct gfar_mii_info *mii_info)
{
	u16 ctrl;
	u32 features = mii_info->phyinfo->features;
	
	ctrl = phy_read(mii_info, MII_BMCR);

	ctrl &= ~(BMCR_FULLDPLX|BMCR_SPEED100|BMCR_SPEED1000|BMCR_ANENABLE);

	/* First reset the PHY */
	phy_write(mii_info, MII_BMCR, ctrl | BMCR_RESET);

	switch(mii_info->speed) {
		case SPEED_1000:
			if(features & (SUPPORTED_1000baseT_Half
						| SUPPORTED_1000baseT_Full)) {
				ctrl |= BMCR_SPEED1000;
				break;
			}
			mii_info->speed = SPEED_100;
		case SPEED_100:
			if (features & (SUPPORTED_100baseT_Half
						| SUPPORTED_100baseT_Full)) {
				ctrl |= BMCR_SPEED100;
				break;
			}
			mii_info->speed = SPEED_10;
		case SPEED_10:
			if (features & (SUPPORTED_10baseT_Half
						| SUPPORTED_10baseT_Full))
				break;
		default: /* Unsupported speed! */
			printk(KERN_ERR "%s: Bad speed!\n", 
					mii_info->dev->name);
			break;
	}

	if (mii_info->duplex == DUPLEX_FULL)
		ctrl |= BMCR_FULLDPLX;

	phy_write(mii_info, MII_BMCR, ctrl);
}


/* Enable and Restart Autonegotiation */
static void genmii_restart_aneg(struct gfar_mii_info *mii_info)
{
	u16 ctl;

	ctl = phy_read(mii_info, MII_BMCR);
	ctl |= (BMCR_ANENABLE | BMCR_ANRESTART);
	phy_write(mii_info, MII_BMCR, ctl);
}


static int gbit_config_aneg(struct gfar_mii_info *mii_info)
{
	u16 adv;
	u32 advertise;

	/* don't try to configure the PHY if there is none */
	if (mii_info->mii_id == FSL_GIANFAR_BRD_HAS_NO_PHY)
		return 0;

	if(mii_info->autoneg) {
		/* Configure the ADVERTISE register */
		config_genmii_advert(mii_info);
		advertise = mii_info->advertising;

		adv = phy_read(mii_info, MII_1000BASETCONTROL);
		adv &= ~(MII_1000BASETCONTROL_FULLDUPLEXCAP |
				MII_1000BASETCONTROL_HALFDUPLEXCAP);
		if (advertise & SUPPORTED_1000baseT_Half)
			adv |= MII_1000BASETCONTROL_HALFDUPLEXCAP;
		if (advertise & SUPPORTED_1000baseT_Full)
			adv |= MII_1000BASETCONTROL_FULLDUPLEXCAP;
		phy_write(mii_info, MII_1000BASETCONTROL, adv);

		/* Start/Restart aneg */
		genmii_restart_aneg(mii_info);
	} else
		genmii_setup_forced(mii_info);

	return 0;
}

static int marvell_config_aneg(struct gfar_mii_info *mii_info)
{
	/* The Marvell PHY has an errata which requires
	 * that certain registers get written in order
	 * to restart autonegotiation */
	phy_write(mii_info, MII_BMCR, BMCR_RESET);

	phy_write(mii_info, 0x1d, 0x1f);
	phy_write(mii_info, 0x1e, 0x200c);
	phy_write(mii_info, 0x1d, 0x5);
	phy_write(mii_info, 0x1e, 0);
	phy_write(mii_info, 0x1e, 0x100);

	gbit_config_aneg(mii_info);

	return 0;
}
static int genmii_config_aneg(struct gfar_mii_info *mii_info)
{
	if (mii_info->autoneg) {
		config_genmii_advert(mii_info);
		genmii_restart_aneg(mii_info);
	} else
		genmii_setup_forced(mii_info);

	return 0;
}


static int genmii_update_link(struct gfar_mii_info *mii_info)
{
	u16 status;

	/* Do a fake read */
	phy_read(mii_info, MII_BMSR);

	/* Read link and autonegotiation status */
	status = phy_read(mii_info, MII_BMSR);
	if ((status & BMSR_LSTATUS) == 0)
		mii_info->link = 0;
	else
		mii_info->link = 1;

	/* If we are autonegotiating, and not done, 
	 * return an error */
	if (mii_info->autoneg && !(status & BMSR_ANEGCOMPLETE))
		return -EAGAIN;

	return 0;
}

static int genmii_read_status(struct gfar_mii_info *mii_info)
{
	int err;

	/* Update the link, but return if there
	 * was an error */
	err = genmii_update_link(mii_info);
	if (err)
		return err;

	/*
	 * defect 248525, if you are going to do the trick of determining
	 * what your current speed and duplex is from the standard PHY
	 * registers, then you have to take into account not just what
	 * the partner is advertising but also what you are. and then
	 * you have to select the highest common mode where modes are
	 * are ordered 100FD, 100HD, 10FD and 10HD.
	 */

	if (mii_info->autoneg) {
		u16 peer_status;
		u16 local_status;
		u16 common_status;

		peer_status = phy_read(mii_info, MII_LPA);
		local_status = phy_read(mii_info, MII_ADVERTISE);
		common_status = peer_status & local_status;

		if (common_status & LPA_100FULL) {
			mii_info->duplex = DUPLEX_FULL;
			mii_info->speed = SPEED_100;
		} else if (common_status & LPA_100HALF) {
			mii_info->duplex = DUPLEX_HALF;
			mii_info->speed = SPEED_100;
		} else if (common_status & LPA_10FULL) {
			mii_info->duplex = DUPLEX_FULL;
			mii_info->speed = SPEED_10;
		} else { 	/* 10 HD is the only choice left */
			mii_info->duplex = DUPLEX_HALF;
			mii_info->speed = SPEED_10;
		}
	}

#if 0
	/* old broken code */
	if (mii_info->autoneg) {
		u16 status;

		status = phy_read(mii_info, MII_LPA);

		if (status & (LPA_10FULL | LPA_100FULL))
			mii_info->duplex = DUPLEX_FULL;
		else
			mii_info->duplex = DUPLEX_HALF;
		if (status & (LPA_100FULL | LPA_100HALF))
			mii_info->speed = SPEED_100;
		else
			mii_info->speed = SPEED_10;
		mii_info->pause = 0;
	}
#endif
	/* On non-aneg, we assume what we put in BMCR is the speed,
	 * though magic-aneg shouldn't prevent this case from occurring
	 */

	return 0;
}
static int marvell_read_status(struct gfar_mii_info *mii_info)
{
	u16 status;
	int err;

	/* Update the link, but return if there
	 * was an error */
	err = genmii_update_link(mii_info);
	if (err)
		return err;

	/* If the link is up, read the speed and duplex */
	/* If we aren't autonegotiating, assume speeds 
	 * are as set */
	if (mii_info->autoneg && mii_info->link) {
		int speed;
		status = phy_read(mii_info, MII_M1011_PHY_SPEC_STATUS);

#if 0
		/* If speed and duplex aren't resolved,
		 * return an error.  Isn't this handled
		 * by checking aneg?
		 */
		if ((status & MII_M1011_PHY_SPEC_STATUS_RESOLVED) == 0)
			return -EAGAIN;
#endif

		/* Get the duplexity */
		if (status & MII_M1011_PHY_SPEC_STATUS_FULLDUPLEX)
			mii_info->duplex = DUPLEX_FULL;
		else
			mii_info->duplex = DUPLEX_HALF;

		/* Get the speed */
		speed = status & MII_M1011_PHY_SPEC_STATUS_SPD_MASK;
		switch(speed) {
			case MII_M1011_PHY_SPEC_STATUS_1000:
				mii_info->speed = SPEED_1000;
				break;
			case MII_M1011_PHY_SPEC_STATUS_100:
				mii_info->speed = SPEED_100;
				break;
			default:
				mii_info->speed = SPEED_10;
				break;
		}
		mii_info->pause = 0;
	}

	return 0;
}


static int cis820x_read_status(struct gfar_mii_info *mii_info)
{
	u16 status;
	int err;

	/* Update the link, but return if there
	 * was an error */
	err = genmii_update_link(mii_info);
	if (err)
		return err;

	/* If the link is up, read the speed and duplex */
	/* If we aren't autonegotiating, assume speeds 
	 * are as set */
	if (mii_info->autoneg && mii_info->link) {
		int speed;

		status = phy_read(mii_info, MII_CIS8201_AUX_CONSTAT);
		if (status & MII_CIS8201_AUXCONSTAT_DUPLEX)
			mii_info->duplex = DUPLEX_FULL;
		else
			mii_info->duplex = DUPLEX_HALF;

		speed = status & MII_CIS8201_AUXCONSTAT_SPEED;

		switch (speed) {
		case MII_CIS8201_AUXCONSTAT_GBIT:
			mii_info->speed = SPEED_1000;
			break;
		case MII_CIS8201_AUXCONSTAT_100:
			mii_info->speed = SPEED_100;
			break;
		default:
			mii_info->speed = SPEED_10;
			break;
		}
	}

	return 0;
}

static int marvell_ack_interrupt(struct gfar_mii_info *mii_info)
{
	/* Clear the interrupts by reading the reg */
	phy_read(mii_info, MII_M1011_IEVENT);

	return 0;
}

static int marvell_config_intr(struct gfar_mii_info *mii_info)
{
	if(mii_info->interrupts == MII_INTERRUPT_ENABLED)
		phy_write(mii_info, MII_M1011_IMASK, MII_M1011_IMASK_INIT);
	else
		phy_write(mii_info, MII_M1011_IMASK, MII_M1011_IMASK_CLEAR);

	return 0;
}

static int cis820x_init(struct gfar_mii_info *mii_info)
{
	phy_write(mii_info, MII_CIS8201_AUX_CONSTAT, 
			MII_CIS8201_AUXCONSTAT_INIT);
	phy_write(mii_info, MII_CIS8201_EXT_CON1,
			MII_CIS8201_EXTCON1_INIT);

	return 0;
}

static int cis820x_ack_interrupt(struct gfar_mii_info *mii_info)
{
	phy_read(mii_info, MII_CIS8201_ISTAT);

	return 0;
}

static int cis820x_config_intr(struct gfar_mii_info *mii_info)
{
	if(mii_info->interrupts == MII_INTERRUPT_ENABLED)
		phy_write(mii_info, MII_CIS8201_IMASK, MII_CIS8201_IMASK_MASK);
	else
		phy_write(mii_info, MII_CIS8201_IMASK, 0);

	return 0;
}

#define DM9161_DELAY 10

static int dm9161_read_status(struct gfar_mii_info *mii_info)
{
	u16 status;
	int err;

	/* Update the link, but return if there
	 * was an error */
	err = genmii_update_link(mii_info);
	if (err)
		return err;

	/* If the link is up, read the speed and duplex */
	/* If we aren't autonegotiating, assume speeds 
	 * are as set */
	if (mii_info->autoneg && mii_info->link) {
		status = phy_read(mii_info, MII_DM9161_SCSR);
		if (status & (MII_DM9161_SCSR_100F | MII_DM9161_SCSR_100H))
			mii_info->speed = SPEED_100;
		else
			mii_info->speed = SPEED_10;

		if (status & (MII_DM9161_SCSR_100F | MII_DM9161_SCSR_10F))
			mii_info->duplex = DUPLEX_FULL;
		else
			mii_info->duplex = DUPLEX_HALF;
	}

	return 0;
}


static int dm9161_config_aneg(struct gfar_mii_info *mii_info)
{
	struct dm9161_private *priv = mii_info->priv;

	if(0 == priv->resetdone)
		return -EAGAIN;

	return 0;
}

static void dm9161_timer(unsigned long data)
{
	struct gfar_mii_info *mii_info = (struct gfar_mii_info *)data;
	struct dm9161_private *priv = mii_info->priv;
	u16 status = phy_read(mii_info, MII_BMSR);

	if (status & BMSR_ANEGCOMPLETE) {
		priv->resetdone = 1;
	} else
		mod_timer(&priv->timer, jiffies + DM9161_DELAY * HZ);
}

static int dm9161_init(struct gfar_mii_info *mii_info)
{
	struct dm9161_private *priv;

	/* Allocate the private data structure */
	priv = kmalloc(sizeof(struct dm9161_private), GFP_KERNEL);

	if (NULL == priv)
		return -ENOMEM;

	mii_info->priv = priv;

	/* Reset is not done yet */
	priv->resetdone = 0;

	/* Isolate the PHY */
	phy_write(mii_info, MII_BMCR, BMCR_ISOLATE);

	/* Do not bypass the scrambler/descrambler */
	phy_write(mii_info, MII_DM9161_SCR, MII_DM9161_SCR_INIT);

	/* Clear 10BTCSR to default */
	phy_write(mii_info, MII_DM9161_10BTCSR, MII_DM9161_10BTCSR_INIT);

	/* Reconnect the PHY, and enable Autonegotiation */
	phy_write(mii_info, MII_BMCR, BMCR_ANENABLE);

	/* Start a timer for DM9161_DELAY seconds to wait
	 * for the PHY to be ready */
	init_timer(&priv->timer);
	priv->timer.function = &dm9161_timer;
	priv->timer.data = (unsigned long) mii_info;
	mod_timer(&priv->timer, jiffies + DM9161_DELAY * HZ);

	return 0;
}

static void dm9161_close(struct gfar_mii_info *mii_info)
{
	struct dm9161_private *priv = mii_info->priv;

	del_timer_sync(&priv->timer);
	kfree(priv);
}

#if 0
static int dm9161_ack_interrupt(struct gfar_mii_info *mii_info)
{
	phy_read(mii_info, MII_DM9161_INTR);

	return 0;
}
#endif

#ifdef CONFIG_SILKWORM_COMMON
/*
 * Although the BCM5461 does not need special init but the Brocade
 * silkworm board cannot put it in copper mode. This is a workaround.
 *
 */
static int bcm5421_init(struct gfar_mii_info * phy)
{
	u16 data;
	unsigned int id;

	id = (phy_read(phy, MII_PHYSID1) << 16 | phy_read(phy, MII_PHYSID2));

	/* Revision 0 of 5421 needs some fixups */
	if (id == 0x002060e0) {
		/* This is borrowed from MacOS
		 */
		phy_write(phy, 0x18, 0x1007);
		data = phy_read(phy, 0x18);
		phy_write(phy, 0x18, data | 0x0400);
		phy_write(phy, 0x18, 0x0007);
		data = phy_read(phy, 0x18);
		phy_write(phy, 0x18, data | 0x0800);
		phy_write(phy, 0x17, 0x000a);
		data = phy_read(phy, 0x15);
		phy_write(phy, 0x15, data | 0x0200);
	}

	/* Pick up some init code from OF for K2 version */
	if ((id & 0xfffffff0) == 0x002062e0) {
		phy_write(phy, 4, 0x01e1);
		phy_write(phy, 9, 0x0300);
	}

	/* Check if we can enable automatic low power */
#ifdef CONFIG_PPC_PMAC
	if (phy->platform_data) {
		struct device_node *np = of_get_parent(phy->platform_data);
		int can_low_power = 1;
		if (np == NULL || get_property(np, "no-autolowpower", NULL))
			can_low_power = 0;
		if (can_low_power) {
			/* Enable automatic low-power */
			phy_write(phy, 0x1c, 0x9002);
			phy_write(phy, 0x1c, 0xa821);
			phy_write(phy, 0x1c, 0x941d);
		}
	}
#endif /* CONFIG_PPC_PMAC */

	return 0;
}

static int marvell88e116_init(struct gfar_mii_info *mii_info)
{
	/* don't try to init the PHY if there is none */
	if (mii_info->mii_id == FSL_GIANFAR_BRD_HAS_NO_PHY)
		return 0;

	/* Reset the Phy first */
	(void)reset_mii_phy(mii_info);

	/* choose page 3 */
	phy_write(mii_info, 22, 0x3);

	/* programe LED to brocade convention LED[0] Link LED[1] activity */ 
	phy_write(mii_info, 16, 0x1040);

	/* restore page 0 */
	 phy_write(mii_info, 22, 0x0);

	return 0;
}

static int bcm5461_init(struct gfar_mii_info *mii_info)
{
	struct gfar_private *priv = netdev_priv(mii_info->dev);

	/* don't try to init the PHY if there is none */
	if (mii_info->mii_id == FSL_GIANFAR_BRD_HAS_NO_PHY)
		return 0;

	/* Reset the Phy first */
	(void)reset_mii_phy(mii_info);

	if (!(priv->einfo->board_flags & FSL_GIANFAR_BRD_COPPER_FORCED)) {
		/* The phy is already strapped in proper copper mode */
		return 0;
	}

	/* Forced into Copper mode */
	phy_write(mii_info, 0x1c, 0x7c00); 

	udelay(200);

	phy_write(mii_info, 0x1c, 0xfc00); 

	udelay(200);

	phy_write(mii_info, 0x1c, 0xfc08); 

	udelay(200);

	phy_write(mii_info, 0x0, 0x1140); 

	udelay(200);

	phy_write(mii_info, 0x1c, 0x7c00); 

	return 0;
}
static int reset_mii_phy(struct gfar_mii_info* phy)
{
	u16 val;
	int limit = 10000;
	
	val = phy_read(phy, MII_BMCR);
	val &= ~(BMCR_ISOLATE | BMCR_PDOWN);
	val |= BMCR_RESET;
	phy_write(phy, MII_BMCR, val);

	udelay(100);

	while (limit--) {
		val = phy_read(phy, MII_BMCR);
		if ((val & BMCR_RESET) == 0)
			break;
		udelay(10);
	}
	if ((val & BMCR_ISOLATE) && limit > 0)
		phy_write(phy, MII_BMCR, val & ~BMCR_ISOLATE);
	
	return (limit <= 0);
}

#if 0
static int bcm5221_init(struct gfar_mii_info* phy)
{
	u16 data;

	data = phy_read(phy, MII_BCM5221_TEST);
	phy_write(phy, MII_BCM5221_TEST,
		data | MII_BCM5221_TEST_ENABLE_SHADOWS);

	data = phy_read(phy, MII_BCM5221_SHDOW_AUX_STAT2);
	phy_write(phy, MII_BCM5221_SHDOW_AUX_STAT2,
		data | MII_BCM5221_SHDOW_AUX_STAT2_APD);

	data = phy_read(phy, MII_BCM5221_SHDOW_AUX_MODE4);
	phy_write(phy, MII_BCM5221_SHDOW_AUX_MODE4,
		data | MII_BCM5221_SHDOW_AUX_MODE4_CLKLOPWR);

	data = phy_read(phy, MII_BCM5221_TEST);
	phy_write(phy, MII_BCM5221_TEST,
		data & ~MII_BCM5221_TEST_ENABLE_SHADOWS);

	return 0;
}
#endif /* 0 */
#endif

/* Cicada 820x */
static struct phy_info phy_info_cis820x = {
	0x000fc440,
	"Cicada Cis8204",
	0x000fffc0,
	.features	= MII_GBIT_FEATURES,
	.init		= &cis820x_init,
	.config_aneg	= &gbit_config_aneg,
	.read_status	= &cis820x_read_status,
	.ack_interrupt	= &cis820x_ack_interrupt,
	.config_intr	= &cis820x_config_intr,
};

static struct phy_info phy_info_dm9161 = {
	.phy_id		= 0x0181b880,
	.name		= "Davicom DM9161E",
	.phy_id_mask	= 0x0ffffff0,
	.init		= dm9161_init,
	.config_aneg	= dm9161_config_aneg,
	.read_status	= dm9161_read_status,
	.close		= dm9161_close,
};

static struct phy_info phy_info_marvell = {
	.phy_id		= 0x01410c00,
	.phy_id_mask	= 0xffffff00,
	.name		= "Marvell 88E1101/88E1111",
	.features	= MII_GBIT_FEATURES,
	.config_aneg	= &marvell_config_aneg,
	.read_status	= &marvell_read_status,
	.ack_interrupt	= &marvell_ack_interrupt,
	.config_intr	= &marvell_config_intr,
};

#ifdef CONFIG_SILKWORM_COMMON
/* Broadcom BCM 5421 */
static struct phy_info phy_info_bcm5421 = {
	.phy_id		= 0x002060e0,
	.phy_id_mask	= 0xfffffff0,
	.name		= "BCM5421",
	.features	= MII_GBIT_FEATURES,
	.init		= bcm5421_init,
	.config_aneg	= gbit_config_aneg,
	.read_status	= bcm54xx_read_link
};

static struct phy_info phy_info_bcm5461= {
	.phy_id		= 0x002060c0,
	.phy_id_mask	= 0xfffffff0,
	.name		= "BCM5461",
	.features	= MII_GBIT_FEATURES,
	.init		= bcm5461_init,
	.config_aneg	= gbit_config_aneg,
	.read_status	= bcm54xx_read_link
};

static struct phy_info phy_info_bcm54610= {
    .phy_id     = 0x0143bd60,
    .phy_id_mask    = 0xfffffff0,
    .name       = "BCM54610",
    .features   = MII_GBIT_FEATURES,
    .init       = bcm5461_init,
    .config_aneg    = gbit_config_aneg,
    .read_status    = bcm54xx_read_link
};


static struct phy_info phy_info_bcm5221= {
	.phy_id		= 0x004061e0,
	.phy_id_mask	= 0xfffffff0,
	.name		= "BCM5221",
	.features	= MII_BASIC_FEATURES,
	/* .init		= bcm5221_init, */
	.config_aneg	= genmii_config_aneg,
	/* .read_status	= bcm54xx_read_link */
	.read_status	= genmii_read_status
};
/* MARVEL  MACE */
static struct phy_info phy_info_marvell88e116 = {
	.phy_id		= 0x01410e10,
	.phy_id_mask	= 0xfffffff0,
	.name		= "Marvell 88E1116",
	.features	= MII_GBIT_FEATURES,
	.init			= &marvell88e116_init,
	.config_aneg	= &marvell_config_aneg,
	.read_status	= &marvell_read_status,
	.ack_interrupt	= &marvell_ack_interrupt,
	.config_intr	= &marvell_config_intr,
};

#endif

static struct phy_info phy_info_genmii= {
	.phy_id		= 0x00000000,
	.phy_id_mask	= 0x00000000,
	.name		= "Generic MII",
	.features	= MII_BASIC_FEATURES,
	.config_aneg	= genmii_config_aneg,
	.read_status	= genmii_read_status,
};

static struct phy_info *phy_info[] = {
	&phy_info_cis820x,
	&phy_info_marvell,
	&phy_info_dm9161,
#ifdef CONFIG_SILKWORM_COMMON
	&phy_info_bcm5421,
	&phy_info_bcm5461,
	&phy_info_bcm54610,
	&phy_info_bcm5221,
	&phy_info_marvell88e116,
#endif
	&phy_info_genmii,
	NULL
};

u16 phy_read(struct gfar_mii_info *mii_info, u16 regnum)
{
	u16 retval;
	unsigned long flags;

	spin_lock_irqsave(&mii_info->mdio_lock, flags);
	retval = mii_info->mdio_read(mii_info->dev, mii_info->mii_id, regnum);
	spin_unlock_irqrestore(&mii_info->mdio_lock, flags);

	return retval;
}

void phy_write(struct gfar_mii_info *mii_info, u16 regnum, u16 val)
{
	unsigned long flags;

	spin_lock_irqsave(&mii_info->mdio_lock, flags);
	mii_info->mdio_write(mii_info->dev, 
			mii_info->mii_id, 
			regnum, val);
	spin_unlock_irqrestore(&mii_info->mdio_lock, flags);
}

/* Use the PHY ID registers to determine what type of PHY is attached
 * to device dev.  return a struct phy_info structure describing that PHY
 */
struct phy_info * get_phy_info(struct gfar_mii_info *mii_info)
{
	u16 phy_reg;
	u32 phy_ID;
	int i;
	struct phy_info *theInfo = NULL;
	struct net_device *dev = mii_info->dev;

	/* Grab the bits from PHYIR1, and put them in the upper half */
	phy_reg = phy_read(mii_info, MII_PHYSID1);
	phy_ID = (phy_reg & 0xffff) << 16;

	/* Grab the bits from PHYIR2, and put them in the lower half */
	phy_reg = phy_read(mii_info, MII_PHYSID2);
	phy_ID |= (phy_reg & 0xffff);

	/* loop through all the known PHY types, and find one that */
	/* matches the ID we read from the PHY. */
	for (i = 0; phy_info[i]; i++)
		if (phy_info[i]->phy_id == 
				(phy_ID & phy_info[i]->phy_id_mask)) {
			theInfo = phy_info[i];
			break;
		}

	/* This shouldn't happen, as we have generic PHY support */
	if (theInfo == NULL) {
		printk("%s: PHY id %x is not supported!\n", dev->name, phy_ID);
		return NULL;
	} else {
		printk("%s: PHY is %s (%x)\n", dev->name, theInfo->name,
		       phy_ID);
	}

	return theInfo;
}

#if 0
static int bcm54xx_setup_aneg(struct gfar_mii_info *phy, u32 advertise)
{
	u16 ctl, adv;
	
	phy->autoneg = 1;
	phy->speed = SPEED_10;
	phy->duplex = DUPLEX_HALF;
	phy->pause = 0;
	phy->advertising = advertise;

	/* Setup standard advertise */
	adv = phy_read(phy, MII_ADVERTISE);
	adv &= ~(ADVERTISE_ALL | ADVERTISE_100BASE4);
	if (advertise & ADVERTISED_10baseT_Half)
		adv |= ADVERTISE_10HALF;
	if (advertise & ADVERTISED_10baseT_Full)
		adv |= ADVERTISE_10FULL;
	if (advertise & ADVERTISED_100baseT_Half)
		adv |= ADVERTISE_100HALF;
	if (advertise & ADVERTISED_100baseT_Full)
		adv |= ADVERTISE_100FULL;
	phy_write(phy, MII_ADVERTISE, adv);

	/* Setup 1000BT advertise */
	adv = phy_read(phy, MII_1000BASETCONTROL);
	adv &= ~(MII_1000BASETCONTROL_FULLDUPLEXCAP|MII_1000BASETCONTROL_HALFDUPLEXCAP);
	if (advertise & SUPPORTED_1000baseT_Half)
		adv |= MII_1000BASETCONTROL_HALFDUPLEXCAP;
	if (advertise & SUPPORTED_1000baseT_Full)
		adv |= MII_1000BASETCONTROL_FULLDUPLEXCAP;
	phy_write(phy, MII_1000BASETCONTROL, adv);

	/* Start/Restart aneg */
	ctl = phy_read(phy, MII_BMCR);
	ctl |= (BMCR_ANENABLE | BMCR_ANRESTART);
	phy_write(phy, MII_BMCR, ctl);

	return 0;
}
#endif

static int bcm54xx_read_link(struct gfar_mii_info *phy)
{
	int err;
	int link_mode;	
	u16 val;

	/* don't try to read the PHY if there is none */
	if (phy->mii_id == FSL_GIANFAR_BRD_HAS_NO_PHY)
		return 0;

	/* Update the link, but return if there
	 * was an error */
	err = genmii_update_link(phy);
	if (err)
		return err;
	
	if (phy->autoneg) {
	    	val = phy_read(phy, MII_BCM5400_AUXSTATUS);
		link_mode = ((val & MII_BCM5400_AUXSTATUS_LINKMODE_MASK) >>
			     MII_BCM5400_AUXSTATUS_LINKMODE_SHIFT);
		phy->duplex = phy_BCM5400_link_table[link_mode][0] ? DUPLEX_FULL : DUPLEX_HALF;
		phy->speed = phy_BCM5400_link_table[link_mode][2] ?
				SPEED_1000 :
				(phy_BCM5400_link_table[link_mode][1] ? SPEED_100 : SPEED_10);
		val = phy_read(phy, MII_LPA);
		phy->pause = ((val & LPA_PAUSE) != 0);
	}
	/* On non-aneg, we assume what we put in BMCR is the speed,
	 * though magic-aneg shouldn't prevent this case from occurring
	 */

	return 0;
}

/*
 * Marvel PHY used in Mace platforms stuck rarely causes BP-CP connectivity
 * issues. To avoid that, we reset the PHY every time we retry XCP connection
 * in Lance blade driver. This is solution is borrowed from NOS where the
 * similar issue was seen with Marvel switch in teh Mercury chassis.
 */
void gianfar_reset_marvel_phy(char *name)
{
	struct net_device *ndev = NULL;
	struct gfar_private *priv = NULL;
	ndev = __dev_get_by_name(name);
	if (ndev) {
		priv = (struct gfar_private *) netdev_priv(ndev);
		if (priv) {
			if (reset_mii_phy(priv->mii_info)) {
				printk(KERN_ERR "MARVEL PHY reset failed\n");
			}
		}
	}
	
}	

EXPORT_SYMBOL(gianfar_reset_marvel_phy);
