/*
 * drivers/net/ibm_emac/ibm_emac_phy.c
 *
 * Driver for PowerPC 4xx on-chip ethernet controller, PHY support.
 * Borrowed from sungem_phy.c, though I only kept the generic MII
 * driver for now.
 * 
 * This file should be shared with other drivers or eventually
 * merged as the "low level" part of miilib
 * 
 * (c) 2003, Benjamin Herrenscmidt (benh@kernel.crashing.org)
 * (c) 2004-2005, Eugene Surovegin <ebs@ebshome.net>
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/netdevice.h>
#include <linux/mii.h>
#include <linux/ethtool.h>
#include <linux/delay.h>

#include <asm/ocp.h>

#include <brocade/platform.h>
#include <brocade/silkworm.h>
#include <brocade/silkworm_features.h>
#include "ibm_emac_core.h"
#include "ibm_emac_phy.h"

#define MII_BASIC_FEATURES	(SUPPORTED_10baseT_Half | SUPPORTED_10baseT_Full | \
				 SUPPORTED_100baseT_Half | SUPPORTED_100baseT_Full | \
				 SUPPORTED_Autoneg | SUPPORTED_TP | SUPPORTED_MII)
#define MII_GBIT_FEATURES	(MII_BASIC_FEATURES | \
				 SUPPORTED_1000baseT_Half | SUPPORTED_1000baseT_Full)

#define NL "\n"
#define PHY_DBG(f,x...) printk("emac" f, ##x)
extern sw_bd silkworm;

static inline int phy_read(struct mii_phy *phy, int reg)
{
	return phy->mdio_read(phy->dev, phy->address, reg);
}

static inline void phy_write(struct mii_phy *phy, int reg, int val)
{
	phy->mdio_write(phy->dev, phy->address, reg, val);
}

/*
 * polls MII_BMCR until BMCR_RESET bit clears or operation times out.
 *
 * returns:
 *	>= 0 => success, value in BMCR returned to caller
 *	-EBUSY => failure, RESET bit never cleared
 *	otherwise => failure, lower level PHY read failed
 */

static int mii_spin_reset_complete(struct mii_phy *phy)
{
	int val;
	int limit = 10000;

	while (limit--) {
		val = phy_read(phy, MII_BMCR);
		if ((val >= 0) && ((val & BMCR_RESET) == 0))
			return val;	/* success */
		udelay(10);
	}

	return (val < 0) ? val : -EBUSY;
}

int mii_reset_phy(struct mii_phy *phy)
{
	int val;

	val = phy_read(phy, MII_BMCR);
	val &= ~BMCR_ISOLATE;
	val |= BMCR_RESET;
	phy_write(phy, MII_BMCR, val);

	udelay(300);

	val = mii_spin_reset_complete(phy);

	if (val < 0) {
		PHY_DBG("%d: reset_complete failed in reset %d" NL,
			((struct ocp_enet_private *) (phy->dev->priv))->def->index, val);
	} else {
		if (val & BMCR_ISOLATE)
			phy_write(phy, MII_BMCR, val & ~BMCR_ISOLATE);
	}

	return val < 0;
}

static int genmii_setup_aneg(struct mii_phy *phy, u32 advertise)
{
	int ctl, adv;
	PHY_DBG("%d: genmii_setup_aneg" NL,
		((struct ocp_enet_private *) (phy->dev->priv))->def->index);
	//printk("0x%x: genmii_setup_aneg advertise\n", advertise);

	phy->autoneg = AUTONEG_ENABLE;
	phy->speed = SPEED_10;
	phy->duplex = DUPLEX_HALF;
	phy->pause = phy->asym_pause = 0;
	phy->advertising = advertise;

	/* Setup standard advertise */
	adv = phy_read(phy, MII_ADVERTISE);
	if (adv < 0)
		return adv;
	adv &= ~(ADVERTISE_ALL | ADVERTISE_100BASE4 | ADVERTISE_PAUSE_CAP |
		 ADVERTISE_PAUSE_ASYM);
	if (advertise & ADVERTISED_10baseT_Half)
		adv |= ADVERTISE_10HALF;
	if (advertise & ADVERTISED_10baseT_Full)
		adv |= ADVERTISE_10FULL;
	if (advertise & ADVERTISED_100baseT_Half)
		adv |= ADVERTISE_100HALF;
	if (advertise & ADVERTISED_100baseT_Full)
		adv |= ADVERTISE_100FULL;
	if (advertise & ADVERTISED_Pause)
		adv |= ADVERTISE_PAUSE_CAP;
	if (advertise & ADVERTISED_Asym_Pause)
		adv |= ADVERTISE_PAUSE_ASYM;
	phy_write(phy, MII_ADVERTISE, adv);

	printk(KERN_DEBUG "phy->features = %x, advertise =%x\n", phy->features, advertise);
	if (phy->features &
	    (SUPPORTED_1000baseT_Full | SUPPORTED_1000baseT_Half)) {
		adv = phy_read(phy, MII_CTRL1000);
		if (adv < 0)
			return adv;
		adv &= ~(ADVERTISE_1000FULL | ADVERTISE_1000HALF);
		if (silkworm.board != SWBD134) {
		if (advertise & ADVERTISED_1000baseT_Full)
			adv |= ADVERTISE_1000FULL;
		if (advertise & ADVERTISED_1000baseT_Half)
			adv |= ADVERTISE_1000HALF;
		}
		phy_write(phy, MII_CTRL1000, adv);
	}

	if (silkworm.board == SWBD134 && (phy->features & SUPPORTED_FIBRE)) {
		phy->speed = SPEED_1000;
		phy->duplex = DUPLEX_FULL;
		phy_write(phy, MII_NCONFIG, 0xfc0b);
		ctl = mii_spin_reset_complete(phy);
		printk(KERN_DEBUG "CR value 0x%x\n", ctl);
		if (ctl < 0) {
			PHY_DBG("%d: reset_complete failed in setup_aneg %d" NL,
					((struct ocp_enet_private *) (phy->dev->priv))->def->index, ctl);
		} else {
			phy_write(phy, MII_ADVERTISE, 0x01e0);
			ctl |= (BMCR_ANENABLE | BMCR_ANRESTART);
			phy_write(phy, MII_BMCR, ctl);
		}
	} else {
	/* Start/Restart aneg */
	/* on some PHYs (e.g. National DP83843) a write to MII_ADVERTISE
	 * causes BMCR_RESET to be set on the next read of MII_BMCR, which
	 * if not checked for causes the PHY to be reset below */
	ctl = mii_spin_reset_complete(phy);

	if (ctl < 0) {
		PHY_DBG("%d: reset_complete failed in setup_aneg %d" NL,
			((struct ocp_enet_private *) (phy->dev->priv))->def->index, ctl);
	} else {
		ctl |= (BMCR_ANENABLE | BMCR_ANRESTART);
		printk(KERN_DEBUG "write to CR ctl = 0x%x\n", ctl);
		phy_write(phy, MII_BMCR, ctl);
	}
	}

	return 0;
}

static int genmii_setup_forced(struct mii_phy *phy, int speed, int fd)
{
	int ctl;
	PHY_DBG("%d: genmii_setup_forced" NL,
		((struct ocp_enet_private *) (phy->dev->priv))->def->index);

	phy->autoneg = AUTONEG_DISABLE;
	phy->speed = speed;
	phy->duplex = fd;
	phy->pause = phy->asym_pause = 0;

	/* First reset the PHY */
	mii_reset_phy(phy);

	/* Perform fixups if applicable */
	if (phy->def->ops->init)
		phy->def->ops->init(phy);

	ctl = phy_read(phy, MII_BMCR);
	if (ctl < 0)
		return ctl;
	ctl &= ~(BMCR_FULLDPLX | BMCR_SPEED100 | BMCR_ANENABLE | BMCR_SPEED1000);

	/* Select speed & duplex */
	switch (speed) {
	case SPEED_10:
		break;
	case SPEED_100:
		ctl |= BMCR_SPEED100;
		break;
	case SPEED_1000:
		ctl |= BMCR_SPEED1000;
		break;
	default:
		return -EINVAL;
	}
	if (fd == DUPLEX_FULL)
		ctl |= BMCR_FULLDPLX;
	phy_write(phy, MII_BMCR, ctl);

	return 0;
}

static int genmii_poll_link(struct mii_phy *phy)
{
	int status;

	/* Clear latched value with dummy read */
	phy_read(phy, MII_BMSR);
	status = phy_read(phy, MII_BMSR);
	if (status < 0 || (status & BMSR_LSTATUS) == 0)
		return 0;
	if (phy->autoneg == AUTONEG_ENABLE && !(status & BMSR_ANEGCOMPLETE))
		return 0;
	return 1;
}

static int genmii_read_link(struct mii_phy *phy)
{
	if (phy->autoneg == AUTONEG_ENABLE) {
		if (silkworm.board != SWBD134 || (!(phy->features & SUPPORTED_FIBRE))) {
		int glpa = 0;
		int lpa = phy_read(phy, MII_LPA) & phy_read(phy, MII_ADVERTISE);
		if (lpa < 0)
			return lpa;

		if (phy->features &
		    (SUPPORTED_1000baseT_Full | SUPPORTED_1000baseT_Half)) {
			int adv = phy_read(phy, MII_CTRL1000);
			glpa = phy_read(phy, MII_STAT1000);

			if (glpa < 0 || adv < 0)
				return adv;

			glpa &= adv << 2;
		}

		phy->speed = SPEED_10;
		phy->duplex = DUPLEX_HALF;
		phy->pause = phy->asym_pause = 0;

		if (glpa & (LPA_1000FULL | LPA_1000HALF)) {
			phy->speed = SPEED_1000;
			if (glpa & LPA_1000FULL)
				phy->duplex = DUPLEX_FULL;
		} else if (lpa & (LPA_100FULL | LPA_100HALF)) {
			phy->speed = SPEED_100;
			if (lpa & LPA_100FULL)
				phy->duplex = DUPLEX_FULL;
		} else if (lpa & LPA_10FULL)
			phy->duplex = DUPLEX_FULL;

		if (phy->duplex == DUPLEX_FULL) {
			phy->pause = lpa & LPA_PAUSE_CAP ? 1 : 0;
			phy->asym_pause = lpa & LPA_PAUSE_ASYM ? 1 : 0;
		}
		} else {/*Qilin 1000Base-X auto neg*/
			int lpa = phy_read(phy, MII_LPA) & phy_read(phy, MII_ADVERTISE);
			printk(KERN_DEBUG "phy->speed %dM\n", phy->speed);
			phy_write(phy, MII_NCONFIG, 0xfc0b);
			if (lpa < 0) {
				printk(KERN_ERR "MII_LPA MII_ADVERTISE not matched\n");
				return lpa;
			}
			phy->duplex = DUPLEX_HALF;
			phy->pause = phy->asym_pause = 0;

			if (lpa & 0x0020)
				phy->duplex = DUPLEX_FULL;

			if (phy->duplex == DUPLEX_FULL) {
				phy->pause = lpa & LPA_PAUSE_CAP ? 1 : 0;
				phy->asym_pause = lpa & LPA_PAUSE_ASYM ? 1 : 0;
			}
		}
	} else {
		int bmcr = phy_read(phy, MII_BMCR);
		if (bmcr < 0)
			return bmcr;

		if (bmcr & BMCR_FULLDPLX)
			phy->duplex = DUPLEX_FULL;
		else
			phy->duplex = DUPLEX_HALF;
		if (bmcr & BMCR_SPEED1000)
			phy->speed = SPEED_1000;
		else if (bmcr & BMCR_SPEED100)
			phy->speed = SPEED_100;
		else
			phy->speed = SPEED_10;

		phy->pause = phy->asym_pause = 0;
	}
	return 0;
}

/* Generic implementation for most 10/100/1000 PHYs */
static struct mii_phy_ops generic_phy_ops = {
	.setup_aneg	= genmii_setup_aneg,
	.setup_forced	= genmii_setup_forced,
	.poll_link	= genmii_poll_link,
	.read_link	= genmii_read_link
};

static struct mii_phy_def genmii_phy_def = {
	.phy_id		= 0x00000000,
	.phy_id_mask	= 0x00000000,
	.name		= "Generic MII",
	.ops		= &generic_phy_ops
};

/* CIS8201 */
#define MII_CIS8201_10BTCSR	0x16
#define  TENBTCSR_ECHO_DISABLE	0x2000
#define MII_CIS8201_EPCR	0x17
#define  EPCR_MODE_MASK		0x3000
#define  EPCR_GMII_MODE		0x0000
#define  EPCR_RGMII_MODE	0x1000
#define  EPCR_TBI_MODE		0x2000
#define  EPCR_RTBI_MODE		0x3000
#define MII_CIS8201_ACSR	0x1c
#define  ACSR_PIN_PRIO_SELECT	0x0004

static int cis8201_init(struct mii_phy *phy)
{
	int epcr;

	epcr = phy_read(phy, MII_CIS8201_EPCR);
	if (epcr < 0)
		return epcr;

	epcr &= ~EPCR_MODE_MASK;

	switch (phy->mode) {
	case PHY_MODE_TBI:
		epcr |= EPCR_TBI_MODE;
		break;
	case PHY_MODE_RTBI:
		epcr |= EPCR_RTBI_MODE;
		break;
	case PHY_MODE_GMII:
		epcr |= EPCR_GMII_MODE;
		break;
	case PHY_MODE_RGMII:
	default:
		epcr |= EPCR_RGMII_MODE;
	}

	phy_write(phy, MII_CIS8201_EPCR, epcr);
	
	/* MII regs override strap pins */
	phy_write(phy, MII_CIS8201_ACSR, 
		  phy_read(phy, MII_CIS8201_ACSR) | ACSR_PIN_PRIO_SELECT);

	/* Disable TX_EN -> CRS echo mode, otherwise 10/HDX doesn't work */
	phy_write(phy, MII_CIS8201_10BTCSR,
		  phy_read(phy, MII_CIS8201_10BTCSR) | TENBTCSR_ECHO_DISABLE);

	return 0;
}

/* Marvell sepcific routines */
static int marvell_setup_aneg(struct mii_phy *phy, u32 advertise)
{
	u16 ctl, adv;
	
	phy->autoneg = 1;
	phy->speed = SPEED_10;
	phy->duplex = DUPLEX_HALF;
	phy->pause = 0;
	phy->advertising = advertise;

	/* Setup standard advertise */
	adv = phy_read(phy, MII_ADVERTISE);
	adv &= ~(ADVERTISE_ALL | ADVERTISE_100BASE4);
	if (advertise & ADVERTISED_10baseT_Half)
		adv |= ADVERTISE_10HALF;
	if (advertise & ADVERTISED_10baseT_Full)
		adv |= ADVERTISE_10FULL;
	if (advertise & ADVERTISED_100baseT_Half)
		adv |= ADVERTISE_100HALF;
	if (advertise & ADVERTISED_100baseT_Full)
		adv |= ADVERTISE_100FULL;
	phy_write(phy, MII_ADVERTISE, adv);

	/* Setup 1000BT advertise & enable crossover detect
	 * XXX How do we advertise 1000BT ? Darwin source is
	 * confusing here, they read from specific control and
	 * write to control... Someone has specs for those
	 * beasts ?
	 */
	adv = phy_read(phy, MII_M1011_PHY_SPEC_CONTROL);
	adv |= MII_M1011_PHY_SPEC_CONTROL_AUTO_MDIX;
	adv &= ~(MII_1000BASETCONTROL_FULLDUPLEXCAP |
			MII_1000BASETCONTROL_HALFDUPLEXCAP);
	if (advertise & SUPPORTED_1000baseT_Half)
		adv |= MII_1000BASETCONTROL_HALFDUPLEXCAP;
	if (advertise & SUPPORTED_1000baseT_Full)
		adv |= MII_1000BASETCONTROL_FULLDUPLEXCAP;
	phy_write(phy, MII_1000BASETCONTROL, adv);

	/* Start/Restart aneg */
	ctl = phy_read(phy, MII_BMCR);
	ctl |= (BMCR_ANENABLE | BMCR_ANRESTART);
	phy_write(phy, MII_BMCR, ctl);

	return 0;
}

static int marvell_setup_forced(struct mii_phy *phy, int speed, int fd)
{
	u16 ctl, ctl2;
	
	phy->autoneg = 0;
	phy->speed = speed;
	phy->duplex = fd;
	phy->pause = 0;

	ctl = phy_read(phy, MII_BMCR);
	ctl &= ~(BMCR_FULLDPLX|BMCR_SPEED100|BMCR_SPD2|BMCR_ANENABLE);
	ctl |= BMCR_RESET;

	/* Select speed & duplex */
	switch(speed) {
	case SPEED_10:
		break;
	case SPEED_100:
		ctl |= BMCR_SPEED100;
		break;
	/* I'm not sure about the one below, again, Darwin source is
	 * quite confusing and I lack chip specs
	 */
	case SPEED_1000:
		ctl |= BMCR_SPD2;
	}
	if (fd == DUPLEX_FULL)
		ctl |= BMCR_FULLDPLX;

	/* Disable crossover. Again, the way Apple does it is strange,
	 * though I don't assume they are wrong ;)
	 */
	ctl2 = phy_read(phy, MII_M1011_PHY_SPEC_CONTROL);
	ctl2 &= ~(MII_M1011_PHY_SPEC_CONTROL_MANUAL_MDIX |
		MII_M1011_PHY_SPEC_CONTROL_AUTO_MDIX |
		MII_1000BASETCONTROL_FULLDUPLEXCAP |
		MII_1000BASETCONTROL_HALFDUPLEXCAP);
	if (speed == SPEED_1000)
		ctl2 |= (fd == DUPLEX_FULL) ?
			MII_1000BASETCONTROL_FULLDUPLEXCAP :
			MII_1000BASETCONTROL_HALFDUPLEXCAP;
	phy_write(phy, MII_1000BASETCONTROL, ctl2);

	// XXX Should we set the sungem to GII now on 1000BT ?
	
	phy_write(phy, MII_BMCR, ctl);

	return 0;
}

static int marvell_read_link(struct mii_phy *phy)
{
	u16 status;

	if (phy->autoneg) {
		status = phy_read(phy, MII_M1011_PHY_SPEC_STATUS);
		if ((status & MII_M1011_PHY_SPEC_STATUS_RESOLVED) == 0)
			return -EAGAIN;
		if (status & MII_M1011_PHY_SPEC_STATUS_1000)
			phy->speed = SPEED_1000;
		else if (status & MII_M1011_PHY_SPEC_STATUS_100)
			phy->speed = SPEED_100;
		else
			phy->speed = SPEED_10;
		if (status & MII_M1011_PHY_SPEC_STATUS_FULLDUPLEX)
			phy->duplex = DUPLEX_FULL;
		else
			phy->duplex = DUPLEX_HALF;
		phy->pause = 0; /* XXX Check against spec ! */
	}
	/* On non-aneg, we assume what we put in BMCR is the speed,
	 * though magic-aneg shouldn't prevent this case from occurring
	 */

	return 0;
}

static int bcm5461_init(struct mii_phy* phy)
{
	unsigned int id;
	int phy_val; 
	int i; 
	/*default is copper mode*/
	int phy_mode=0;

	id = (phy_read(phy, MII_PHYSID1) << 16 | phy_read(phy, MII_PHYSID2));

	/* 5461 RGMII needs some fixups */
	if (id == 0x002060c1) {
		/* enable RGMII RXC delayed timing mode */
		phy_write(phy, 0x18, 0xf1e7);
		/* enable read/write clock alignment */
		phy_write (phy, 0x1c, 0x8e00);
	}
	
	switch (silkworm.board) {
        case SWBD117: /* Pharos */
	        if (phy->address == 1) {
                    phy_write(phy, 0x1c, 0xf866);
		}
	        break;
	case SWBD134: /* Qilin */
		phy_write(phy, MII_NCONFIG, 0xf866);

		phy_write(phy, MII_NCONFIG, 0x7c00);
		/*Check whether the fiber is enable or not*/
		for(i=0; i<50000; i++)
		{
			phy_val = phy_read(phy, MII_NCONFIG);
			if(phy_val&0x0010)
			{
				/*Fiber mode was detected*/
				phy_mode = 1;
				break;
			}
		}
		
		/*Qilin was connected to chassis, enable 1000M-X*/
		if(phy_mode)
		{
			phy->features |= SUPPORTED_FIBRE;
		 	phy->features |= SUPPORTED_1000baseT_Half;
			phy->features |= SUPPORTED_1000baseT_Full;
		  	phy->features |= SUPPORTED_Autoneg;
		
		  	/*Power down the copper interface*/	 
		  	phy_write(phy, MII_NCONFIG, 0xfc08);
	
		  	phy_val = phy_read(phy,MII_BMCR);
		  	phy_val |= 0x0800;
		  	phy_write(phy, MII_BMCR, phy_val);
	
		  	/*Power on the fiber interface*/
		  	phy_write(phy, MII_NCONFIG, 0xfc0b);
		  	phy_val = phy_read(phy,MII_BMCR);
		  	phy_val &= ~0x0800;
		  	phy_write(phy, MII_BMCR, phy_val);
		}
		/*100Mbase-T mode, enable auto negotiation*/
		else
		{
		 	phy->features |= SUPPORTED_Autoneg;
			phy_write(phy, MII_NCONFIG, 0xf800);
		}
		break;
	default:
		break;
	}

	return 0;

}

static struct mii_phy_ops cis8201_phy_ops = {
	.init		= cis8201_init,
	.setup_aneg	= genmii_setup_aneg,
	.setup_forced	= genmii_setup_forced,
	.poll_link	= genmii_poll_link,
	.read_link	= genmii_read_link
};

static struct mii_phy_def cis8201_phy_def = {
	.phy_id		= 0x000fc410,
	.phy_id_mask	= 0x000ffff0,
	.name		= "CIS8201 Gigabit Ethernet",
	.ops		= &cis8201_phy_ops
};

static struct mii_phy_ops marvell_phy_ops = {
	.setup_aneg	= marvell_setup_aneg,
	.setup_forced	= marvell_setup_forced,
	.poll_link	= genmii_poll_link,
	.read_link	= marvell_read_link
};

static struct mii_phy_def marvell_phy_def = {
	.phy_id		= 0x01410c00,
	.phy_id_mask	= 0xffffff00,
	.name		= "Marvell 88E308X",
	.features	= MII_GBIT_FEATURES,
	.magic_aneg	= 1,
	.ops		= &marvell_phy_ops
};

static struct mii_phy_def phy_info_bcm5421= {
        .phy_id         = 0x0143bc31,
        .phy_id_mask    = 0xffffffff,
        .name           = "BCM5421",
        .features       = MII_BASIC_FEATURES,
	.ops            = &generic_phy_ops
};

static struct mii_phy_ops bcm5461_phy_ops = {
	.init       = bcm5461_init,
	.setup_aneg = genmii_setup_aneg,
	.setup_forced   = genmii_setup_forced,
	.poll_link  = genmii_poll_link,
	.read_link  = genmii_read_link
};

static struct mii_phy_def bcm5461_phy_def = {
	.phy_id     = 0x002060c0,
	.phy_id_mask    = 0xfffffff0,
	.name       = "BCM5461",
	.ops        = &bcm5461_phy_ops
};

static struct mii_phy_def *mii_phy_table[] = {
	&cis8201_phy_def,
	&marvell_phy_def,
	&phy_info_bcm5421,
	&bcm5461_phy_def,
	&genmii_phy_def,
	NULL
};

int mii_phy_probe(struct mii_phy *phy, int address)
{
	struct mii_phy_def *def;
	int i;
	u32 id;

	phy->autoneg = AUTONEG_DISABLE;
	phy->advertising = 0;
	phy->address = address;
	phy->speed = SPEED_10;
	phy->duplex = DUPLEX_HALF;
	phy->pause = phy->asym_pause = 0;

	/* Take PHY out of isolate mode and reset it. */
	if (mii_reset_phy(phy))
		return -ENODEV;

	/* Read ID and find matching entry */
	id = (phy_read(phy, MII_PHYSID1) << 16) | phy_read(phy, MII_PHYSID2);
	for (i = 0; (def = mii_phy_table[i]) != NULL; i++) {
		if ((id & def->phy_id_mask) == def->phy_id)
			break;
	}
	/* Should never be NULL (we have a generic entry), but... */
	if (!def)
		return -ENODEV;

	phy->def = def;

	/* Determine PHY features if needed */
	phy->features = def->features;
	if (!phy->features) {
		u16 bmsr = phy_read(phy, MII_BMSR);
		if (bmsr & BMSR_ANEGCAPABLE)
			phy->features |= SUPPORTED_Autoneg;
		if (bmsr & BMSR_10HALF)
			phy->features |= SUPPORTED_10baseT_Half;
		if (bmsr & BMSR_10FULL)
			phy->features |= SUPPORTED_10baseT_Full;
		if (bmsr & BMSR_100HALF)
			phy->features |= SUPPORTED_100baseT_Half;
		if (bmsr & BMSR_100FULL)
			phy->features |= SUPPORTED_100baseT_Full;
		if (bmsr & BMSR_ESTATEN) {
			u16 esr = phy_read(phy, MII_ESTATUS);
			if (esr & ESTATUS_1000_TFULL)
				phy->features |= SUPPORTED_1000baseT_Full;
			if (esr & ESTATUS_1000_THALF)
				phy->features |= SUPPORTED_1000baseT_Half;
			if (esr & 0x8000) /*ESTATUS_1000_XFULL*/
			        phy->features |= 0x10000;
			if (esr & 0x4000) /*ESTATUS_1000_XFULL*/
			        phy->features |= 0x20000;
		}		
		phy->features |= SUPPORTED_MII;
	}

	/*Qilin enable the 1000M */
	if (silkworm.board == SWBD134) {
		phy->features |= SUPPORTED_1000baseT_Full;
		phy->features |= SUPPORTED_1000baseT_Half;
	}
	/* Setup default advertising */
	phy->advertising = phy->features;

	return 0;
}

MODULE_LICENSE("GPL");
