#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <asm/time.h>
#include <linux/kallsyms.h>

extern unsigned tb_ticks_per_jiffy;
unsigned tb_cycles_per_sec;

#define CYCLESFORPERCENT (tb_ticks_per_jiffy/100)  /* Just for calculating % */
cycles_t time_cycles = 0;

#define SIRQ_ARRAY_MAX_TOP  10  /* Maximum number of Softirq callers we want to monitor */
#define SIRQ_ARRAY_MAX_CUR  20  


typedef struct softirq_caller_struct {
	void  *softirqcaller;
	unsigned long long count;
	unsigned long max_cycles;
	unsigned long long cpu_cycles;
	unsigned int print;
} softirq_caller_struct_t;

int sort_callers(void *a, void *b);


softirq_caller_struct_t softirq_latency_cur[SIRQ_ARRAY_MAX_CUR];
softirq_caller_struct_t softirq_count_cur[SIRQ_ARRAY_MAX_CUR];
softirq_caller_struct_t softirq_latency_top[SIRQ_ARRAY_MAX_TOP];
softirq_caller_struct_t softirq_count_top[SIRQ_ARRAY_MAX_TOP];

static inline void shift_softirq(softirq_caller_struct_t *list, int start, int size)
{
	int i;

	if ((size-1) <= start)
		return;
	for (i=size-1; i>start; i--)
		memcpy(&list[i], &list[i-1], sizeof(softirq_caller_struct_t));
}

static inline int find_softirq_func(void *func, softirq_caller_struct_t *s, int size)
{
	int i;

	for (i=0; i<size; i++)
	{
		if (s[i].max_cycles == 0)
			return -1;
		if (s[i].softirqcaller == func)
			return i;
	}
	return -1;
}

static inline void swapsoftirq(softirq_caller_struct_t *x,softirq_caller_struct_t *y)
{
	softirq_caller_struct_t temp;

	memcpy(&temp, x, sizeof(softirq_caller_struct_t));
	memcpy(x, y, sizeof(softirq_caller_struct_t));
	memcpy(y, &temp, sizeof(softirq_caller_struct_t));
}
 
static inline cycles_t calc_cycles(cycles_t start, cycles_t end)
{
	if (end >= start)
		return (end - start);
	else
		return (0xffffffffUL - start + end + 1);
}

static int our_symbol_print(unsigned long addr, char *buffer)
{
	char str1[KSYM_NAME_LEN+1];
	const char *name;
	char *our_mod_name;
	unsigned long size, offset;

	name = kallsyms_lookup((unsigned long)(addr), 
			       &size,&offset,&our_mod_name,str1);

        if (!name)
		sprintf(buffer, "0x%lx", addr);
	else {
		if (our_mod_name)
			sprintf(buffer, "%s [%s]", name, our_mod_name);
		else
			sprintf(buffer, "%s", name);
	}
 
	return 0;
}

void kick_watchdog(void);
static void print_latency_softirqs(softirq_caller_struct_t *s)
{
	int i;
	char buf[256];
	unsigned int max_cpu = 0;
	unsigned int most_sig_percent =0, least_sig_percent=0;
		
	kick_watchdog();
	printk("   Maximum Softirq Latency (cycles)\n");
	kick_watchdog();
	printk("   Latency |  CPU %% | Softirq\n");
	kick_watchdog();
	for (i =0; i < SIRQ_ARRAY_MAX_TOP; i++) {
		max_cpu = softirq_latency_top[i].max_cycles / (unsigned int)(CYCLESFORPERCENT);
		most_sig_percent = (max_cpu / 100 );
		least_sig_percent = max_cpu - (most_sig_percent * 100);
		if (most_sig_percent >= 100)
		{
			most_sig_percent = 100;
			least_sig_percent = 0;
		}
		if (most_sig_percent < 50)
			break;
		our_symbol_print((unsigned long)(softirq_latency_top[i].softirqcaller), buf);
		kick_watchdog();
		printk("%10lu   %3u.%02u   %p  %s\n", softirq_latency_top[i].max_cycles, most_sig_percent, least_sig_percent, softirq_latency_top[i].softirqcaller, buf);
		kick_watchdog();
	}
	printk("\n");
	kick_watchdog();
}

static void print_count_softirqs(softirq_caller_struct_t *s)
{
	int i;
	char buf[256];
	unsigned int max_cpu = 0;
	unsigned int most_sig_percent =0, least_sig_percent=0;
		
	kick_watchdog();
	printk("Softirqs Activated Per Second\n");
	kick_watchdog();
	printk("Tot Cycles |  CPU %% |  Count | Softirq\n");
	kick_watchdog();
	for (i =0; i < SIRQ_ARRAY_MAX_TOP; i++) {
		max_cpu = (unsigned int) (s[i].cpu_cycles) / (unsigned int)(CYCLESFORPERCENT);
		most_sig_percent = (max_cpu / 100 );
		least_sig_percent = max_cpu - (most_sig_percent * 100);
		if (most_sig_percent >= 100)
		{
			most_sig_percent = 100;
			least_sig_percent = 0;
		}
		if (most_sig_percent < 50)
			break;
		our_symbol_print((unsigned long)(s[i].softirqcaller),buf);
		kick_watchdog();
		printk("%10llu   %3u.%02u %8llu   %p %s\n", s[i].cpu_cycles, most_sig_percent, least_sig_percent, s[i].count, s[i].softirqcaller, buf);
		kick_watchdog();
	}
	printk("\n");
	kick_watchdog();
}

void do_1sec_processing(void)
{
	int i, top_cnt, dup_index, again, print_lat, print_cnt;
	unsigned int max_cpu;

	/* insert into max latency list */
	top_cnt = 0;
	again = 0;
	print_lat = 0;

	/* go through top latency list to see if new entries from the 1 second list can be added */
	for (i=0; i < SIRQ_ARRAY_MAX_TOP; i++) {

		/* exit if nothing to do */
		if ((softirq_latency_top[SIRQ_ARRAY_MAX_TOP-1].max_cycles >= softirq_latency_cur[top_cnt].max_cycles) || 
				(top_cnt >= SIRQ_ARRAY_MAX_TOP))
			break;

		/* this happens when an entry is already higher on the list. Rerun with the next entry on current 1 second list */
		if (again)
		{
			i--;
			again = 0;
		}

		/* if we have a high latency entry, put it on the list */
		if (softirq_latency_top[i].max_cycles < softirq_latency_cur[top_cnt].max_cycles) {
			dup_index = find_softirq_func(softirq_latency_cur[top_cnt].softirqcaller, softirq_latency_top, SIRQ_ARRAY_MAX_TOP);
			/* check if function is already on the list */
			if (dup_index != -1)
			{
				/* shift the list down to insert */
				if (dup_index >= i)
				{
					/* don't update print field to avoid printing again */
					shift_softirq(softirq_latency_top, i, dup_index+1);
					softirq_latency_top[i].softirqcaller = softirq_latency_cur[top_cnt].softirqcaller;
					softirq_latency_top[i].count = softirq_latency_cur[top_cnt].count;
					softirq_latency_top[i].max_cycles = softirq_latency_cur[top_cnt].max_cycles;
					softirq_latency_top[i].cpu_cycles = softirq_latency_cur[top_cnt].cpu_cycles;
				}
				/* skip this current entry and try the next as existing entry is higher */
				else
					again = 1;
			}
			/* put a new entry on the list */
			else
			{
				shift_softirq(softirq_latency_top, i, SIRQ_ARRAY_MAX_TOP);
				memcpy(&softirq_latency_top[i], &softirq_latency_cur[top_cnt], sizeof(softirq_caller_struct_t));
			}
			top_cnt++;
		}
		if (softirq_latency_top[i].print == 0)
		{
			max_cpu = softirq_latency_top[i].max_cycles / (unsigned int)(CYCLESFORPERCENT) / 100;
			if (max_cpu > 50)
			{
				softirq_latency_top[i].print = 1;
				print_lat = 1;
			}
		}
	}

	/* insert into count list */
	top_cnt = 0;
	again = 0;
	print_cnt = 0;

	/* go through top max count list to see if new entries from the 1 second list can be added */
	for (i=0; i < SIRQ_ARRAY_MAX_TOP; i++) {

		/* exit if nothing to do */
		if ((softirq_count_top[SIRQ_ARRAY_MAX_TOP-1].cpu_cycles >= softirq_count_cur[top_cnt].cpu_cycles) || 
				(top_cnt >= SIRQ_ARRAY_MAX_TOP))
			break;

		/* this happens when an entry is already higher on the list. Rerun with the next entry on current 1 second list */
		if (again)
		{
			i--;
			again = 0;
		}

		/* skip entries with only 1 count as this is picked up in the latency list */
		if (softirq_count_cur[top_cnt].count == 1)
		{
			top_cnt++;
			again = 1;
			continue;
		}

		/* if we have a high latency entry, put it on the list */
		if (softirq_count_top[i].cpu_cycles < softirq_count_cur[top_cnt].cpu_cycles) {
			dup_index = find_softirq_func(softirq_count_cur[top_cnt].softirqcaller, softirq_count_top, SIRQ_ARRAY_MAX_TOP);

			/* check if function is already on the list */
			if (dup_index != -1)
			{
				/* skip this current entry and try the next as existing entry is higher */
				if (dup_index >= i)
				{
					/* don't update print field to avoid printing again */
					shift_softirq(softirq_count_top, i, dup_index+1);
					softirq_count_top[i].softirqcaller = softirq_count_cur[top_cnt].softirqcaller;
					softirq_count_top[i].count = softirq_count_cur[top_cnt].count;
					softirq_count_top[i].max_cycles = softirq_count_cur[top_cnt].max_cycles;
					softirq_count_top[i].cpu_cycles = softirq_count_cur[top_cnt].cpu_cycles;
				}
				/* skip this current entry and try the next as existing entry is higher */
				else
					again = 1;
			}
			/* put a new entry on the list */
			else
			{
				shift_softirq(softirq_count_top, i, SIRQ_ARRAY_MAX_TOP);
				memcpy(&softirq_count_top[i], &softirq_count_cur[top_cnt], sizeof(softirq_caller_struct_t));
			}
			top_cnt++;
		}
		if (softirq_count_top[i].print == 0)
		{
			max_cpu = (unsigned int)(softirq_count_top[i].cpu_cycles) / (unsigned int)(CYCLESFORPERCENT) / 100;
			if (max_cpu > 50)
			{
				softirq_count_top[i].print = 1;
				print_cnt = 1;
			}
		}
	}

	/* clear 1 second lists */
	memset(softirq_latency_cur, 0, sizeof(softirq_latency_cur));
	memset(softirq_count_cur, 0, sizeof(softirq_count_cur));

	if (print_lat)
		print_latency_softirqs(softirq_latency_top);
	if (print_cnt)
		print_count_softirqs(softirq_count_top);

}

void insert_softirq_caller(void *func, cycles_t start, cycles_t end)
{
	int i, dup_index;
	int found;
	cycles_t diff;
	softirq_caller_struct_t *s, tmp;

	diff = calc_cycles(start, end);

	/* insert into max latency list */
	s = softirq_latency_cur;
	dup_index = find_softirq_func(func, s, SIRQ_ARRAY_MAX_CUR);

	/* if function is already on list, then increment count and reorder it */
	if (dup_index != -1)
	{
		s[dup_index].cpu_cycles += diff;

		/* check if we need to reorder the list */
		if (s[dup_index].max_cycles < diff)
		{
			/* find new spot on list */
			for (i=0; i < SIRQ_ARRAY_MAX_CUR; i++) {
				if (s[i].max_cycles < diff)
					break;
			}

			/* if we need to move up, then shift everything down so we can move it to the proper spot */
			if (i < dup_index)
			{
				memcpy(&tmp, &s[dup_index], sizeof(softirq_caller_struct_t));
				tmp.max_cycles = diff;
				shift_softirq(s, i, dup_index+1);
				memcpy(&s[i], &tmp, sizeof(softirq_caller_struct_t));
			}
		}
	}
	/* not on list yet.  Add it to the ordered list */
	else
	{
		/* find correct spot to keep list ordered */
		for (i=0; i < SIRQ_ARRAY_MAX_CUR; i++) {
			if (s[i].max_cycles < diff) {
				/* insert new entry */
				shift_softirq(s, i, SIRQ_ARRAY_MAX_CUR);
				s[i].softirqcaller =  func;
				s[i].count = 1;
				s[i].max_cycles = diff;
				s[i].cpu_cycles = diff;
				s[i].print = 0;
				break;
			}
		}
	}

	/* insert into count list */
	s = softirq_count_cur;

	found = 0;

	/* check if function is already on the list */
	dup_index = find_softirq_func(func, s, SIRQ_ARRAY_MAX_CUR);
	if (dup_index != -1)
	{
		s[dup_index].count++;
		s[dup_index].cpu_cycles += diff;

		/* find correct spot on ordered list */
		for (i=0; i < SIRQ_ARRAY_MAX_CUR; i++) {
			if (s[i].cpu_cycles < s[dup_index].cpu_cycles)
				break;
		}

		/* shift everything down and put it on the list */
		if (i < dup_index)
		{
			memcpy(&tmp, &s[dup_index], sizeof(softirq_caller_struct_t));
			shift_softirq(s, i, dup_index+1);
			memcpy(&s[i], &tmp, sizeof(softirq_caller_struct_t));
		}
		found = 1;
	} 
	/* new funtion to add to list */
	else 
	{
		/* if we have a free spot, put it in immediately */
		for (i=0; i < SIRQ_ARRAY_MAX_CUR; i++) {
			if (s[i].cpu_cycles == 0 ) {
				s[i].softirqcaller =  func;
				s[i].count = 1;
				s[i].max_cycles = diff;
				s[i].cpu_cycles = diff;
				s[i].print = 0;
				found = 1;
				break;
			}
		}
	}

	/* out of space, so just replace the last entry */
	if (!found) {
		s[SIRQ_ARRAY_MAX_CUR-1].softirqcaller =  func;
		s[SIRQ_ARRAY_MAX_CUR-1].count = 1;
		s[SIRQ_ARRAY_MAX_CUR-1].max_cycles = diff;
		s[SIRQ_ARRAY_MAX_CUR-1].cpu_cycles = diff;
		s[SIRQ_ARRAY_MAX_CUR-1].print = 0;
	}

	/* check if 1 sec has elapsed, and do the calculations  */
	if (calc_cycles(time_cycles, end) >= tb_cycles_per_sec ) { 
		time_cycles = end;
		do_1sec_processing();
	}

}

EXPORT_SYMBOL(insert_softirq_caller);

static const char *desc_array[NR_SOFTIRQS] = {
	"HI",
	"TIMER",
	"NET_TX",
	"NET_RX",
	"SCSI",
	"TASKLET",
};

/*
 * /proc/softirqs  ... display the number of softirqs
 */

static int show_softirqs(struct seq_file *p, void *v)
{
	int i, j, ret = 0;
	char buf[256];
	unsigned int max_cpu = 0;
	unsigned int most_sig_percent =0, least_sig_percent=0;
		
	seq_printf(p, "                ");
	for_each_cpu(i)
		seq_printf(p, "CPU%-8d", i);
	seq_printf(p, "\n");
		
	for (i = 0; i < NR_SOFTIRQS; i++) {
		seq_printf(p, "%-9s", desc_array[i]);
		for_each_cpu(j)
			seq_printf(p, " %10u", kstat_softirqs_cpu(i, j));
		seq_printf(p, "\n");
	}
	seq_printf(p, "\n");

	seq_printf(p, "   Maximum Softirq Latency (cycles)\n");
	seq_printf(p, "   Latency |  CPU %% | Softirq\n");
	for (i =0; i < SIRQ_ARRAY_MAX_TOP; i++) {
		max_cpu = softirq_latency_top[i].max_cycles;
		seq_printf(p, "%10u  ", max_cpu);
		max_cpu = max_cpu / (unsigned int)(CYCLESFORPERCENT);
		most_sig_percent = (max_cpu / 100 );
		least_sig_percent = max_cpu - (most_sig_percent * 100);
		if (most_sig_percent >= 100)
		{
			most_sig_percent = 100;
			least_sig_percent = 0;
		}
		seq_printf(p, " %3u.", most_sig_percent);
		seq_printf(p, "%02u ", least_sig_percent);
		seq_printf(p, "  %p ", softirq_latency_top[i].softirqcaller);
		ret  = our_symbol_print((unsigned long)(softirq_latency_top[i].softirqcaller), buf);
		seq_printf(p, "%s\n", buf);
	}
	seq_printf(p, "\ncycles/sec=%d\n\n", tb_cycles_per_sec);
#if 0
	seq_printf(p, "   Latency |  CPU %% |   Count | Softirq\n");
	for (i =0; i < SIRQ_ARRAY_MAX_CUR; i++) 
	{
		if (softirq_latency_cur[i].max_cycles)
		{
			max_cpu = softirq_latency_cur[i].max_cycles;
			seq_printf(p, "%10u  ", max_cpu);
			max_cpu = max_cpu / (unsigned int)(CYCLESFORPERCENT);
			most_sig_percent = (max_cpu / 100 );
			least_sig_percent = max_cpu - (most_sig_percent * 100);
			if (most_sig_percent >= 100)
			{
				seq_printf(p, " 100.00 ");
			}
			else
			{
				seq_printf(p, " %3u.", most_sig_percent);
				seq_printf(p, "%02u ", least_sig_percent);
			}
			seq_printf(p, " %8llu ", softirq_latency_cur[i].count);		
			seq_printf(p, "  %p ", softirq_latency_cur[i].softirqcaller);
			ret  = our_symbol_print((unsigned long)(softirq_latency_cur[i].softirqcaller), buf);
			seq_printf(p, "%s\n", buf);
		}
	}
	seq_printf(p, "\n");
#endif

	seq_printf(p, "Softirqs Activated Per Second\n");
	seq_printf(p, "Tot Cycles |  CPU %% |  Count | Softirq\n");
	for (i =0; i < SIRQ_ARRAY_MAX_TOP; i++) {
		seq_printf(p, "%10llu ", softirq_count_top[i].cpu_cycles);
		max_cpu = (unsigned int) (softirq_count_top[i].cpu_cycles) / (unsigned int)(CYCLESFORPERCENT);
		most_sig_percent = (max_cpu / 100 );
		least_sig_percent = max_cpu - (most_sig_percent * 100);
		if (most_sig_percent >= 100)
		{
			most_sig_percent = 100;
			least_sig_percent = 0;
		}
		seq_printf(p, "  %3u.", most_sig_percent);
		seq_printf(p, "%02u ", least_sig_percent);
		seq_printf(p, "%8llu   ", softirq_count_top[i].count);		
		seq_printf(p, "%p ", softirq_count_top[i].softirqcaller);
		ret = our_symbol_print((unsigned long)(softirq_count_top[i].softirqcaller),buf);
		seq_printf(p, "%s\n", buf);
	}
	seq_printf(p, "\n");
	
#if 0
	seq_printf(p, "   Count |  CPU %% | Avg Cycles | Softirq\n");
	for (i =0; i < SIRQ_ARRAY_MAX_CUR; i++) {
		if (softirq_count_cur[i].count != 0 ) 
		{
			seq_printf(p, "%8llu   ", softirq_count_cur[i].count);		
			max_cpu = (unsigned int) (softirq_count_cur[i].max_cycles) / (unsigned int)(CYCLESFORPERCENT);
			most_sig_percent = (max_cpu / 100 );
			least_sig_percent = max_cpu - (most_sig_percent * 100);
			seq_printf(p, "%3u.", most_sig_percent);
			seq_printf(p, "%02u ", least_sig_percent);
			average_cpu = (unsigned int)(softirq_count_cur[i].cpu_cycles) / (unsigned int)(softirq_count_cur[i].count);
			seq_printf(p, " %11u ", average_cpu);
			seq_printf(p, "  %p ", softirq_count_cur[i].softirqcaller);
			ret = our_symbol_print((unsigned long)(softirq_count_cur[i].softirqcaller),buf);
			seq_printf(p, "%s\n", buf);
		}
	}
	seq_printf(p, "\n");
#endif
	return 0;
}

static int softirqs_open(struct inode *inode, struct file *file)
{
	return single_open(file, show_softirqs, NULL);
}

static int softirqs_write(struct file *file, const char __user *user, size_t insize, 
			  loff_t *offset)
{
	memset(softirq_latency_top, 0, sizeof(softirq_latency_top));
	memset(softirq_count_top, 0, sizeof(softirq_count_top));
	return  SIRQ_ARRAY_MAX_TOP;
}
		
static struct file_operations proc_softirqs_operations = {
	.open = softirqs_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
	.write = softirqs_write,
};

struct proc_dir_entry *sentry = NULL;
static int __init proc_softirqs_init(void)
{
/*	 proc_create("softirqs", 0, NULL, &proc_softirqs_operations); */
	sentry = create_proc_entry("softirqs", 0, NULL); 
	if (sentry) {
	  sentry->proc_fops = &proc_softirqs_operations;
	  //	  sentry->owner = THIS_MODULE;
	  printk(KERN_DEBUG "Inserted Softirq counter module\n");
	} else {
	  printk(KERN_ERR "Unable to create softirqs entry in /proc \n");
	}
	
	tb_cycles_per_sec = tb_ticks_per_jiffy * HZ;
	time_cycles = get_cycles();
	memset(softirq_latency_cur, 0, sizeof(softirq_latency_cur));
	memset(softirq_count_cur, 0, sizeof(softirq_count_cur));
	memset(softirq_latency_top, 0, sizeof(softirq_latency_top));
	memset(softirq_count_top, 0, sizeof(softirq_count_top));
	return 0;
}
module_init(proc_softirqs_init); 
